/*
 * dcs_io.h
 *
 * Copyright 2019-2023 Ichiji Tadokoro. All Rights Reserved.
 */

#ifndef _DCS_IO_H_
#define _DCS_IO_H_

#ifdef __cplusplus
extern "C" {
#endif

#include "dcs.h"

#define DCS_MAX_TIO		100	/* 最大ファイル・オープン数 */

/* オープン・モード */
#define DMD_READ		0	/* 入力モード("rt" "rb") */
#define DMD_WRITE		1	/* 出力モード("wt" "wb") */
#define DMD_APPEND		2	/* 追加モード("at" "ab") */
#define DMD_MODIFY		3	/* 更新モード("r+t" "r+b") */

/* 処理タイプ */
#define DOP_NOTHING		0	/* 未処理 */
#define DOP_COPY		1	/* コピー */
#define DOP_ADDBOM		2	/* BOM 追加 */
#define DOP_DELBOM		3	/* BOM 削除 */
#define DOP_CONV		4	/* コード変換 */

/* DCSW.DLL 共通 ======================================= */
/* 文字コード判定（ファイル・パス） */
long dcs_io_path(const char *spath, long stype);
/* 処理ステータスからメッセージを取得 */
long dcs_io_statusto_message(char *str, long sts, long ctype);
/* コード・タイプから名前を取得 */
long dcs_io_typeto_name(char *name, long ctype);
/* コード名からタイプを取得 */
long dcs_io_nameto_type(const char *name);
/* 不特定のコード名を取得 */
long dcs_io_multi_type_names(char *names, long size, long ctype);
/* すべてのコード名を取得 */
long dcs_io_all_type_names(char *names, long size);
/* すべてのコード名の長さを取得 */
long dcs_io_all_type_names_length();

/* DCS_IO 固有 ========================================= */
/* 入力モード判定 */
long dcs_io_is_can_read(long mode);
/* 出力モード判定 */
long dcs_io_is_can_write(long mode);
/* BOM 追加／削除判定 */
long dcs_io_cmp_bom(long stype, long dtype);
/* ファイルのコード変換 */
long dcs_io_convert_file(const char *spath, long stype, const char *dpath, long dtype, bool is_forced);

/* テキスト入出力オープン */
long dcs_io_open(const char *path, long mode, long stype, long dtype, bool is_forced, long *pstype);
/* 入力ファイル巻き戻し */
long dcs_io_rewind(long fd);
/* 変換後サイズ取得 */
long dcs_io_get_size(long fd);
/* 入力文字コード取得 */
long dcs_io_get_stype(long fd);
/* 出力文字コード取得 */
long dcs_io_get_dtype(long fd);
/* テキスト入出力１行入力（改行文字を除く）*/
long dcs_io_get_line(long fd, char *str, long size, long *psize);
/* エラー番号取得 */
long dcs_io_get_errno(long fd);
/* エラー・メッセージ取得 */
long dcs_io_get_strerror(long fd, long err_no, char *str, long size, long ctype);
/* テキスト入出力クローズ */
long dcs_io_close(long fd);

/* 処理タイプ取得 */
long dcs_io_get_ope_type();
/* 判定結果の取得 */
long dcs_io_is_decided();
/* 判定コード取得 */
long dcs_io_get_ctype();


#ifdef __cplusplus
}
#endif

#endif /* _DCS_IO_H_ */
