/*
 * _code_tables.c
 *
 * Copyright 2019-2023 Ichiji Tadokoro. All Rights Reserved.
 */

#include"_dcs.h"

#include "_euc2unibecommon_table00.h"
#include "_euc2unibe20932_table00.h"
#include "_euc2unibe20932_table8f.h"
#include "_euc2unibe51932_table00.h"
#include "_euc2unibems_table00.h"
#include "_euc2unibems_table8f.h"
#include "_jis2unibe50221_table.h"
#include "_jis2unibe2022jp2004_table.h"
#include "_sjis2unibe932_table.h"
#include "_unibe2euccommon_table.h"
#include "_unibe2euc20932_table.h"
#include "_unibe2euc51932_table.h"
#include "_unibe2eucms_table.h"
#include "_unibe2jis50221_table.h"
#include "_unibe2jis50221_table_bits.h"
#include "_unibe2jis2022jp2004_table.h"
#include "_unibe2jis2022jp2004_table_srg.h"
#include "_unibe2jis2022jp2004_table_bits.h"
#include "_unibe2sjis932_table.h"
#include "_unibe_def_table.h"

/**
 * 文字コード判定や文字コード変換等を行なうための Dcs 情報を生成する。
 * @param なし
 * @return 生成された Dcs 情報が返る。メモリー・アロケート・エラーの時は NULL。
 */
Dcs *create_dcs() {
	Dcs *dcs = calloc(sizeof(Dcs), 1);
	if (dcs != NULL) {
		if ((dcs->_block = ll_create()) == NULL) {
			release_dcs(dcs);
			return NULL;
		}
		dcs->_unibe_def_table = (uint16_t *)_unibe_def_table;
		dcs->_unibe_def_table_size = _unibe_def_table_size;
		dcs->_unibe2jis_table_srg[2] = (uint16_t *)_unibe2jis2022jp2004_table_srg;
		dcs->_unibe2jis_table_srg_size[2] = _unibe2jis2022jp2004_table_srg_size;
		dcs->_sjis2unibe_table = (uint16_t *)_sjis2unibe932_table;
		dcs->_sjis2unibe_table_size = _sjis2unibe932_table_size;
		dcs->_unibe2sjis_table = (uint16_t *)_unibe2sjis932_table;
		dcs->_unibe2sjis_table_size = _unibe2sjis932_table_size;
		set_jis50221_to(dcs);
		set_to_jis50221(dcs);
		set_euccommon_to(dcs);
		set_to_euccommon(dcs);
		dcs->version = DCS_VERSION;
	}
	return dcs;
}

/**
 * 生成された Dcs 情報を解放する。
 * @param 解放する Dcs 情報
 * @return なし。
 */
void release_dcs(Dcs *dcs) {
	if (dcs->_block != NULL) {
		ll_release((ll_list **)&dcs->_block);
	}
	free(dcs);
}
