/*
 * _jisto_jis.c
 *
 * Copyright 2019-2023 Ichiji Tadokoro. All Rights Reserved.
 */

#include "_dcs.h"

int _jisto_jis(Dcs *dcs) {
	int c, tsts, sts = DCS_SUCCESS;
	int flag_sshift = STYPE_ASCII;
	int flag_dshift = STYPE_ASCII;
	const unsigned char *tp, *sp, *ep;
	unsigned char *dbuf, *dp;
	int ltype = dcs->_type;
	ll_list *tlist, *clist = ll_first((ll_list *)dcs->_block);
	BufferInfo *tbinfo, *binfo = ll_get(clist);

	sp = binfo->buf + binfo->offset;
	ep = binfo->buf + binfo->length;
	reset_no_char(dcs, sp);
	dbuf = dp = _buffser_info.buf;

	for(;;) {
		/* 残りが最大長に満たない時、次の入力バッファに移動 */
		check_ll_buf(dcs, sp, ep, 4, clist, tlist, binfo, tbinfo);
		/* 出力バッファ・サイズを超える時、ライター出力 */
		check_info_buf_goto(dcs, dp, dbuf, sizeof(_buffser_info.buf) - _MAX_JIS_OFFSET, sts, loop_end);
		if (sp >= ep) {
			break;
		}
		tp = sp;
		if ((c = *sp++) == _ESC_CHAR) {	/* ｴｽｹｰﾌﾟ(0x1b) */
			if (sp + 2 > ep) {
				*dp++ = _ESC_CHAR;
				if (sp < ep) {
					*dp++ = *sp++;
					if (sp < ep) {
						*dp++ = *sp++;
					}
				}
				goto loop_end;
			}
			if ((c = *sp++) == '$') {       /* JIS漢字ｼﾌﾄIN */
				switch(c = *sp++) {
				case 'J':	/* 78JIS??? */
				case '@':	/* 78JIS */
				case 'B':	/* 83JIS */
					flag_sshift = STYPE_X0208_1983;
					break;
				case '(':	/* JIS X 0213 */
					if (sp + 1 > ep) {
						*dp++ = _ESC_CHAR;
						*dp++ = '$';
						*dp++ = '(';
						goto loop_end;
					} else {
						switch(c = *sp++) {
						case 'D': /* 補助漢字 */
							flag_sshift = STYPE_X0212;
							break;
						case 'Q': /* 第1面 */
						case 'O': /* JIS X 0213:2000第1面 */
							set_jis2022jp200401_to(dcs);
							flag_sshift = STYPE_X0213_2004;
							break;
						case 'P': /* 第2面 */
							set_jis2022jp200402_to(dcs);
							flag_sshift = STYPE_X0213_2000_2;
							break;
						default:
							sp--;
							*dp++ = _ESC_CHAR;
							*dp++ = '$';
							*dp++ = '(';
							break;
						}
					}
					break;
				default:
					sp--;
					*dp++ = _ESC_CHAR;
					*dp++ = '$';
					break;
				}
			} else if (c == '(') {  /* JIS漢字ｼﾌﾄOUT */
				switch(c = *sp++) {
				case 'J':	/* JISﾛｰﾏ字 */
				case 'H':	/* JISﾛｰﾏ字 */
				case 'B':	/* ASCII */
					flag_sshift = STYPE_ASCII;
					break;
				case 'I':	/* JIS半角ｶﾅ */
					flag_sshift = STYPE_KANA;
					break;
				default:
					sp--;
					*dp++ = _ESC_CHAR;
					*dp++ = '(';
					break;
				}
			} else {
				sp--;
				*dp++ = _ESC_CHAR;
				break;
			}
		} else { /* if ((c = *sp++) == _ESC_CHAR) */
			switch(flag_sshift) {
			case STYPE_X0212:
			case STYPE_X0208_1983:
			case STYPE_X0213_2004:
			case STYPE_X0213_2000_2:
				for(;;) {
					/* 残りが最大長に満たない時、次の入力バッファに移動 */
					check_ll_buf(dcs, sp, ep, 2, clist, tlist, binfo, tbinfo);
					/* 出力バッファ・サイズを超える時、ライター出力 */
					check_info_buf_goto(dcs, dp, dbuf, sizeof(_buffser_info.buf) - _MAX_JIS_OFFSET, sts, loop_end);
					if (sp >= ep) {
						sts = DCS_EUEXPEOD;
						goto loop_end;
					} else {
						c = ((c<<8)|(*sp++ & 0xff));
						jis_to_unibe_short_sp(dcs, dp, c, tp);
						unsigned short _ch2_, _ch_ = (unsigned short)unibe_char_to_short(dp);;
						if (_ch_ == NO_CHAR) {
						/* 不正文字の時 */
							/* flag_dshift が STYPE_X0208_1983 でなければシフトコードを出力 */
							_check_jis_to_zen(dp, flag_dshift);
							if (!dcs_is_suppress(dcs)) {
								*dp = NO_CHAR1;
								*(dp + 1) = NO_CHAR2;
								(dp) += 2;
							}
						} else if (is_unibe_surrogate_high(_ch_)) {
						/* Unicode(BE)サロゲート・ペア上位ビットの時 */
							/* flag_dshift が STYPE_X0208_1983 でなければシフトコードを出力 */
							_check_jis_to_zen(dp, flag_dshift);
							/* 変換後のJISはサロゲート文字ではない */
							unibe_to_jis_short_sp(dcs, dp, c, tp);
							(dp) += 2;
						} else {
							/* JIS to JIS の場合はテーブルは固定 */
							_ch2_ = (_ch_ >= (dcs)->_unibe2jis_table_size? 0: (dcs)->_unibe2jis_table[_ch_]);
							if (_ch2_ == 0 && (dcs)->_unibe2jis_table == (unsigned short *)_unibe2jis2022jp2004_table) {
								_ch2_ = (_ch_ >= (dcs)->_unibe2jis_table_srg_size[2]? 0: (dcs)->_unibe2jis_table_srg[2][_ch_]);
							}
							/* 変換 JIS コード・チェック */
							_check_conv_jis_sp(dcs, dp, _ch2_, _ch_, tp, flag_dshift);
						}
					}
					if (sp >= ep) {
						goto loop_end;
					} else if ((c = *sp++) == _ESC_CHAR) {
						sp--;
						break;
					}
					tp = sp - 1;
				}
				break;
			case STYPE_KANA:
				for(;;) {
					/* 残りが最大長に満たない時、次の入力バッファに移動 */
					check_ll_buf(dcs, sp, ep, 1, clist, tlist, binfo, tbinfo);
					/* 出力バッファ・サイズを超える時、ライター出力 */
					check_info_buf_goto(dcs, dp, dbuf, sizeof(_buffser_info.buf) - _MAX_JIS_OFFSET, sts, loop_end);
					if (is_jis_kana(c) == 0) {
						set_no_char_unibe(dcs, dp, c, sp-1);
						if (!dcs_is_suppress(dcs)) {
							*dp++ = NO_CHAR;
						}
					} else {
						/* JIS シフト 半角かな */
						_check_jis_to_kana(dp, flag_dshift);
						*dp++ = c;
					}
					if (sp >= ep) {
						goto loop_end;
					} else if (*sp == _ESC_CHAR) {
						break;
					}
					c = *sp++;
				}
				break;
			case STYPE_ASCII:
			default:
				for (;;) {
					/* 残りが最大長に満たない時、次の入力バッファに移動 */
					check_ll_buf(dcs, sp, ep, 2, clist, tlist, binfo, tbinfo);
					/* 出力バッファ・サイズを超える時、ライター出力 */
					check_info_buf_goto(dcs, dp, dbuf, sizeof(_buffser_info.buf) - _MAX_JIS_OFFSET, sts, loop_end);
					/* JIS シフト ASCII */
					_check_jis_to_ascii(dp, flag_dshift);
					if (c > _JIS_ROMAN_MAX) {
						if (!dcs_is_suppress(dcs)) {
							*dp++ = NO_CHAR;
						}
					} else if (c != _SO_CHAR) {
						cnvln_stype_to_jis(dp, c, sp, ep, ltype, jis);
					} else {		/* JIS半角ｶﾅｼﾌﾄOUT */
						for(;;) {
							/* 残りが最大長に満たない時、次の入力バッファに移動 */
							check_ll_buf(dcs, sp, ep, 1, clist, tlist, binfo, tbinfo);
							/* 出力バッファ・サイズを超える時、ライター出力 */
							check_info_buf_goto(dcs, dp, dbuf, sizeof(_buffser_info.buf) - _MAX_JIS_OFFSET, sts, loop_end);
							if (sp >= ep) {
								goto loop_end;
							} else if ((c = *sp++) == _SI_CHAR) {	/* JIS半角ｶﾅｼﾌﾄIN */
								break;
							} else if (is_jis_kana(c) == 0) {
								if (!dcs_is_suppress(dcs)) {
									*dp++ = NO_CHAR;
								}
								break;
							} else {
								*dp++ = c;
							}
						}
					}
					if (sp >= ep) {
						goto loop_end;
					} else if (*sp == _ESC_CHAR) {
						break;
					}
					c = *sp++;
				}
				break;
			}
		} /* if ((c = *sp++) == _ESC_CHAR) */
	}
loop_end:

	if (sts == DCS_SUCCESS || sts == DCS_EUEXPEOD) {
		if (flag_dshift != STYPE_ASCII) {
			*dp++ = (_ESC_CHAR);		/* ｴｽｹｰﾌﾟ(0x1b) */
			*dp++ = ('(');			/* JISｼﾌﾄOUT */
			*dp++ = (_SHIFT_OUT);	/* ﾛｰﾏ字 */
		}
		/* 出力バッファの残りをライター出力 */
		check_info_buf_rem(dcs, dp, dbuf, tsts);
		if (tsts != DCS_SUCCESS) {
			sts = tsts;
		}
	}
	DCS_RETURN(dcs, sts);
}
