/*
 * _jisto_unibe.c
 *
 * Copyright 2019-2023 Ichiji Tadokoro. All Rights Reserved.
 */

#include "_dcs.h"

int _jisto_unibe(Dcs *dcs, bool is_bom) {
	int c, tsts, sts = DCS_SUCCESS;
	int flag_shift = STYPE_ASCII;
	const unsigned char *tp, *sp, *ep;
	unsigned char *dbuf, *dp;
	int ltype = dcs->_type;
	ll_list *tlist, *clist = ll_first((ll_list *)dcs->_block);
	BufferInfo *tbinfo, *binfo = ll_get(clist);

	sp = binfo->buf + binfo->offset;
	ep = binfo->buf + binfo->length;
	reset_no_char(dcs, sp);
	dbuf = dp = _buffser_info.buf;
	if (is_bom) {
		*dp++ = UTF16_B0;
		*dp++ = UTF16_B1;
	}
	for(;;) {
		/* 残りが最大長に満たない時、次の入力バッファに移動 */
		check_ll_buf(dcs, sp, ep, 4, clist, tlist, binfo, tbinfo);
		/* 出力バッファ・サイズを超える時、ライター出力 */
		check_info_buf_goto(dcs, dp, dbuf, sizeof(_buffser_info.buf) - _MAX_UNIBE_OFFSET, sts, loop_end);
		if (sp >= ep) {
			break;
		}
		tp = sp;
		if ((c = *sp++) == _ESC_CHAR) {	/* ｴｽｹｰﾌﾟ(0x1b) */
			if (sp + 2 > ep) {
				*dp++ = _ESC_CHAR;
				if (sp < ep) {
					*dp++ = *sp++;
					if (sp < ep) {
						*dp++ = *sp++;
					}
				}
				goto loop_end;
			}
			if ((c = *sp++) == '$') {       /* JIS漢字ｼﾌﾄIN */
				switch(c = *sp++) {
				case 'J':	/* 78JIS??? */
				case '@':	/* 78JIS */
				case 'B':	/* 83JIS */
					flag_shift = STYPE_X0208_1983;
					break;
				case '(':	/* JIS X 0213 */
					if (sp + 1 > ep) {
						*dp++ = _ESC_CHAR;
						*dp++ = '$';
						*dp++ = '(';
						goto loop_end;
					} else {
						switch(c = *sp++) {
						case 'D': /* 補助漢字 */
							flag_shift = STYPE_X0212;
							break;
						case 'Q': /* 第1面 */
						case 'O': /* JIS X 0213:2000第1面 */
							set_jis2022jp200401_to(dcs);
							flag_shift = STYPE_X0213_2004;
							break;
						case 'P': /* 第2面 */
							set_jis2022jp200402_to(dcs);
							flag_shift = STYPE_X0213_2000_2;
							break;
						default:
							sp--;
							*dp++ = _ESC_CHAR;
							*dp++ = '$';
							*dp++ = '(';
							break;
						}
					}
					break;
				default:
					sp--;
					*dp++ = _ESC_CHAR;
					*dp++ = '$';
					break;
				}
			} else if (c == '(') {  /* JIS漢字ｼﾌﾄOUT */
				switch(c = *sp++) {
				case 'J':	/* JISﾛｰﾏ字 */
				case 'H':	/* JISﾛｰﾏ字 */
				case 'B':	/* ASCII */
					flag_shift = STYPE_ASCII;
					break;
				case 'I':	/* JIS半角ｶﾅ */
					flag_shift = STYPE_KANA;
					break;
				default:
					sp--;
					*dp++ = _ESC_CHAR;
					*dp++ = '(';
					break;
				}
			} else {
				sp--;
				*dp++ = _ESC_CHAR;
			}
		} else { /* if ((c = *sp++) == _ESC_CHAR) */
			switch(flag_shift) {
			case STYPE_X0212:
			case STYPE_X0208_1983:
			case STYPE_X0213_2004:
			case STYPE_X0213_2000_2:
				for(;;) {
					/* 残りが最大長に満たない時、次の入力バッファに移動 */
					check_ll_buf(dcs, sp, ep, 4, clist, tlist, binfo, tbinfo);
					/* 出力バッファ・サイズを超える時、ライター出力 */
					check_info_buf_goto(dcs, dp, dbuf, sizeof(_buffser_info.buf) - _MAX_UNIBE_OFFSET, sts, loop_end);
					if (sp >= ep) {
						sts = DCS_EUEXPEOD;
						goto loop_end;
					} else {
						c = ((c<<8)|(*sp++ & 0xff));
						jis_to_unibe_short_sp(dcs, dp, c, tp);
						c = char_to_short(dp);
						if (is_unibe_surrogate_high(c)) {
						/* サロゲート・ペアの時 */
							dp += 4;
						} else {
							if (!dcs_is_suppress(dcs) || !is_unibe_no_char(dp)) {
								dp += 2;
							}
						}
					}
					if (sp >= ep) {
						goto loop_end;
					} else if ((c = *sp++) == _ESC_CHAR) {
						sp--;
						break;
					}
					tp = sp - 1;
				}
				break;
			case STYPE_KANA:
				for(;;) {
					/* 残りが最大長に満たない時、次の入力バッファに移動 */
					check_ll_buf(dcs, sp, ep, 1, clist, tlist, binfo, tbinfo);
					/* 出力バッファ・サイズを超える時、ライター出力 */
					check_info_buf_goto(dcs, dp, dbuf, sizeof(_buffser_info.buf) - _MAX_UNIBE_OFFSET, sts, loop_end);
					if (is_jis_kana(c) == 0) {
						set_no_char_unibe(dcs, dp, c, sp-1);
						if (!dcs_is_suppress(dcs)) {
							*dp++ = 0;
							*dp++ = NO_CHAR;
						}
					} else {
						*dp++ = 0xff;
						*dp++ = jis_kana_to_unibe(c);
					}
					if (sp >= ep) {
						goto loop_end;
					} else if (*sp == _ESC_CHAR) {
						break;
					}
					c = *sp++;
				}
				break;
			case STYPE_ASCII:
			default:
				for (;;) {
					/* 残りが最大長に満たない時、次の入力バッファに移動 */
					check_ll_buf(dcs, sp, ep, 2, clist, tlist, binfo, tbinfo);
					/* 出力バッファ・サイズを超える時、ライター出力 */
					check_info_buf_goto(dcs, dp, dbuf, sizeof(_buffser_info.buf) - _MAX_UNIBE_OFFSET, sts, loop_end);
					if (c > _JIS_ROMAN_MAX) {
						if (!dcs_is_suppress(dcs)) {
							*dp++ = 0;
							*dp++ = NO_CHAR;
						}
					} else if (c != _SO_CHAR) {
						cnvln_stype_to_unibe(dp, c, sp, ep, ltype, jis);
					} else {		/* JIS半角ｶﾅｼﾌﾄOUT */
						for(;;) {
							/* 残りが最大長に満たない時、次の入力バッファに移動 */
							check_ll_buf(dcs, sp, ep, 1, clist, tlist, binfo, tbinfo);
							/* 出力バッファ・サイズを超える時、ライター出力 */
							check_info_buf_goto(dcs, dp, dbuf, sizeof(_buffser_info.buf) - _MAX_UNIBE_OFFSET, sts, loop_end);
							if (sp >= ep) {
								goto loop_end;
							} else if ((c = *sp++) == _SI_CHAR) {	/* JIS半角ｶﾅｼﾌﾄIN */
								break;
							} else if (is_jis_kana(c) == 0) {
								if (!dcs_is_suppress(dcs)) {
									*dp++ = 0;
									*dp++ = NO_CHAR;
								}
								break;
							} else {
								jis_to_unibe_short_sp(dcs, dp, c, tp);
								dp += 2;
							}
						}
					}
					if (sp >= ep) {
						goto loop_end;
					} else if (*sp == _ESC_CHAR) {
						break;
					}
					c = *sp++;
				}
				break;
			}
		} /* if ((c = *sp++) == _ESC_CHAR) */
	}
loop_end:

	if (sts == DCS_SUCCESS || sts == DCS_EUEXPEOD) {
		/* 出力バッファの残りをライター出力 */
		check_info_buf_rem(dcs, dp, dbuf, tsts);
		if (tsts != DCS_SUCCESS) {
			sts = tsts;
		}
	}
	DCS_RETURN(dcs, sts);
}
