/*
 * _unileto_euc.c
 *
 * Copyright 2019-2023 Ichiji Tadokoro. All Rights Reserved.
 */

#include "_dcs.h"

int _unileto_euc(Dcs *dcs) {
	int tsts, sts = DCS_SUCCESS;
	int c, c2, bytes;
	const unsigned char *sp, *ep;
	unsigned char *dbuf, *dp;
	int ltype = dcs->_type;
	ll_list *tlist, *clist = ll_first((ll_list *)dcs->_block);
	BufferInfo *tbinfo, *binfo = ll_get(clist);

	sp = binfo->buf + binfo->offset;
	ep = binfo->buf + binfo->length;
	reset_no_char(dcs, sp);
	dbuf = dp = _buffser_info.buf;

	for(;;) {
		/* 残りが最大長に満たない時、次の入力バッファに移動 */
		check_ll_buf(dcs, sp, ep, 4, clist, tlist, binfo, tbinfo);
		/* 出力バッファ・サイズを超える時、ライター出力 */
		check_info_buf_break(dcs, dp, dbuf, sizeof(_buffser_info.buf) - _MAX_EUC_OFFSET, sts);
		if (sp + 1 >= ep) {
			if (sp < ep) {
				sts = DCS_EUEXPEOD;
			}
			break;
		}
		c = unile_to_unibe_char(sp);
		sp += 2;
		if ((c & 0xff80) == 0) {
			cnvln_stype_to_euc(dp, c, sp, ep, ltype, unile);
		} else if (((c>>8)&0xff) == _ZEN_TILDE_0 && (c&0xff) == _ZEN_TILDE_1) {
		/* 全角'~'の時 */
			*dp++ = '~';
		} else {
			UNILE_NFD_TO_NFC(dcs, c, sp, ep);
			unibe_to_euc_short_sp(dcs, dp, c, sp - 2, bytes);
			if (!dcs_is_suppress(dcs) || !is_euc_no_char(dp)) {
				dp += bytes;
			}
			/* Unicode(BE)サロゲート・ペア上位ビットの判定 */
			if (is_unibe_surrogate_high(c)) {
				if ((sp + 1) >= ep) {
					sts = DCS_EUEXPEOD;
					break;
				}
				c2 = unile_to_unibe_char(sp);
				/* Unicode(BE)サロゲート・ペア下位ビットの判定 */
				if (is_unibe_surrogate_low(c2)) {
					sp += 2;
				}
			}
		}
	}
	if (sts == DCS_SUCCESS || sts == DCS_EUEXPEOD) {
		/* 出力バッファの残りをライター出力 */
		check_info_buf_rem(dcs, dp, dbuf, tsts);
		if (tsts != DCS_SUCCESS) {
			sts = tsts;
		}
	}
	DCS_RETURN(dcs, sts);
}
