/*
 * dcs_judges_convert.c
 *
 * Copyright 2019-2023 Ichiji Tadokoro. All Rights Reserved.
 */

#include <errno.h>
#include "_jdcs.h"
#include "_dcs.h"

/* 改行コード・タイプ UNIBE, UNILE 無効 */
#define judge_unicode(cinfo, tltype) \
	((cinfo)->type == DCS_UTF16BE || (cinfo)->type == DCS_UTF16LE? 0: tltype)

/* EOD の前 */
#define FILL_BUFFER_AND_RETURN(dcs, ltype, tltype, atype, sts, is_keep)	{ \
	if ((is_keep || (ltype) != NULL) && DCS_CONFIRMED_CODE_TYPE(sts)) { \
		int _sts_ = _fill_buffer(dcs, sts); \
		if (_sts_ != DCS_SUCCESS) { return _sts_; }} \
	CHECKLN_SETA_AND_RETURN(dcs, ltype, tltype, atype, sts); }

/* EOD の後 */
#define CHECKLN_SETA_AND_RETURN(dcs, ltype, tltype, atype, sts)	{ \
	if ((ltype) != NULL) { *ltype = _check_ln(dcs, tltype, sts); } \
	if ((atype) != NULL) { *(atype) = (sts); } return (sts); }

#define SETA_AND_RETURN(dcs, atype, sts)	{ \
	if ((atype) != NULL) { *(atype) = (sts); } return (sts); }

int _dcs_judges(Dcs *dcs, int *atype, int *ltype, int stype, bool is_forced, bool is_keep);
int _dcs_convert(Dcs *dcs, int stype, int dtype, bool is_forced);
int _check_ln(Dcs *dcs, int ltype, int stype);
int _fill_buffer(Dcs *dcs, int stype);
int _tcompare(const void *info1, const void *info2);

/*
 * コード判定（バッファー）
 */
int dcs_judges_buffer(Dcs *dcs, int *atype, int *ltype, const char *buffer, size_t size, int stype) {
	int sts;
	dcs->_sp = (unsigned char *)buffer;
	dcs->_ep = (unsigned char *)buffer + size;
	_set_data_type(dcs, _DTYPE_BUFFER);
	sts = _dcs_judges(dcs, atype, ltype, stype, false, ltype != NULL);
	_clear_data_type(dcs, _DTYPE_BUFFER);
	ll_clear(dcs->_block);
	DCS_RETURN(dcs, sts);
}

/*
 * コード判定（ストリ−ム）
 */
int dcs_judges_stream(Dcs *dcs, int *atype, int *ltype, FILE *sfp, int stype) {
	int sts;

	dcs->_sfp = sfp;
	_set_data_type(dcs, _DTYPE_STREAM);
	sts = _dcs_judges(dcs, atype, ltype, stype, false, ltype != NULL);
	_clear_data_type(dcs, _DTYPE_STREAM);
	ll_clear(dcs->_block);
	DCS_RETURN(dcs, sts);
}

/*
 * コード判定（ファイル）
 */
int dcs_judges_file(Dcs *dcs, int *atype, int *ltype, const char *spath, int stype) {
	int sts;
	FILE *fp = fopen(spath, "rb");
	if (fp == NULL) {
		dcs->_dcs_errno = errno;
		sts = DCS_EROPEN;
		DCS_RETURN(dcs, sts);
	}
	_set_data_type(dcs, _DTYPE_FILE);
	sts = dcs_judges_stream(dcs, atype, ltype, fp, stype);
	_clear_data_type(dcs, _DTYPE_FILE);
	fclose(fp);
	DCS_RETURN(dcs, sts);
}

/*
 * コード変換（バッファー）
 */
int dcs_convert_buffer(Dcs *dcs, const char *buffer, size_t size, int stype, int dtype, bool is_forced,
		int (*observer)(int sts, void *user_info), int (*writer)(const char *buf, size_t size, void *user_info), void *user_info) {
	int sts;
	dcs->_sp = (unsigned char *)buffer;
	dcs->_ep = (unsigned char *)buffer + size;
	dcs->_writer = writer;
	dcs->_observer = observer;
	dcs->_user_info = user_info;
	_set_data_type(dcs, _DTYPE_BUFFER);
	sts = _dcs_convert(dcs, stype, dtype, is_forced);
	_clear_data_type(dcs, _DTYPE_BUFFER);
	ll_clear(dcs->_block);
	DCS_RETURN(dcs, sts);
}

/*
 * コード変換（ストリーム）
 */
int dcs_convert_stream(Dcs *dcs, FILE *sfp, int stype, int dtype, bool is_forced,
		int (*observer)(int sts, void *user_info), int (*writer)(const char *buf, size_t size, void *user_info), void *user_info) {
	int sts;
	dcs->_sfp = sfp;
	dcs->_writer = writer;
	dcs->_observer = observer;
	dcs->_user_info = user_info;
	_set_data_type(dcs, _DTYPE_STREAM);
	sts = _dcs_convert(dcs, stype, dtype, is_forced);
	_clear_data_type(dcs, _DTYPE_STREAM);
	ll_clear(dcs->_block);
	DCS_RETURN(dcs, sts);
}

/*
 * コード変換（ファイル）
 */
int dcs_convert_file(Dcs *dcs, const char *spath, int stype, int dtype, bool is_forced,
		int (*observer)(int sts, void *user_info), int (*writer)(const char *buf, size_t size, void *user_info), void *user_info) {
	int sts;
	FILE *fp = fopen(spath, "rb");
	if (fp == NULL) {
		dcs->_dcs_errno = errno;
		sts = DCS_EROPEN;
		DCS_RETURN(dcs, sts);
	}
	_set_data_type(dcs, _DTYPE_FILE);
	sts = dcs_convert_stream(dcs, fp, stype, dtype, is_forced, observer, writer, user_info);
	_clear_data_type(dcs, _DTYPE_FILE);
	fclose(fp);
	dcs->_dcs_errno = errno;
	DCS_RETURN(dcs, sts);
}

BufferInfo _buffser_info;

#define get_one_buffer(dcs, sp, ep, alen, rlen) { \
	if (_is_type_buffer(dcs)) { \
	/* バッファーの時（JNI共通） */ \
		rlen = min(ep - sp, dcs->_ep - dcs->_sp); \
		if (ep - sp < dcs->_ep - dcs->_sp) { \
			rlen = ep - sp; \
		} else { \
			rlen = dcs->_ep - dcs->_sp; \
			/* 長さが等しくても EOF */ \
			is_eof = true; \
		} \
		memcpy(sp, dcs->_sp, rlen); \
		dcs->_sp += rlen; \
		sp += rlen; \
	} else if (_is_type_stream(dcs)) { \
		/* ストリームの時（ファイル、JNIファイル共通） */ \
		rlen = fread(sp, 1, ep - sp, dcs->_sfp); \
		if (feof(dcs->_sfp)) { \
			is_eof = true; \
		} else if(ferror(dcs->_sfp)) { \
			dcs->_dcs_errno = errno; \
			ll_clear((ll_list *)dcs->_block); \
			return DCS_EREAD; \
		} \
		sp += rlen; \
	} else {/* if (_is_type_jni_stream(dcs)) */ \
	/* JNI ストリームの時 */ \
		rlen = min(ep - sp, alen); \
		rlen = (*jni_info->jenv)->CallIntMethod( \
							jni_info->jenv, \
							jni_info->jistream, \
							jni_info->mid_buf, \
							jni_info->jbuf, \
							0, \
							rlen); \
		/* IOException とは限らない */ \
		if ((*jni_info->jenv)->ExceptionCheck(jni_info->jenv)) { \
			(*jni_info->jenv)->ExceptionClear(jni_info->jenv); \
			ll_clear((ll_list *)dcs->_block); \
			return DCS_EREAD; \
		} \
		if (rlen < 0) { \
			is_eof = true; \
		} else { \
			(*jni_info->jenv)->GetByteArrayRegion(jni_info->jenv, jni_info->jbuf, 0, rlen, sp); \
			sp += rlen; \
		} \
	} }

#define _DEB_TRACE	0

/* 共通コード判定 */
int _dcs_judges(Dcs *dcs, int *atype, int *ltype, int stype, bool is_forced, bool is_keep) {
	int i, three_bytes;
	int c, c2, c3, c4, ch;
	unsigned char *sp, *ep, *tp, *han_kana_p;
	int ecnt;
	int err_no;
	int32_t maxb;
	int flag_btype;
	int kana_bits;
	int32_t ubacnt, ulacnt, ubacnt_all, ulacnt_all;
	int32_t uba_ulz_cnt, ula_ubz_cnt;
	int dtype, tltype;
	int ub_mac_cnt, ul_mac_cnt, u8_mac_cnt;
	int ub_surrog_cnt, ul_surrog_cnt, u8_surrog_cnt;
	long long _source_size;
	bool is_ascii, is_all_ascii, is_eof, is_jis;
	bool is_utfmac_d;
	ll_list *clist = ll_first((ll_list *)dcs->_block);
	CodeInfo code_info[_I_MAX];
	CodeInfo *sjis_info = &code_info[_I_SJIS];
	CodeInfo *euc_info = &code_info[_I_EUC];
	CodeInfo *euc1_info = &code_info[_I_EUC1];
	CodeInfo *euc2_info = &code_info[_I_EUC2];
	CodeInfo *euc3_info = &code_info[_I_EUC3];
	CodeInfo *ub_info = &code_info[_I_UTF16BE];
	CodeInfo *ul_info = &code_info[_I_UTF16LE];
	CodeInfo *u8_info = &code_info[_I_UTF8];
	CodeInfo *cinfo_sp[] = {
			sjis_info, euc_info, ub_info, u8_info,
	};
	CodeInfo *cinfo[] = {
			&code_info[0], &code_info[1], &code_info[2],
			&code_info[3], &code_info[4], &code_info[5],
			&code_info[6], &code_info[7],
	};
	int types[] = {
			DCS_SJIS, DCS_EUC, DCS_EUC1, DCS_EUC2, DCS_EUC3,
			DCS_UTF16BE, DCS_UTF16LE, DCS_UTF8,
	};
	int nd_types[] = {
			DCS_ND_SJIS, DCS_ND_EUC, DCS_ND_EUC1, DCS_ND_EUC2, DCS_ND_EUC3,
			DCS_ND_UTF16BE, DCS_ND_UTF16LE, DCS_ND_UTF8,
	};
	JniJudgedStreamInfo * jni_info = (JniJudgedStreamInfo *)dcs->_jni_info;
	size_t rlen, alen = (_is_type_jni_stream(dcs)?
			(*jni_info->jenv)->GetArrayLength(jni_info->jenv, jni_info->jbuf): 0);

	if (atype != NULL) {
		*atype = DCS_UNKNOWN;
	}
	if (ltype != NULL) {
		*ltype = tltype = LTYPE_NON;
	}
	memset(&_buffser_info, 0, sizeof(_buffser_info));
	dcs->_is_surrogate = dcs->_utfmac_exists = false;
	sp = _buffser_info.buf;
	*sp++ = c = _dcs_getc(dcs);
	if (c == EOF) {
		err_no = errno;
		if (_is_type_file(dcs) && ferror(dcs->_sfp)) {
		/* ファイル・アクセス・エラーの時 */
			dcs->_dcs_errno = err_no;
			ll_clear((ll_list *)dcs->_block);
			return DCS_EREAD;
		}
		if ((atype) != NULL) {
			*(atype) = (DCS_EMPTY);
		}
		return (DCS_EMPTY);
	}
	*sp++ = c2 = _dcs_getc(dcs);
	if (c == UTF16_B0 && c2 == UTF16_B1) {
	/* UTF16BE BOM がある時 */
		FILL_BUFFER_AND_RETURN(dcs, ltype, tltype, atype, DCS_UTF16BE_BOM, is_keep);
	}
	if (c == UTF16_B1 && c2 == UTF16_B0) {
	/* UTF16LE BOM がある時 */
		FILL_BUFFER_AND_RETURN(dcs, ltype, tltype, atype, DCS_UTF16LE_BOM, is_keep);
	}
	c3 = 0;
	if (c == UTF8_B0 && c2 == UTF8_B1) {
		*sp++ = c3 = _dcs_getc(dcs);
		if (c3 == UTF8_B2) {
		/* UTF8 BOM がある時 */
			FILL_BUFFER_AND_RETURN(dcs, ltype, tltype, atype, DCS_UTF8_BOM, is_keep);
		}
	}
	if (c2 == EOF || c3 == EOF) {
		err_no = errno;
		if (_is_type_file(dcs) && ferror(dcs->_sfp)) {
		/* ファイル・アクセス・エラーの時 */
			dcs->_dcs_errno = err_no;
			ll_clear((ll_list *)dcs->_block);
			return DCS_EREAD;
		}
		sp--;
	}
	memset(code_info, 0, sizeof(code_info));
	for(i=0; i<sizeof(code_info)/sizeof(code_info[0]); i++) {
		code_info[i].type = types[i];
		code_info[i].nd_type = nd_types[i];
	}
	kana_bits = 0;
	set_jis50221_to(dcs);
	set_euccommon_to(dcs);
	is_eof = is_jis = false;
	is_all_ascii = true;
	ub_mac_cnt = ul_mac_cnt = u8_mac_cnt = 0;
	ub_surrog_cnt = ul_surrog_cnt = u8_surrog_cnt = 0;
	ubacnt_all = ulacnt_all = uba_ulz_cnt = ula_ubz_cnt = 0;
	_source_size = 0;

	for(;;) {
		is_ascii = true;
		ubacnt = ulacnt = 0;
		flag_btype = (DCS_ND_ALL & ~(DCS_ND_JIS | DCS_ND_UTF_BOM));
		ep = _buffser_info.buf + sizeof(_buffser_info.buf) - _MAX_OFFSET;
		get_one_buffer(dcs, sp, ep, alen, rlen);
		ep = sp;
		_buffser_info.length = ep - _buffser_info.buf;
		if (is_keep) {
		/* バッファーを保持する時 */
			if ((clist = _add_buffer(clist, &_buffser_info, sizeof(_buffser_info))) == NULL) {
				ll_clear((ll_list *)dcs->_block);
				return DCS_EALLOC;
			}
		}
		/* JIS, SJIS の解析	 */
		sp = han_kana_p = _buffser_info.buf + _buffser_info.s_offset;
		sjis_info->blen = 0;
		_source_size += (ep - sp);
		while(sp < ep) {
			c = *sp++;
			if ((c & ~0x7f) == 0) {
			/* サイン・ビットが立っていない時 */
				if (c == 0) {
				/* UTF16 のASCII がある時 */
					if (((sp - _buffser_info.buf) & 01) == 0) {
					/* UTF16LE の ASCII がある時 */
						ulacnt++;
						ch = *(sp - 2);
						if (is_unibestd_short(dcs, ch)) {
						/* UTF16BE の全角が有効の時 */
							ula_ubz_cnt++;
						}
					} else {
					/* UTF16BE の ASCII がある時 */
						if (sp >= ep) {
						/* 次のバイトがない時 */
							break;
						}
						ubacnt++;
						ch = *sp;
						if (is_unibestd_short(dcs, ch)) {
						/* UTF16LE の全角が有効の時 */
							uba_ulz_cnt++;
						}
						sp++;
					}
					continue;
				}
				if ((c & ~0x1f) == 0) {
				/* 制御文字の時 */
					if (c == _ESC_CHAR) {	/* ｴｽｹｰﾌﾟ(0x1b) */
						if (sp + 2 > ep) {
							sp += 2;
						} else {
							c2 = *sp++;
							if (c2 == '$') {	/* JIS漢字ｼﾌﾄIN */
								c3 = *sp++;
								switch(c3) {
								case 'J':	/* 78JIS??? */
								case '@':	/* 78JIS */
								case 'B':	/* 83JIS */
									is_jis = true;
									break;
								case '(':	/* JIS X 0213 */
									if (sp + 1 > ep) {
										sp++;
									} else {
										c3 = *sp++;
										switch(c3) {
										case 'D': /* 補助漢字 */
											is_jis = true;
											break;
										case 'Q':	/* 第1面 */
										case 'O': /* JIS X 0213:2000第1面 */
										case 'P':	/* 第2面 */
											if (!_BINARY_RANGE(sjis_info->bad_cnt, _source_size)) {
											/* バイナリー域ではない時（SJISで代用） */
												FILL_BUFFER_AND_RETURN(dcs, ltype, tltype, atype, DCS_JIS2, is_keep);
											}
											break;
										default:
											sp--;
											break;
										}
									}
									break;
								default:
									sp--;
									break;
								}
							} else if (c2 == '(') {	/* JIS漢字ｼﾌﾄOUT */
								c3 = *sp++;
								switch(c3) {
								case 'J':	/* JISﾛｰﾏ字 */
								case 'H':	/* JISﾛｰﾏ字 */
								case 'B':	/* ASCII */
								case 'I':	/* JIS半角ｶﾅ */
									is_jis = true;
									break;
								default:
									sp--;
									break;
								}
							} else {
								sp--;
							}
						}
					} else if (c == '\r') {
						if (sp >= ep) {
							euc_info->blen = 1;
						}
					}
				}
				continue;
			} else {
			/* サイン・ビットが立っている時 */
				/* ASCII をクリア */
				is_ascii = is_all_ascii = false;
				if (is_sjis_kana(c)) {
				/* 半角カナの時 */
					if (han_kana_p < sp - 2
					 && (sp >= ep || (c2 = *sp, !is_sjis_kana(c2)))) {
					/* 半角カナが単一の時 */
						sjis_info->tan_han_kana_cnt++;
					}
					han_kana_p = sp - 1;
				} else if (sp >= ep) {
				/* 泣き別れの時 */
					if (is_eof) {
						sjis_info->bad_cnt += 2;
					}
					sjis_info->blen = 1;
					break;
				} else {
					c2 = *sp++;
					ch = ((c << 8)|c2);
					if (is_sjis_short(dcs, ch)) {
					/* SJIS 全角の時 */
						if (is_sjis_zenkana(ch)) {
						/* SJIS 全角かなの時 */
							kana_bits |= DCS_ND_SJIS;
							sjis_info->zen_kana_cnt += 2;
						}
						sjis_info->zen_cnt += 2;
					} else {
						sjis_info->bad_cnt += 2;
					}
				}
			}
		} /* while(sp < ep) */
		if (is_eof) {
			if (!_BINARY_RANGE(sjis_info->bad_cnt, _source_size)) {
			/* バイナリー域ではない時（SJISを兼用） */
				if (is_jis) {
					CHECKLN_SETA_AND_RETURN(dcs, ltype, tltype, atype, DCS_JIS);
				}
				if (is_forced) {
					CHECKLN_SETA_AND_RETURN(dcs, ltype, tltype, atype, stype);
				}
			}
		}
		if (ubacnt > 0 || ulacnt > 0) {
		/* UTF16 の ASCII があった時 */
			is_ascii = is_all_ascii = false;
			ubacnt_all += ubacnt;
			ulacnt_all += ulacnt;
		}
		if (is_ascii) {
		/* ASCII だった時 */
			euc_info->blen = u8_info->blen = 0;
		} else {
		/* ASCII ではなかった時 */
			/* ASCII, JIS をクリア */
			flag_btype &= ~(DCS_ND_ASCII | DCS_ND_JIS);
			/* EUC の解析 */
			sp = han_kana_p = _buffser_info.buf + _buffser_info.e_offset;
			euc_info->blen = 0;
			while(sp < ep) {
				if ((c = *sp++) <= 0x7f) {
					if (c == '\r') {
						if (sp >= ep) {
							euc_info->blen = 1;
						}
					}
					continue;
				}
				three_bytes = 0;
				if (c == 0x8e) {
					if ((sp >= ep)) {
						if (is_eof) {
							euc_info->bad_cnt += 2;
							euc1_info->bad_cnt += 2;
							euc2_info->bad_cnt += 2;
							euc3_info->bad_cnt += 2;
						}
						euc_info->blen = 1;
						break;
					} else {
						c2 = *sp++;
						if (is_sjis_kana(c2)) {
							if (han_kana_p < sp - 4 && (sp + 1 >= ep
							 || !(*sp == 0x8e && is_sjis_kana(*(sp + 1))))) {
							/* 半角カナが単一の時 */
								euc_info->tan_han_kana_cnt += 2;
							}
							han_kana_p = sp - 2;
						} else {
							euc_info->bad_cnt += 2;
						}
					}
					continue;
				} else if (c == 0x8f) {
					if (sp + 2 > ep) {
						if (is_eof) {
							euc1_info->bad_cnt += 3;
							euc3_info->bad_cnt += 3;
						}
						euc_info->blen = 2 - (ep - sp);
						break;
					}
					three_bytes = 1;
					c = ((c << 8) | *sp++);
				} else if (is_eucms_byte(c)) {
					/* EUCJP-MSコード(8bit(0x80-0x9f))の判定 */
					continue;
				} else if (sp >= ep) {
					if (is_eof) {
						euc_info->bad_cnt += 2;
						euc1_info->bad_cnt += 2;
						euc2_info->bad_cnt += 2;
						euc3_info->bad_cnt += 2;
					}
					euc_info->blen = 1;
					break;
				}
				c = ((c << 8) | *sp++);
				if (is_euc_long(dcs, c)) {
					if (is_euc_zenkana(c)) {
						kana_bits |= DCS_ND_EUC;
						euc_info->zen_kana_cnt += 2;
					}
					euc_info->zen_cnt += 2 + three_bytes;
				} else {
					euc_info->bad_cnt += 2 + three_bytes;
					set_euc20932_to(dcs);
					if (is_euc_long(dcs, c)) {
						euc1_info->zen_cnt += 2 + three_bytes;
					} else {
						euc1_info->bad_cnt += 2 + three_bytes;
					}
					if (three_bytes == 0) {
					/* 0x8fxxxx コードではない時 */
						set_euc51932_to(dcs);
						if (is_euc_long(dcs, c)) {
							euc2_info->zen_cnt += 2;
						} else {
							euc2_info->bad_cnt += 2;
						}
					} else {
						euc2_info->bad_cnt += 2 + three_bytes;
					}
					set_eucms_to(dcs);
					if (is_euc_long(dcs, c)) {
						/* 拡張 Unicode(BE)2 判定 */
						if (!is_unibeex_short(dcs, _euc_to_unibe(dcs, c))) {
							euc3_info->zen_cnt += 2 + three_bytes;
						}
					} else {
						euc3_info->bad_cnt += 2 + three_bytes;
					}
					set_euccommon_to(dcs);
				} /* if (is_euc_long(dcs, c)) */
			} /* while(sp < ep) */
			if (is_eof) {
				/* 共通コードのカウントを他のコードに加える */
				euc1_info->zen_cnt += euc_info->zen_cnt;
				euc2_info->zen_cnt += euc_info->zen_cnt;
				euc3_info->zen_cnt += euc_info->zen_cnt;
				euc1_info->zen_kana_cnt += euc_info->zen_kana_cnt;
				euc2_info->zen_kana_cnt += euc_info->zen_kana_cnt;
				euc3_info->zen_kana_cnt += euc_info->zen_kana_cnt;
				/* 共通コードが泣き別れの時は、全てが泣き別れ */
				/* 全角の数が等しい時は、共通コードを採る */
				if (stype != DCS_EUC1) {
				/* EUC1 の指示がない時 */
					if (euc1_info->zen_cnt == euc_info->zen_cnt) {
						flag_btype &= ~DCS_ND_EUC1;
					}
				}
				if (stype != DCS_EUC2) {
				/* EUC2 の指示がない時 */
					if (euc2_info->zen_cnt == euc_info->zen_cnt) {
						flag_btype &= ~DCS_ND_EUC2;
					}
				}
				if (stype != DCS_EUC3) {
				/* EUC3 の指示がない時 */
					if (euc3_info->zen_cnt == euc_info->zen_cnt) {
						flag_btype &= ~DCS_ND_EUC3;
					} else if (euc3_info->zen_cnt > euc_info->zen_cnt) {
						if ((flag_btype & DCS_ND_EUC1) != 0) {
							if (euc1_info->zen_cnt == euc3_info->zen_cnt
							 && euc1_info->bad_cnt == euc3_info->bad_cnt) {
							/* EUC1, EUC3 の区別が付かない時 */
								flag_btype &= ~DCS_ND_EUC3;
							}
						}
						if ((flag_btype & DCS_ND_EUC2) != 0) {
							if (euc2_info->zen_cnt == euc3_info->zen_cnt
							 && euc2_info->bad_cnt == euc3_info->bad_cnt) {
							/* EUC2, EUC3 の区別が付かない時 */
								flag_btype &= ~DCS_ND_EUC3;
							}
						}
					}
				}
			}
			/* UTF-8 の解析	 */
			sp = _buffser_info.buf + _buffser_info.u8_offset;
			u8_info->blen = 0;
			is_utfmac_d = false;
			ecnt = 0;
			while(sp < ep) {
				tp = sp;
				c = *sp++;
				if ((c & 0x80) == 0) {
				/* 0xxxxxxx: １バイト文字 */
					if (c == '\r') {
						if (sp >= ep) {
							euc_info->blen = 1;
						}
					}
					continue;
				}
				if (sp >= ep) {
					if (is_eof) {
						u8_info->bad_cnt += 2;
					}
					u8_info->blen = 1;
					break;
				}
				c2 = *sp++;
				if ((c & 0xe0) == 0xc0) {
					if ((c2 & 0xc0) == 0x80) {
					/* 110xxxxx 10xxxxxx: ２バイト文字 */
						ch = ((c&0x1f)<<6|(c2&0x3f));
						if (is_unibe_short(dcs, ch)) {
							u8_info->zen_cnt += 2;
							if (is_unibe_mac(dcs, ch, CSB_MAC_BO | CSB_MAC_TO)) {
							/* UTF8-MAC 短音符（ブレーヴェ）、分音符（ウムラウト）の時 */
								is_utfmac_d = true;
							} else if (is_utfmac_d) {
								if (ch == MAC_BUNONPU || ch == MAC_TANONPU) {
									u8_mac_cnt += 4;
								}
								is_utfmac_d = false;
							}
						} else {
							u8_info->bad_cnt += 2;
						}
					} else {
						u8_info->bad_cnt += 2;
					}
				} else {
					if (sp >= ep) {
						if (is_eof) {
							u8_info->bad_cnt += 3;
						}
						u8_info->blen = 2;
						break;
					}
					c3 = *sp++;
					if ((c & 0xf0) == 0xe0) {
						if ((c2 & 0xc0) == 0x80 && (c3 & 0xc0) == 0x80) {
						/* 1110xxxx 10xxxxxx 10xxxxxx: ３バイト文字 */
							ch = ((c&0x0f)<<12|(c2&0x3f)<<6|(c3&0x3f));
							if (is_unibe_short(dcs, ch)) {
								u8_info->zen_cnt += 3;
								if (is_unibe_zenkana(ch)) {
								/* 全角仮名の時 */
									kana_bits |= DCS_ND_UTF8;
									u8_info->zen_kana_cnt += 3;
									if (is_unibe_mac(dcs, ch, CSB_MAC_D | CSB_MAC_H)) {
									/* UTF8-MAC 濁点、半濁点の時 */
										is_utfmac_d = true;
									} else if (is_utfmac_d) {
										if (ch == MAC_DAKUTEN || ch == MAC_HANDAKUTEN) {
											u8_mac_cnt += 6;
										}
										is_utfmac_d = false;
									}
								}
							} else {
								u8_info->bad_cnt += 3;
							}
						} else {
							u8_info->bad_cnt += 3;
						}
					} else {
						if (sp >= ep) {
							if (is_eof) {
								u8_info->bad_cnt += 4;
							}
							u8_info->blen = 3;
							break;
						}
						c4 = *sp++;
						if ((c & 0xf8) == 0xf0 && (c2 & 0xc0) == 0x80 && (c3 & 0xc0) == 0x80 && (c4 & 0xc0) == 0x80) {
						/* 11110xxx 10xxxxxx 10xxxxxx 10xxxxxx: ４バイト文字 */
							/* サロゲート文字 */
							u8_info->zen_cnt += 4;
							u8_surrog_cnt += 4;
						} else {
							ecnt++;
							sp = tp + 1;
							continue;
						}
					}
				}
			} /* while(...) */

			u8_info->bad_cnt += ecnt;
			if (is_eof) {
				if (u8_info->zen_cnt == 0) {
					/* ASCII 以外の文字が無かった時 */
					flag_btype &= ~DCS_ND_UTF8;
				}
			}
		} /* if (is_ascii) */

		/* Unicode の解析	 */
		sp = _buffser_info.buf + _buffser_info.u16_offset;
		ub_info->blen = 0; /* ul_info 共通 */
		while(sp < ep) {
			c = *sp++;
			if (sp >= ep) {
				if (is_eof) {
					ub_info->bad_cnt += 2;
					ul_info->bad_cnt += 2;
				}
				ub_info->blen = 1;
				break;
			}
			c2 = *sp++;
			/* Big Endian */
			ch = ((c << 8) | c2);

			if (is_unibestd_short(dcs, ch)) {
				if (is_unibe_zenkana(ch)) {
					kana_bits |= DCS_ND_UTF16BE;
					ub_info->zen_kana_cnt += 2;
				}
				ub_info->zen_cnt += 2;
				if (is_unibe_mac(dcs, ch, CSB_MAC_D | CSB_MAC_H | CSB_MAC_BO | CSB_MAC_TO)) {
				/* UTF8-MAC 濁点、半濁点の時 */
					if (sp + 2 <= ep) {
						c3 = (((*sp&0xff)<<8)|(*(sp+1)&0xff));
						if (c3 == MAC_DAKUTEN || c3 == MAC_HANDAKUTEN || c3 == MAC_BUNONPU || c3 == MAC_TANONPU) {
							ub_info->zen_cnt += 2;
							ub_mac_cnt += 4;
							sp += 2;
						}
					}
				}
			} else if (is_unibe_surrogate_high(ch)) {
				if (sp + 2 <= ep) {
					c3 = (((*sp&0xff)<<8)|(*(sp+1)&0xff));
					if (is_unibe_surrogate_low(c3)) {
						ub_info->zen_cnt += 4;
						ub_surrog_cnt += 4;
						sp += 2;
					}
				}
			} else if (ch == REP_CHARBE) {
			/* 置換文字の時 */
				ub_info->zen_cnt += 2;
			} else {
				ub_info->bad_cnt += 2;
			}
			/* Little Endian */
			ch = ((c2 << 8) | c);
			if (is_unibestd_short(dcs, ch)) {
				if (is_unibe_zenkana(ch)) {
					kana_bits |= DCS_ND_UTF16LE;
					ul_info->zen_kana_cnt += 2;
				}
				ul_info->zen_cnt += 2;
				if (is_unibe_mac(dcs, ch, CSB_MAC_D | CSB_MAC_H | CSB_MAC_BO | CSB_MAC_TO)) {
				/* UTF8-MAC 濁点、半濁点の時 */
					if (sp + 2 <= ep) {
						c3 = (((*(sp+1)&0xff)<<8)|(*sp&0xff));
						if (c3 == MAC_DAKUTEN || c3 == MAC_HANDAKUTEN || c3 == MAC_BUNONPU || c3 == MAC_TANONPU) {
							ul_info->zen_cnt += 2;
							ul_mac_cnt += 4;
							sp += 2;
						}
					}
				}
			} else if (is_unibe_surrogate_high(ch)) {
				if (sp + 2 <= ep) {
					c3 = (((*(sp+1)&0xff)<<8)|(*sp&0xff));
					if (is_unibe_surrogate_low(c3)) {
						ul_info->zen_cnt += 4;
						ul_surrog_cnt += 4;
						sp += 2;
					}
				}
			} else if (ch == REP_CHARLE) {
			/* 置換文字の時 */
				ul_info->zen_cnt += 2;
			} else {
				ul_info->bad_cnt += 2;
			}
		} /* while(sp < ep;) */
		if (ub_surrog_cnt > 0 || ul_surrog_cnt > 0 || u8_surrog_cnt > 0) {
			dcs->_is_surrogate = true;
		}
		if (ub_mac_cnt > 0 || ul_mac_cnt > 0 || u8_mac_cnt > 0) {
			dcs->_utfmac_exists = true;
		}
		if (is_eof) {
			if (ubacnt_all > 0 || ulacnt_all > 0) {
			/* UTF16 の ASCII がある時 */
			} else if (is_all_ascii) {
			/* すべて ASCII だった時 */
				if ((ub_info->bad_cnt > 0
				 && ul_info->bad_cnt > 0)
				 || (ub_info->zen_cnt == 0 && ul_info->zen_cnt == 0)) {
				/* UTF16 の全角が無かった時 */
					FILL_BUFFER_AND_RETURN(dcs, ltype, tltype, atype, DCS_ASCII, is_keep);
				}
				if (stype != DCS_UTF16BE && ub_info->bad_cnt > 0) {
				/* UTF16BE の指示がない時
				 * かつ、不正コードがある時 */
					flag_btype &= ~DCS_ND_UTF16BE;
				}
				if (stype != DCS_UTF16LE && ul_info->bad_cnt > 0) {
					/* UTF16LE の指示がない時
					 * かつ、不正コードがある時 */
					flag_btype &= ~DCS_ND_UTF16LE;
				}
			}
			if (ub_info->bad_cnt > ul_info->bad_cnt) {
			/* UTF16BE の不正文字が多い時 */
				/* UTF16LE の全角を優先する */
				ub_info->zen_cnt -= uba_ulz_cnt;
				ub_info->zen_cnt -= ul_surrog_cnt;
			} else if (ul_info->bad_cnt > ub_info->bad_cnt) {
			/* UTF16LE の不正文字が多い時 */
				/* UTF16BE の全角を優先する */
				ul_info->zen_cnt -= ula_ubz_cnt;
				ul_info->zen_cnt -= ub_surrog_cnt;
			} else if (ub_surrog_cnt > ul_surrog_cnt) {
			/* UTF16BE のサロゲート文字が多い時 */
				ub_info->zen_cnt += ub_surrog_cnt;
			} else if (ul_surrog_cnt > ub_surrog_cnt) {
			/* UTF16LE のサロゲート文字が多い時 */
				ul_info->zen_cnt += ul_surrog_cnt;
			}
		}
		if (flag_btype == 0) {
		/* 該当なしの時 */
			FILL_BUFFER_AND_RETURN(dcs, ltype, tltype, atype, DCS_UNKNOWN, is_keep);
		}
		/* 不特定コード・タイプが一意の場合、コード・タイプを決定する */
		switch(flag_btype) {
		case DCS_ND_ASCII:
			FILL_BUFFER_AND_RETURN(dcs, ltype, tltype, atype, DCS_ASCII, is_keep);
		case DCS_ND_SJIS:
			FILL_BUFFER_AND_RETURN(dcs, ltype, tltype, atype, DCS_SJIS, is_keep);
		case DCS_ND_EUC:
			FILL_BUFFER_AND_RETURN(dcs, ltype, tltype, atype, DCS_EUC, is_keep);
		case DCS_ND_EUC1:
			FILL_BUFFER_AND_RETURN(dcs, ltype, tltype, atype, DCS_EUC1, is_keep);
		case DCS_ND_EUC2:
			FILL_BUFFER_AND_RETURN(dcs, ltype, tltype, atype, DCS_EUC2, is_keep);
		case DCS_ND_EUC3:
			FILL_BUFFER_AND_RETURN(dcs, ltype, tltype, atype, DCS_EUC3, is_keep);
		case DCS_ND_UTF16BE:
			FILL_BUFFER_AND_RETURN(dcs, ltype, 0, atype, DCS_UTF16BE, is_keep);
		case DCS_ND_UTF16LE:
			FILL_BUFFER_AND_RETURN(dcs, ltype, 0, atype, DCS_UTF16LE, is_keep);
		case DCS_ND_UTF8:
			FILL_BUFFER_AND_RETURN(dcs, ltype, tltype, atype, DCS_UTF8, is_keep);
		default:
			break;
		}
		if (!is_eof) {
		/* EOF ではない時 */
			/* 最大値を取得 */
			maxb = 0;
			for(i=0; i<sizeof(cinfo_sp)/sizeof(cinfo_sp[0]); i++) {
				if ((flag_btype & cinfo_sp[i]->nd_type) != 0) {
				/* 処理対象の時 */
					maxb = cinfo_sp[i]->blen;
					for(i++; i<sizeof(cinfo_sp)/sizeof(cinfo_sp[0]); i++) {
						if ((flag_btype & cinfo_sp[i]->nd_type) != 0) {
						/* 処理対象の時 */
							if (maxb < cinfo_sp[i]->blen) {
								maxb = cinfo_sp[i]->blen;
							}
						}
					}
				}
			}
			_buffser_info.offset = _MAX_OFFSET;
			_buffser_info.s_offset = _MAX_OFFSET - sjis_info->blen;
			_buffser_info.e_offset = _MAX_OFFSET - euc_info->blen;
			_buffser_info.u8_offset  = _MAX_OFFSET - u8_info->blen;
			_buffser_info.u16_offset = _MAX_OFFSET - ub_info->blen;
			/* 残りを先頭へコピー */
			memcpy(_buffser_info.buf+ _MAX_OFFSET - maxb, ep - maxb, maxb);
			sp = _buffser_info.buf + _MAX_OFFSET;
			/* 続きを読み込む */
			continue;
		}
		/* これより、全て表示指定と返却値が別れる */
		if (atype != NULL) {
		/* 全て表示指定ありの時 */
			*atype = flag_btype;
		}
		for(i=0; i<sizeof(code_info)/sizeof(code_info[0]); i++) {
			if ((flag_btype & code_info[i].nd_type) == 0) {
				/* 無効データの件数を無効とする */
				code_info[i].zen_cnt = _MIN_VAL;
				code_info[i].bad_cnt = _MAX_VAL;
			} else {
				/* ソート第２キー：全角＋全角かなー１文字半角カナ */
				code_info[i].key = code_info[i].zen_cnt
						+ code_info[i].zen_kana_cnt
						- code_info[i].tan_han_kana_cnt;
			}
		}
		qsort(cinfo, sizeof(cinfo)/sizeof(cinfo[0]), sizeof(cinfo[0]), _tcompare);
		for(i=0; i<sizeof(cinfo)/sizeof(cinfo[0]); i++) {
			if (cinfo[i]->bad_cnt > 0) {
				/* 不正文字がある時 */
				/* 又は泣き別れの時 */
				break;
			}
			if (cinfo[i]->type == stype) {
				/* 指示にマッチしたタイプの時 */
				CHECKLN_SETA_AND_RETURN(dcs, ltype, judge_unicode(cinfo[i], tltype), atype, cinfo[i]->type);
			}
		}
		if (_BINARY_RANGE(cinfo[0]->bad_cnt, _source_size)) {
		/* バイナリー域の時 */
			SETA_AND_RETURN(dcs, atype, DCS_UNKNOWN);
		}
		if (cinfo[0]->bad_cnt < cinfo[1]->bad_cnt) {
		/* 不正文字が少ない時 */
			CHECKLN_SETA_AND_RETURN(dcs, ltype, judge_unicode(cinfo[0], tltype), atype, cinfo[0]->type);
		}
		if (ub_info->bad_cnt > 0) {
			flag_btype &= ~DCS_ND_UTF16BE;
		}
		if (ul_info->bad_cnt > 0) {
			flag_btype &= ~DCS_ND_UTF16LE;
		}
#if ! _KEEP_ASCII
		if (cinfo[0]->zen_cnt > 0 && cinfo[0]->bad_cnt == 0) {
		/* 全角がある時、かつ不正文字がない時 */
			flag_btype &= ~DCS_ND_ASCII;
		}
#endif
		if (cinfo[0]->key > cinfo[1]->key) {
		/* 全角、全角仮名が多い時 */
			if ((cinfo[0]->type == DCS_UTF16BE || cinfo[0]->type == DCS_UTF16LE)
			 && (flag_btype & DCS_ND_ASCII) != 0
			 && (flag_btype & (DCS_ND_UTF16BE | DCS_ND_UTF16LE)) != 0) {
			/* ASCII と Unicode の時 */
				CHECKLN_SETA_AND_RETURN(dcs, ltype, judge_unicode(cinfo[0], tltype), atype, (flag_btype & (DCS_ND_ASCII | DCS_ND_UTF16BE | DCS_ND_UTF16LE)));
			}
			CHECKLN_SETA_AND_RETURN(dcs, ltype, judge_unicode(cinfo[0], tltype), atype, cinfo[0]->type);
		}
		if (cinfo[0]->zen_kana_cnt > cinfo[1]->zen_kana_cnt) {
		/* 全角仮名が多い時 */
			CHECKLN_SETA_AND_RETURN(dcs, ltype, judge_unicode(cinfo[0], tltype), atype, cinfo[0]->type);
		}
		if (cinfo[0]->tan_han_kana_cnt == 0 && cinfo[1]->tan_han_kana_cnt > 0) {
		/* 単一半角カナが無く、かつ後続に単一半角カナがある時 */
			SETA_AND_RETURN(dcs, atype, cinfo[0]->nd_type);
		}
#if _KEEP_ASCII
		/* ASCII を保つ */
		dtype = (flag_btype & DCS_ND_ASCII);
#else
		dtype = 0;
#endif
		for(i=0;; i++) {
			if (cinfo[i]->type == stype) {
			/* 指示にマッチしたタイプの時 */
				CHECKLN_SETA_AND_RETURN(dcs, ltype, judge_unicode(cinfo[i], tltype), atype, cinfo[i]->type);
			}
			switch(cinfo[i]->type) {
			case DCS_JIS:
				dtype |= DCS_ND_JIS;
				break;
			case DCS_SJIS:
				dtype |= DCS_ND_SJIS;
				break;
			case DCS_EUC:
				dtype |= DCS_ND_EUC;
				break;
			case DCS_EUC1:
				dtype |= DCS_ND_EUC1;
				break;
			case DCS_EUC2:
				dtype |= DCS_ND_EUC2;
				break;
			case DCS_EUC3:
				dtype |= DCS_ND_EUC3;
				break;
			case DCS_UTF16BE:
				dtype |= DCS_ND_UTF16BE;
				break;
			case DCS_UTF16LE:
				dtype |= DCS_ND_UTF16LE;
				break;
			case DCS_UTF8:
				dtype |= DCS_ND_UTF8;
				break;
			}
			if (i + 1 >= sizeof(cinfo) / sizeof(cinfo[0])) {
			/* 最後のデータの時 */
				break;
			}
			if (cinfo[i]->zen_cnt > cinfo[i+1]->zen_cnt) {
			/* 全角が多い時 */
				break;
			}
			if (cinfo[i]->bad_cnt < cinfo[i+1]->bad_cnt) {
			/* 不正文字が少ない時 */
				break;
			}
		}
		SETA_AND_RETURN(dcs, atype, dtype);
	} /* for(;;) */
} /* _dcs_judges */

#define _LN_CHECK_AND_RERURN(ltype) { \
	if ((ltype) == LTYPE_ALL) { \
		return LTYPE_ALL; \
	} }

int _check_ln(Dcs *dcs, int ltype, int stype) {
	int c, c2;
	const unsigned char *sp, *ep;
	ll_list *tlist, *clist = ll_first((ll_list *)dcs->_block);
	BufferInfo *tbinfo, *binfo = ll_get(clist);

	sp = binfo->buf + binfo->offset;;
	ep = binfo->buf + binfo->length;

	if (stype == DCS_UTF16BE) {
		for(;;) {
			/* 残りが最大長に満たない時、次の入力バッファに移動 */
			check_ll_buf(dcs, sp, ep, 4, clist, tlist, binfo, tbinfo);
			if (sp + 1 >= ep) {
				return ltype;
			}
			c = *sp++;
			c2 = *sp++;
			if (c == 0) {
				if (c2 == '\n') {
					_LN_CHECK_AND_RERURN(ltype |= LTYPE_LF);
				} else if (c2 == '\r') {
					if (sp + 1 >= ep) {
						ltype |= LTYPE_CR;
						return ltype;
					}
					c = *sp++;
					if (c == 0) {
						c2 = *sp++;
						if (c2 == '\n') {
							_LN_CHECK_AND_RERURN(ltype |= LTYPE_CRLF);
						} else {
							_LN_CHECK_AND_RERURN(ltype |= LTYPE_CR);
						}
					} else {
						sp--;
						_LN_CHECK_AND_RERURN(ltype |= LTYPE_CR);
					}
				}
			}
		}
	} else if (stype == DCS_UTF16LE) {
		for(;;) {
			/* 残りが最大長に満たない時、次の入力バッファに移動 */
			check_ll_buf(dcs, sp, ep, 4, clist, tlist, binfo, tbinfo);
			if (sp + 1 >= ep) {
				return ltype;
			}
			c = *sp++;
			c2 = *sp++;
			if (c2 == 0) {
				if (c == '\n') {
					_LN_CHECK_AND_RERURN(ltype |= LTYPE_LF);
				} else if (c == '\r') {
					if (sp + 1 >= ep) {
						ltype |= LTYPE_CR;
						return ltype;
					}
					c = *sp++;
					if (c == '\n') {
						c2 = *sp++;
						if (c2 == 0) {
							_LN_CHECK_AND_RERURN(ltype |= LTYPE_CRLF);
						} else {
							_LN_CHECK_AND_RERURN(ltype |= LTYPE_CR);
						}
					} else {
						sp--;
						_LN_CHECK_AND_RERURN(ltype |= LTYPE_CR);
					}
				}
			}
		}
	} else {
		if (!DCS_CONFIRMED_CODE_TYPE(stype)) {
		/* 未確定コードがあった時 */
			return DCS_ECTYPE;
		}
		for(;;) {
			/* 残りが最大長に満たない時、次の入力バッファに移動 */
			check_ll_buf(dcs, sp, ep, 2, clist, tlist, binfo, tbinfo);
			if (sp >= ep) {
				return ltype;
			}
			c = *sp++;
			if (c == '\n') {
				_LN_CHECK_AND_RERURN(ltype |= LTYPE_LF);
			} else if (c == '\r') {
				if (sp >= ep) {
					return ltype |= LTYPE_CR;
				}
				c = *sp++;
				if (c == '\n') {
					_LN_CHECK_AND_RERURN(ltype |= LTYPE_CRLF);
				} else {
					sp--;
					_LN_CHECK_AND_RERURN(ltype |= LTYPE_CR);
				}
			}
		}
	}
} /* _check_ln */

/* 共通コード変換 */
int _dcs_convert(Dcs *dcs, int stype, int dtype, bool is_forced) {
	int sts = DCS_SUCCESS;
	int cmp;
	int estype = _dcs_judges(dcs, NULL, NULL, stype, is_forced, true);

	if (dcs->_observer != NULL
	 && dcs->_observer(estype, dcs->_user_info) != DCS_OBS_CONTINUE) {
		return estype;
	}
	if (!DCS_CONFIRMED_CODE_TYPE(estype)) {
	/* 入力タイプが確定できなかった時 */
		return estype;
	}
	if (estype == DCS_ASCII) {
		estype = DCS_SJIS;
	}
	if (dtype == DCS_ASCII) {
		dtype = DCS_SJIS;
	}
	if (estype == dtype) {
	/* 入出力のコードが等しい時 */
		if (dcs_is_utfmac_to(dcs) && dcs_utfmac_exists(dcs)) {
			switch(dtype) {
			case DCS_UTF16BE:
			case DCS_UTF16BE_BOM:
				sts = _unibeto_unibe(dcs, dtype==DCS_UTF16BE_BOM);
				break;
			case DCS_UTF16LE:
			case DCS_UTF16LE_BOM:
				sts = _unileto_unile(dcs, dtype==DCS_UTF16LE_BOM);
				break;
			case DCS_UTF8:
			case DCS_UTF8_BOM:
				sts = _utf8to_utf8(dcs, dtype==DCS_UTF8_BOM);
				break;
			}
		} else {
			sts = _dcs_copy(dcs, estype);
		}
	} else if ((cmp = dcs_cmp_bom(estype, dtype)) == CBOM_ADD) {
	/* 同一 UTF コードで BOM 追加の時 */
		if (dcs_is_utfmac_to(dcs) && dcs_utfmac_exists(dcs)) {
			switch(dtype) {
			case DCS_UTF16BE_BOM:
				sts = _unibeto_unibe(dcs, true);
				break;
			case DCS_UTF16LE_BOM:
				sts = _unileto_unile(dcs, true);
				break;
			case DCS_UTF8_BOM:
				sts = _utf8to_utf8(dcs, true);
				break;
			}
		} else {
			sts = _add_bom(dcs, dtype);
		}
	} else if (cmp == CBOM_DELETE) {
	/* 同一 UTF コードで BOM 削除の時 */
		if (dcs_is_utfmac_to(dcs) && dcs_utfmac_exists(dcs)) {
			switch(dtype) {
			case DCS_UTF16BE:
				sts = _unibeto_unibe(dcs, false);
				break;
			case DCS_UTF16LE:
				sts = _unileto_unile(dcs, false);
				break;
			case DCS_UTF8:
				sts = _utf8to_utf8(dcs, false);
				break;
			}
		} else {
			sts = _delete_bom(dcs, dtype);
		}
	} else {
		switch(estype) {
		case DCS_JIS:
			set_jis50221_to(dcs);
			break;
		case DCS_JIS2:
			set_jis2022jp200401_to(dcs);
			break;
		case DCS_EUC:
			set_euccommon_to(dcs);
			break;
		case DCS_EUC1:
			set_euc20932_to(dcs);
			break;
		case DCS_EUC2:
			set_euc51932_to(dcs);
			break;
		case DCS_EUC3:
			set_eucms_to(dcs);
			break;
		}
		switch(dtype) {
		case DCS_JIS:
			if (dcs->_is_surrogate) {
				set_to_jis2022jp2004(dcs);
			} else {
				set_to_jis50221(dcs);
			}
			break;
		case DCS_JIS2:
			set_to_jis2022jp2004(dcs);
			break;
		case DCS_EUC:
			set_to_euccommon(dcs);
			break;
		case DCS_EUC1:
			set_to_euc20932(dcs);
			break;
		case DCS_EUC2:
			set_to_euc51932(dcs);
			break;
		case DCS_EUC3:
			set_to_eucms(dcs);
			break;
		}
		switch(estype) {
		case DCS_EMPTY:
			break;
		case DCS_JIS:
		case DCS_JIS2:
			switch(dtype) {
			case DCS_JIS:
			case DCS_JIS2:
				sts = _jisto_jis(dcs);
				break;
			case DCS_ASCII:
				/* no break: SJIS とみなす */
			case DCS_SJIS:
				sts = _jisto_sjis(dcs);
				break;
			case DCS_EUC:
			case DCS_EUC1:
			case DCS_EUC2:
			case DCS_EUC3:
				sts = _jisto_euc(dcs);
				break;
			case DCS_UTF16BE:
			case DCS_UTF16BE_BOM:
				sts = _jisto_unibe(dcs, dtype==DCS_UTF16BE_BOM);
				break;
			case DCS_UTF16LE:
			case DCS_UTF16LE_BOM:
				sts = _jisto_unile(dcs, dtype==DCS_UTF16LE_BOM);
				break;
			case DCS_UTF8:
			case DCS_UTF8_BOM:
				sts = _jisto_utf8(dcs, dtype==DCS_UTF8_BOM);
				break;
			default:
				break;
			}
			break;
		case DCS_ASCII:
			/* no break: SJIS とみなす */
		case DCS_SJIS:
			switch(dtype) {
			case DCS_JIS:
			case DCS_JIS2:
				sts = _sjisto_jis(dcs);
				break;
			case DCS_EUC:
			case DCS_EUC1:
			case DCS_EUC2:
			case DCS_EUC3:
				sts = _sjisto_euc(dcs);
				break;
			case DCS_UTF16BE:
			case DCS_UTF16BE_BOM:
				sts = _sjisto_unibe(dcs, dtype==DCS_UTF16BE_BOM);
				break;
			case DCS_UTF16LE:
			case DCS_UTF16LE_BOM:
				sts = _sjisto_unile(dcs, dtype==DCS_UTF16LE_BOM);
				break;
			case DCS_UTF8:
			case DCS_UTF8_BOM:
				sts = _sjisto_utf8(dcs, dtype==DCS_UTF8_BOM);
				break;
			default:
				break;
			}
			break;
		case DCS_EUC:
		case DCS_EUC1:
		case DCS_EUC2:
		case DCS_EUC3:
			switch(dtype) {
			case DCS_JIS:
			case DCS_JIS2:
				sts = _eucto_jis(dcs);
				break;
			case DCS_ASCII:
				/* no break: SJIS とみなす */
			case DCS_SJIS:
				sts = _eucto_sjis(dcs);
				break;
			case DCS_EUC:
			case DCS_EUC1:
			case DCS_EUC2:
			case DCS_EUC3:
				sts = _eucto_euc(dcs);
				break;
			case DCS_UTF16BE:
			case DCS_UTF16BE_BOM:
				sts = _eucto_unibe(dcs, dtype==DCS_UTF16BE_BOM);
				break;
			case DCS_UTF16LE:
			case DCS_UTF16LE_BOM:
				sts = _eucto_unile(dcs, dtype==DCS_UTF16LE_BOM);
				break;
			case DCS_UTF8:
			case DCS_UTF8_BOM:
				sts = _eucto_utf8(dcs, dtype==DCS_UTF8_BOM);
				break;
			default:
				break;
			}
			break;
		case DCS_UTF16BE:
		case DCS_UTF16BE_BOM:
			switch(dtype) {
			case DCS_JIS:
			case DCS_JIS2:
				sts = _unibeto_jis(dcs);
				break;
			case DCS_ASCII:
				/* no break: SJIS とみなす */
			case DCS_SJIS:
				sts = _unibeto_sjis(dcs);
				break;
			case DCS_EUC:
			case DCS_EUC1:
			case DCS_EUC2:
			case DCS_EUC3:
				sts = _unibeto_euc(dcs);
				break;
			case DCS_UTF16LE:
			case DCS_UTF16LE_BOM:
				sts = _unibeto_unile(dcs, dtype==DCS_UTF16LE_BOM);
				break;
			case DCS_UTF8:
			case DCS_UTF8_BOM:
				sts = _unibeto_utf8(dcs, dtype==DCS_UTF8_BOM);
				break;
			default:
				break;
			}
			break;
		case DCS_UTF16LE:
		case DCS_UTF16LE_BOM:
			switch(dtype) {
			case DCS_JIS:
			case DCS_JIS2:
				sts = _unileto_jis(dcs);
				break;
			case DCS_ASCII:
				/* no break: SJIS とみなす */
			case DCS_SJIS:
				sts = _unileto_sjis(dcs);
				break;
			case DCS_EUC:
			case DCS_EUC1:
			case DCS_EUC2:
			case DCS_EUC3:
				sts = _unileto_euc(dcs);
				break;
			case DCS_UTF16BE:
			case DCS_UTF16BE_BOM:
				sts = _unileto_unibe(dcs, dtype==DCS_UTF16BE_BOM);
				break;
			case DCS_UTF8:
			case DCS_UTF8_BOM:
				sts = _unileto_utf8(dcs, dtype==DCS_UTF8_BOM);
				break;
			default:
				break;
			}
			break;
		case DCS_UTF8:
		case DCS_UTF8_BOM:
			switch(dtype) {
			case DCS_JIS:
			case DCS_JIS2:
				sts = _utf8to_jis(dcs);
				break;
			case DCS_ASCII:
				/* no break: SJIS とみなす */
			case DCS_SJIS:
				sts = _utf8to_sjis(dcs);
				break;
			case DCS_EUC:
			case DCS_EUC1:
			case DCS_EUC2:
			case DCS_EUC3:
				sts = _utf8to_euc(dcs);
				break;
			case DCS_UTF16BE:
			case DCS_UTF16BE_BOM:
				sts = _utf8to_unibe(dcs, dtype==DCS_UTF16BE_BOM);
				break;
			case DCS_UTF16LE:
			case DCS_UTF16LE_BOM:
				sts = _utf8to_unile(dcs, dtype==DCS_UTF16LE_BOM);
				break;
			default:
				break;
			}
			break;
		}
	}
	return sts;
} /* _dcs_convert */

/* バッファー確保 */
int _fill_buffer(Dcs *dcs, int stype) {
	int c, c2, c3, c4;
	unsigned short ch;
	uint64_t lch;
	unsigned char *sp, *ep;
	bool is_eof = false;
	bool is_utfmac_d;
	ll_list *tlist, *clist = dcs->_block;
	JniJudgedStreamInfo * jni_info = (JniJudgedStreamInfo *)dcs->_jni_info;
	size_t rlen, alen = (_is_type_jni_stream(dcs)?
			(*jni_info->jenv)->GetArrayLength(jni_info->jenv, jni_info->jbuf): 0);

	for(;;) {
		memset(&_buffser_info, 0, sizeof(_buffser_info));
		_buffser_info.offset = _MAX_OFFSET;
		sp = _buffser_info.buf + _MAX_OFFSET;
		ep = _buffser_info.buf + sizeof(_buffser_info.buf) - _MAX_OFFSET;
		get_one_buffer(dcs, sp, ep, alen, rlen);

		if (sp > _buffser_info.buf + _MAX_OFFSET) {
			ep = sp;
			_buffser_info.length = ep - _buffser_info.buf;
			if ((clist = _add_buffer(clist, &_buffser_info, sizeof(_buffser_info))) == NULL) {
				ll_clear((ll_list *)dcs->_block);
				return DCS_EALLOC;
			}
		}
		if (is_eof) {
			if ((!dcs->_is_surrogate || !dcs->_utfmac_exists) && DCS_UTF16BE <= stype && stype <= DCS_UTF8_BOM) {
			/* サロゲートの判定が必要な時 */
				BufferInfo *binfo, *tbinfo;

				clist = ll_first((ll_list *)dcs->_block);
				binfo = ll_get(clist);
				sp = binfo->buf + binfo->offset;;
				ep = binfo->buf + binfo->length;
				if (stype == DCS_UTF16BE || stype == DCS_UTF16BE_BOM) {
				/* UTF16BE の時 */
					for(;;) {
						/* 残りが最大長に満たない時、次の入力バッファに移動 */
						check_ll_buf(dcs, sp, ep, 4, clist, tlist, binfo, tbinfo);
						if (sp >= ep) {
							break;
						}
						ch = unibe_char_to_short(sp);
						if (is_unibe_surrogate_high(ch)) {
							dcs->_is_surrogate = true;
							if (dcs->_utfmac_exists) {
								goto loop_end;
							}
						} else {
							if (is_unibe_mac(dcs, ch, CSB_MAC_D | CSB_MAC_H | CSB_MAC_BO | CSB_MAC_TO)) {
							/* UTF8-MAC 濁点、半濁点の時 */
								if (sp + 4 <= ep) {
									c3 = (((*(sp+2)&0xff)<<8)|(*(sp+3)&0xff));
									if (c3 == MAC_DAKUTEN || c3 == MAC_HANDAKUTEN || c3 == MAC_BUNONPU || c3 == MAC_TANONPU) {
										dcs->_utfmac_exists = true;
										if (dcs->_is_surrogate) {
											goto loop_end;
										}
										sp += 2;
									}
								}
							}
						}
						sp += 2;
					}
				} else if (stype == DCS_UTF16LE || stype == DCS_UTF16LE_BOM) {
				/* UTF16LE の時 */
					for(;;) {
						/* 残りが最大長に満たない時、次の入力バッファに移動 */
						check_ll_buf(dcs, sp, ep, 4, clist, tlist, binfo, tbinfo);
						if (sp >= ep) {
							break;
						}
						ch = unile_to_unibe_char(sp);
						if (is_unibe_surrogate_high(ch)) {
							dcs->_is_surrogate = true;
							if (dcs->_utfmac_exists) {
								goto loop_end;
							}
						} else {
							if (is_unibe_mac(dcs, ch, CSB_MAC_D | CSB_MAC_H | CSB_MAC_BO | CSB_MAC_TO)) {
							/* UTF8-MAC 濁点、半濁点の時 */
								if (sp + 4 <= ep) {
									c3 = (((*(sp+3)&0xff)<<8)|(*(sp+2)&0xff));
									if (c3 == MAC_DAKUTEN || c3 == MAC_HANDAKUTEN || c3 == MAC_BUNONPU || c3 == MAC_TANONPU) {
										dcs->_utfmac_exists = true;
										if (dcs->_is_surrogate) {
											goto loop_end;
										}
										sp += 2;
									}
								}
							}
						}
						sp += 2;
					}
				} else {
				/* UTF8 の時 */
					is_utfmac_d = false;
					for(;;) {
						/* 残りが最大長に満たない時、次の入力バッファに移動 */
						check_ll_buf(dcs, sp, ep, 4, clist, tlist, binfo, tbinfo);
						if (sp >= ep) {
							break;
						}
						c = *sp++;
						if ((c & 0x80) == 0 || c == _UYEN) {
						/* 0xxxxxxx: １バイト文字 */
						} else if (sp >= ep) {
							return DCS_EUEXPEOD;
						} else {
							c2 = *sp++;
							if ((c & 0xe0) == 0xc0 && (c2 & 0xc0) == 0x80) {
							/* 110xxxxx 10xxxxxx: ２バイト文字 */
								ch = ((c&0x1f)<<6|(c2&0x3f));
								if (is_unibe_short(dcs, ch)) {
									if (is_unibe_mac(dcs, ch, CSB_MAC_BO | CSB_MAC_TO)) {
									/* UTF8-MAC の時 */
										is_utfmac_d = true;
									} else if (is_utfmac_d) {
										if (ch == MAC_DAKUTEN || ch == MAC_TANONPU) {
											dcs->_utfmac_exists = true;
											if (dcs->_is_surrogate) {
												goto loop_end;
											}
										}
										is_utfmac_d = false;
									}
								}
							} else if (sp >= ep) {
								return DCS_EUEXPEOD;
							} else {
								c3 = *sp++;
								if ((c & 0xf0) == 0xe0 && (c2 & 0xc0) == 0x80 && (c3 & 0xc0) == 0x80) {
								/* 1110xxxx 10xxxxxx 10xxxxxx: ３バイト文字 */
									ch = ((c&0x0f)<<12|(c2&0x3f)<<6|(c3&0x3f));
									if (is_unibe_short(dcs, ch)) {
										if (is_unibe_zenkana(ch)) {
										/* 全角仮名の時 */
											if (is_unibe_mac(dcs, ch, CSB_MAC_D | CSB_MAC_H)) {
											/* UTF8-MAC の時 */
												is_utfmac_d = true;
											} else if (is_utfmac_d) {
												if (ch == MAC_DAKUTEN || ch == MAC_HANDAKUTEN) {
													dcs->_utfmac_exists = true;
													if (dcs->_is_surrogate) {
														goto loop_end;
													}
												}
												is_utfmac_d = false;
											}
										}
									}
								} else if (sp >= ep) {
									return DCS_EUEXPEOD;
								} else {
									c4 = *sp++;
									/* 11110xxx 10xxxxxx 10xxxxxx 10xxxxxx: ４バイト文字 */
									if (((c) & 0xf8) == 0xf0 && ((c2) & 0xc0) == 0x80 && ((c3) & 0xc0) == 0x80 && ((c4) & 0xc0) == 0x80) {
										lch = (((c & 07) << 18)|((c2 & 0x3f) << 12)|((c3 & 0x3f) << 6)|(c4 & 0x3f));
										lch -= 0x10000;
										ch = ((lch >> 10)|0xd800);
										if (is_unibe_surrogate_high(ch)) {
											dcs->_is_surrogate = true;
											if (dcs->_utfmac_exists) {
												goto loop_end;
											}
										}
									}
								}
							}
						}
					}
				}
			}
loop_end:
			return DCS_SUCCESS;
		}
	}
} /* _fill_buffer */

int _tcompare(const void *info1, const void *info2) {
	CodeInfo *tinfo1 = *(CodeInfo **)info1, *tinfo2 = *(CodeInfo **)info2;
	/* 比較キー：不正文字数、全角仮名数、全角数、単一半角カナ数 */
	return(
			tinfo1->bad_cnt != tinfo2->bad_cnt? tinfo1->bad_cnt - tinfo2->bad_cnt:
			tinfo1->key != tinfo2->key? tinfo2->key - tinfo1->key: 0);
}
