package rabbit.dcs;

import android.Manifest;
import android.app.AlertDialog;
import android.content.Context;
import android.content.DialogInterface;
import android.content.Intent;
import android.content.pm.ActivityInfo;
import android.content.pm.PackageManager;
import android.graphics.Bitmap;
import android.graphics.Color;
import android.graphics.Point;
import android.graphics.Typeface;
import android.graphics.drawable.BitmapDrawable;
import android.os.AsyncTask;
import android.os.Build;
import android.os.Bundle;
import android.text.Spannable;
import android.text.SpannableStringBuilder;
import android.text.TextPaint;
import android.text.style.ForegroundColorSpan;
import android.text.style.StyleSpan;
import android.view.Display;
import android.view.GestureDetector;
import android.view.Menu;
import android.view.MenuInflater;
import android.view.MenuItem;
import android.view.MotionEvent;
import android.view.View;
import android.view.ViewParent;
import android.view.WindowManager;
import android.view.animation.DecelerateInterpolator;
import android.widget.Button;
import android.widget.CheckBox;
import android.widget.CompoundButton;
import android.widget.HorizontalScrollView;
import android.widget.ImageButton;
import android.widget.LinearLayout;
import android.widget.TextView;
import android.widget.ToggleButton;

import androidx.annotation.NonNull;
import androidx.appcompat.app.AppCompatActivity;
import androidx.core.app.ActivityCompat;
import androidx.recyclerview.widget.LinearLayoutManager;
import androidx.recyclerview.widget.RecyclerView;

import com.google.android.material.snackbar.Snackbar;

import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.File;
import java.io.FileFilter;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.nio.charset.Charset;
import java.nio.charset.StandardCharsets;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Calendar;
import java.util.Comparator;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Locale;
import java.util.Properties;
import java.util.concurrent.Semaphore;

import rabbit.dcs.databinding.ActivityMainBinding;

public class MainActivity extends AppCompatActivity implements SelectionOwner {
	public static final boolean _ERROR_LOG = true;
	public static final boolean _DEBUG_LOG = false;
	public static final boolean _DEBUG_LOG_NOW = false;

	public static final boolean _DEBUG_SORT_DIR = false;

	static final String _INTERNAL_STORAGE_RKEY = "^/storage/emulated/[0-9]+/.*";

	// Used to load the 'dcs' library on application startup.
	static {
		System.loadLibrary("dcs");
	}

	static MainActivity _mainActivity;

	static Values _values;

	static final int _OPTION_YES = 0;

	static final int _OPTION_NO = 1;

	static final int _OPTION_CANCEL = 3;

	static final int _CHECK_TIME = 500;

	static final int _DURATION_TIME = 800;

	static final int _MAX_SCROLL_DEPTH = 5000;

	static final int _MAX_CONVERT_COUNT = 1200;

	static final int _MAX_DISPLAY_COUNT = 3000;

	static final int _PROGRESS_CYCLE = 10;

	static final int S_SPRING = 0;
	static final int S_SUMMER = 1;
	static final int S_FALL = 2;
	static final int S_WINTER = 3;

	static final String _LOG_DIR = "dcs_log";

	static final String _LOG_NAME = "log.txt";

	static final String KEY_LIST = "LIST";
	static final String KEY_EDIT = "EDIT";

	public static final int WARNING_COLOR = 0xff5050ff;

	private ActivityMainBinding _binding;

	static final int _HSV_ROWOFFSET = 94;
	static final int _ROW_SCROLL_OFFSET = 170;
	static final int _CB_WIDTH = 96;

	static final int _PATH_TEXT_SIZE = 20;

	static final float _W_BGCOLOR = 0.6F;
	static final float _W_BGCOLOR_TEXT = 1.1F;
	static final float _W_FGCOLOR_TEXT = 0.3F;
	static final float _W_BGCOLOR_BUTTON = 0.5F;
	static final float _W_FGCOLOR_BUTTON = 1.8F;
	static final float _W_BGCOLOR_BUTTON_DISABLED = 0.8F;
	static final float _W_FGCOLOR_BUTTON_DISABLED = 1.0F;

	static final String _ROOT_PATH = "/storage";

	TreeNode _node;

	int _stepCount;

	boolean _isNoType;

	boolean _isSelectTtype;

	Hashtable<String, String> _hSuffix;

	SelectionControler _selectionControler;

	ScrollManager _scrollManager = new ScrollManager();

	RowInfo[] _ainfo;

	Hashtable<String, TreeNode> _hNode;

	static final String _PROPERTIES_NAME = ".properties";

	Properties _prop = new Properties();

	static final String _property_key_cpath = "cpath";
	static final String _property_key_help_zoom = "help_zoom";
	static final String _property_key_help_y = "help_y";
	static final String _property_key_condttype = "condttype";
	static final String _property_key_ttypes = "ttypes";
	static final String _property_key_btypes = "btypes";
	static final String _property_key_notype = "notype";
	static final String _property_key_stype = "stype";
	static final String _property_key_dtype = "dtype";

	static final String _property_key_ln_judge = "ln_judge";
	static final String _property_key_mac_judge = "mac_judge";

	static final String _property_key_ln_edit = "ln_edit";
	static final String _property_key_ln_dest = "ln_dest";
	static final String _property_key_mac1char = "mac1char";
	static final String _property_key_overr_same = "overr_same";
	// JIS, JIS2: EUC, EUC1, EUC2, EUC3
	static final String _property_key_overr_similar = "overr_similar";

	static final String _property_key_pinning_conv = "pinning_conv";
	static final String _property_key_pinning_main = "pinning_main";

	private static final Integer[] icons = {
			R.drawable.i000_wagahaiwa,
			R.drawable.i001_takekurabe,
			R.drawable.i002_butokai,
			R.drawable.i003_umareizuru,
			R.drawable.i004_bokutokidan,
			R.drawable.i005_hatukoi,
			R.drawable.i006_dotei,
			R.drawable.i007_takasebune,
			R.drawable.i008_tumitobatu,
			R.drawable.i009_nitirin,
			R.drawable.i010_tonomonogatari,
			R.drawable.i011_oninohanasi,
			R.drawable.i012_sangetuki,
			R.drawable.i013_remon,
			R.drawable.i014_onsyunokanatani,
			R.drawable.i015_miminasihoiti,
	};

	int _ii;

	static final int _ICON_SIZE_1 = 15;

	static int _text_highlight_color;

	public static void setParentBackground(View child, int bgcolor) {
		View view = child;
		for(;;) {
			ViewParent vp = view.getParent();
			if (!(vp instanceof View)) {
				break;
			}
			view = (View)vp;
			view.setBackgroundColor(bgcolor);
		}
	}

	public static class ForegroundColorBoldSpan extends ForegroundColorSpan {
		public ForegroundColorBoldSpan(int color) {
			super(color);
		}
		@Override
		public void updateDrawState(TextPaint textPaint) {
			super.updateDrawState(textPaint);
			textPaint.setTypeface(Typeface.defaultFromStyle(Typeface.BOLD));
		}
	}

	public static void setStatusColor(SpannableStringBuilder ssb, int sts, String strType, String strLType, String strMac, String strMessage) {
		ssb.append(' ');
		if (sts == RowInfo.STS_SUCCESS) {
			ssb.append(strType).append(' ').append(strLType).append(' ').append(strMac);
		} else {
			int len = ssb.length();
			boolean isWarning = true;
			// ワーニングの背景色を設定
			ssb.append(strType);
			int tsts = (sts & (RowInfo.STS_WNOLN | RowInfo.STS_WMIXEDLN));
			if (isWarning && tsts == 0) {
				ssb.setSpan(
						new ForegroundColorBoldSpan(MainActivity.WARNING_COLOR),
						len, // start
						ssb.length(), // end
						Spannable.SPAN_EXCLUSIVE_EXCLUSIVE
				);
			}
			ssb.append(' ').append(strLType);
			int len2 = ssb.length();
			ssb.append(' ').append(strMac).append(' ').append(strMessage);

			isWarning = (tsts != 0);
			tsts = (sts & RowInfo.STS_WUTF8_MAC);
			if (isWarning && tsts == 0) {
				ssb.setSpan(
						new ForegroundColorBoldSpan(MainActivity.WARNING_COLOR),
						len, // start
						len2, // end
						Spannable.SPAN_EXCLUSIVE_EXCLUSIVE
				);
			} else if (!isWarning && tsts != 0) {
				ssb.setSpan(
						new ForegroundColorBoldSpan(MainActivity.WARNING_COLOR),
						len2, // start
						len2 + 1 + strMac.length(), // end
						Spannable.SPAN_EXCLUSIVE_EXCLUSIVE
				);
			}
		}
	}

	public static void setButtonColor(Button btn) {
		int fgcolor_button, bgcolor_button;

		if (btn.isEnabled()) {
			bgcolor_button = MainActivity.getSeasonColor(MainActivity._W_BGCOLOR_BUTTON);
			fgcolor_button = MainActivity.getSeasonColor(MainActivity._W_FGCOLOR_BUTTON);
		} else {
			bgcolor_button = MainActivity.getSeasonColor(MainActivity._W_BGCOLOR_BUTTON_DISABLED);
			fgcolor_button = MainActivity.getSeasonColor(MainActivity._W_FGCOLOR_BUTTON_DISABLED);
		}
        btn.setBackgroundColor(bgcolor_button);
		btn.setTextColor(fgcolor_button);
	}

	@Override
	protected void onStop() {
		super.onStop();
		try {
			saveProperties();
		} catch(IOException e) {
if (MainActivity._ERROR_LOG) {
RLog.e("Dcs", e.getMessage(), e);
}
		}
	}

	@Override
	protected void onDestroy () {
		super.onDestroy();
		setRequestedOrientation(ActivityInfo.SCREEN_ORIENTATION_UNSPECIFIED);
	}

	public static MainActivity getContext() {
		return _mainActivity;
	}

	@Override
	public void finish() {
		if (!_isPinning) {
			// ピン止めでなければ閉じる
			super.finish();
		}
	}

	@Override
	public void finishActivity (int requestCode) {
		if (!_isPinning) {
			// ピン止めでなければ閉じる
			super.finishActivity(requestCode);
		}
	}

	boolean _isPinning;

	/*
    初期表示はこのタイミングでないと RecyclerView のサイズが取れない
    ２回め以降は Mainadapter.onBindViewHolder から setListeners を呼び出す
     */
	boolean _isInit;

	@Override
	public void onWindowFocusChanged(boolean hasFocus) {
		super.onWindowFocusChanged(hasFocus);

		if (_isInit) {
			return;
		}
		_isInit = true;
		/*
		/storage/emulated/0/Android/data/rabbit.dcs/cache
		上記のパスに他のソフトから簡単にアクセスできないのでここで createLog() を呼び出して下記の中に出力していたが
		システムのファイル・マネージャーでアクセスできたので onCreate() での上記のパスに統一した
		/storage/emulated/0/dcs_log
		 */
		if (_logStackTrace != null) {
			SpannableStringBuilder ssb = new SpannableStringBuilder();
			int i, len = _logStackTrace.length;
			for(i=0;i<len;i++) {
				ssb.append(_logStackTrace[i].toString()).append('\n');
			}
			AlertDialog.Builder builder = new AlertDialog.Builder(MainActivity.this)
					.setTitle(_values.logGenerateErrorTitle)
					.setMessage(ssb.toString())
					.setPositiveButton(_values.ok, new DialogInterface.OnClickListener() {
						@Override
						public void onClick(DialogInterface dialog, int which) {
						}
					});
			builder.show();
			_logStackTrace = null;
		}
		RecyclerView rvList = findViewById(R.id.rvList);
		int j, jlen = rvList.getChildCount();
		for(j=0;j<jlen;j++) {
			View parent = rvList.getChildAt(j);
			MainAdapter.ViewHolder holder = (MainAdapter.ViewHolder) rvList.getChildViewHolder(parent);
			// リスナー登録
			setListeners(parent, holder);
		}
		// 更新メニューを無効にする
		setSelectedCount(getSelectedCount());
//
		TreeNode node = _node.getRoot();
		node = node.getChild(0);
		if (node.rowInfo.isEmptyDirectory) {
		// 内部ストレージが無効の時
			createPermiDialog().show();
		}
	}

	//V11	@RequiresApi(api = Build.VERSION_CODES.R)
	@Override
	protected void onCreate(Bundle savedInstanceState) {
		super.onCreate(savedInstanceState);

		setRequestedOrientation(ActivityInfo.SCREEN_ORIENTATION_PORTRAIT);
		_mainActivity = this;
        _seasonIndex = -1;
		Locale locale = Locale.getDefault();
		_values = (Locale.JAPAN.equals(locale) ? new ValuesJp() : new ValuesEn());
		createLog();
		loadProperties();
		if (!checkPermission()) {
		// パーミッションの変更がなかった時
			// 初期表示
			initializing();
		}
	}

	/* 初期表示 */
	void initializing() {
		_binding = ActivityMainBinding.inflate(getLayoutInflater());
		// ルート表示ボタン
		createTouchListener(_binding.ibRoot, Color.argb(100, 200, 255, 100), new TouchCallback() {
			@Override
			public void execute() {
				synchronized (_wdlg[0]) {
					_wdlg[0].dismiss();
					// ２回目以降はルートも生成(_wdlg[0].isCancel()判定のため)
					_wdlg[0] = createWaitingDialog(true);
					_wdlg[0].setWaiting();
				}
				drawRoot();
			}
		});
		// すべて選択テキスト・ビュー
		_binding.tvSelectAll.setOnClickListener(new View.OnClickListener() {
			@Override
			public void onClick(View view) {
				CheckBox cbSelectAll = findViewById(R.id.cbSelectAll);
				cbSelectAll.performClick();
			}
		});
		// 戻るボタン
		int color = getSeasonColor(0.8F);
		createTouchListener(_binding.ibRevert, Color.argb(100, (color>>16)&0xff, (color>>8)&0xff, color&0xff), new TouchCallback() {
			public void execute() {
				LinearLayout llPath = findViewById(R.id.llPath);
				gotoWithIndex(llPath.getChildCount() - 2);
			}
		});
		// ピン止め
		ToggleButton tbPinninng = _binding.tbPinning;
		tbPinninng.setTextOn("");
		tbPinninng.setTextOff("");
		_isPinning = String.valueOf(true).equals(
				getProperty(MainActivity._property_key_pinning_main));
		tbPinninng.setChecked(_isPinning);
		tbPinninng.setBackgroundResource(_isPinning? R.drawable.pinning1: R.drawable.pinning0);
		tbPinninng.setOnCheckedChangeListener(new CompoundButton.OnCheckedChangeListener() {
			@Override
			public void onCheckedChanged(CompoundButton cbPinning, boolean isChecked) {
				tbPinninng.setBackgroundResource(isChecked?
						R.drawable.pinning1: R.drawable.pinning0);
				_isPinning = isChecked;
				setProperty(MainActivity._property_key_pinning_main, String.valueOf(isChecked));
			}
		});
		// 最下端移動ボタン
		createTouchListener(_binding.ibDown, Color.argb(100, 200, 255, 100), new TouchCallback() {
			@Override
			public void execute() {
				setDownList(_binding.rvList);
			}
		});
		// 最上端移動ボタン
		createTouchListener(_binding.ibUp, Color.argb(100, 200, 255, 100), new TouchCallback() {
			@Override
			public void execute() {
				setUpList(_binding.rvList);
			}
		});
		int bgcolor0 = MainActivity.getSeasonColor(1.1F);
		_binding.rvList.setBackgroundColor(bgcolor0);
		_binding.rvList.setOnTouchListener(new View.OnTouchListener() {
			@Override
			public boolean onTouch(View view, MotionEvent evt) {
				switch (evt.getAction()) {
					default:
						break;
					case MotionEvent.ACTION_DOWN:
						_scrollManager.isContains(_binding.rvList, evt.getRawX(), evt.getRawY(), "initializing");
						break;
					case MotionEvent.ACTION_MOVE:
						_scrollManager.moveTo(_binding.rvList, evt.getRawX(), evt.getRawY(), false, "initializing");
						break;
					case MotionEvent.ACTION_UP:
						_scrollManager._isInside = false;
						break;
					case MotionEvent.ACTION_CANCEL:
						break;
				}
				return false;
			}
		});
		setContentView(_binding.getRoot());
		// 選択チェック・ボックス共通処理
		_selectionControler = new SelectionControler(this) {

			@Override
			public int getSelectedCount() {
				return MainActivity.this.getSelectedCount();
			}

			@Override
			public void setSelectedCount(int count) {
				MainActivity.this.setSelectedCount(count);
			}

			@Override
			public RecyclerView getRecyclerViewList() {
				return findViewById(R.id.rvList);
			}

			@Override
			public CheckBox getCheckBoxSelectAll() {
				return findViewById(R.id.cbSelectAll);
			}
		};
//
		setTitle(MainActivity._values.title);
		// すべて選択
		TextView tvSelectAll = findViewById(R.id.tvSelectAll);
		tvSelectAll.setText(MainActivity._values.selectAll);
		// 選択数
		TextView tvValue = findViewById(R.id.tvValue);
		tvValue.setText(MainActivity._values.value); // "　選択数"
		// 日本語／英語でフォント・サイズを変える
		if (_values instanceof ValuesJp) {
			tvSelectAll.setTextSize(10F);
			tvValue.setTextSize(12F);
		} else {
			tvSelectAll.setTextSize(12F);
			tvValue.setTextSize(12F);
		}
		// 件
		TextView tvValue2 = findViewById(R.id.tvValue2);
		tvValue2.setText(MainActivity._values.value2); // "件"
//
		// 階層パス
		LinearLayout llPath = findViewById(R.id.llPath);
		_text_highlight_color = MainActivity.getSeasonColor(0.3F);
		llPath.setBackgroundColor(bgcolor0);
		setParentBackground(llPath, bgcolor0);
		// 戻るボタン
		setBackground(_binding.ibRevert);
		// 最上／下端移動ボタン
		LinearLayout llUp = findViewById(R.id.llUp);
		LinearLayout llDown = findViewById(R.id.llDown);
//
		ImageButton ibUp = findViewById(R.id.ibUp);
		ImageButton ibDown = findViewById(R.id.ibDown);
		int iDown, iUp;
		switch (MainActivity.getSeasonIndex()) {
			default:
			case S_SPRING:
				llUp.setBackgroundResource(R.drawable.shape_sp);
				llDown.setBackgroundResource(R.drawable.shape_sp);
				break;
			case S_SUMMER:
				llUp.setBackgroundResource(R.drawable.shape_su);
				llDown.setBackgroundResource(R.drawable.shape_su);
				break;
			case S_FALL:
				llUp.setBackgroundResource(R.drawable.shape_fa);
				llDown.setBackgroundResource(R.drawable.shape_fa);
				break;
			case S_WINTER:
				llUp.setBackgroundResource(R.drawable.shape_wi);
				llDown.setBackgroundResource(R.drawable.shape_wi);
				break;
		}
		iUp = R.drawable.ic_baseline_arrow_upward_40;
		iDown = R.drawable.ic_baseline_arrow_downward_40;
		// 背景色を不可視にする（値は何でも可―ImageButton の内部仕様もしくはバグ）
		ibDown.setBackgroundColor(0x202020);
		ibUp.setBackgroundColor(0x202020);
//
		ibUp.setImageResource(iUp);
		ibDown.setImageResource(iDown);
		_node = createRootNode();
		// ルート表示
		drawRoot();
	}

	static void setDownList(RecyclerView rvList) {
		int ccnt = rvList.getChildCount();
		if (ccnt == 0) {
			return;
		}
		RecyclerView.Adapter adapter = rvList.getAdapter();
		if (adapter == null) {
			if (MainActivity._ERROR_LOG) {
				RLog.e("Dcs", "initializing: ibDown.onClick: RecyclerView.getAdapter() == null");
			}
		} else {
			// 最下端表示のビュー
			View view = rvList.getChildAt(ccnt - 1);
			// 最下端表示のビューの全体の位置
			int apos = rvList.getChildAdapterPosition(view);
			int vh = view.getHeight();
			int icnt = adapter.getItemCount();
			// 最下端表示のビューの最低辺までの距離
			int h = vh * (icnt - apos);
            // ページ数
			int times = (h + _MAX_SCROLL_DEPTH - 1) / _MAX_SCROLL_DEPTH;
			int depth = h / times;
			if (times >= 2) {
				rvList.scrollToPosition(icnt - 1 - depth / vh);
			}
			rvList.post(new Runnable() {
				@Override
				public void run() {
					rvList.smoothScrollBy(0, depth, new DecelerateInterpolator(), _DURATION_TIME);
				}
			});
		}
	}

	static void setUpList(RecyclerView rvList) {
		int ccnt = rvList.getChildCount();
		if (ccnt == 0) {
			return;
		}
		RecyclerView.Adapter adapter = rvList.getAdapter();
		if (adapter == null) {
			if (MainActivity._ERROR_LOG) {
				RLog.e("Dcs", "initializing: ibDown.onClick: RecyclerView.getAdapter() == null");
			}
		} else {
			// 最上端表示のビュー
			View view = rvList.getChildAt(0);
			// 最上端表示のビューの全体の位置
			int apos = rvList.getChildAdapterPosition(view);
			int vh = view.getHeight();
			// 最上端表示のビューの最上辺までの距離
			int h = vh * (apos + 1);
			// ページ数
			int times = (h + _MAX_SCROLL_DEPTH - 1) / _MAX_SCROLL_DEPTH;
			int depth = h / times;
			if (times >= 2) {
				rvList.scrollToPosition(depth / vh);
			}
			rvList.post(new Runnable() {
				@Override
				public void run() {
					rvList.smoothScrollBy(0, -depth, new DecelerateInterpolator(), _DURATION_TIME);
				}
			});
		}
	}
	/* イメージ／ボタンの共通処理 */
	static void createTouchListener(ImageButton imageButton, int argb, TouchCallback touchCallback) {
		boolean[] isMoved = new boolean[1];
		imageButton.setOnTouchListener(new View.OnTouchListener() {
			//V11			@RequiresApi(api = Build.VERSION_CODES.R)
			@Override
			public boolean onTouch(View view, MotionEvent evt) {
				try {
					switch (evt.getAction()) {
						case MotionEvent.ACTION_DOWN:
							isMoved[0] = false;
							((ImageButton) view).setColorFilter(Color.argb((argb>>24)&0xff, (argb>>16)&0xff, (argb>>8)&0xff, argb&0xff));
							return false;
						case MotionEvent.ACTION_MOVE:
							isMoved[0] = true;
							return false;
						case MotionEvent.ACTION_UP:
							if (!isMoved[0]) {
								// 移動していなかった時
								touchCallback.execute();
							}
							// no break
						case MotionEvent.ACTION_CANCEL:
							((ImageButton) view).setColorFilter(null);
							break;
					}
				} catch(Exception e) {
					if (MainActivity._ERROR_LOG) {
						RLog.e("Dcs", e.getMessage(), e);
					}
				}
				return false;
			}
		});
	}

	// ルート・ノード生成
	static TreeNode createRootNode() {
		RowInfo info = new RowInfo();
		info.fileName = info.displayName = info.path = _ROOT_PATH;
		info.isDirectory = true;
		return new TreeNode(info);
	}

	// ボタン・イメージ生成
	static void setBackground(ImageButton ib) {
		// 背景色をアイコンの縁に合わせる
		BitmapDrawable bd = (BitmapDrawable) ib.getDrawable();
		Bitmap bm = bd.getBitmap();
		ib.setBackgroundColor(bm.getPixel(0, 0));
	}

	@Override
	protected void onStart() {
		super.onStart();
	}

	@Override
	protected void onResume() {
		super.onResume();
	}

	private static final int REQUEST_CODE = 1000;

	/*
	パーミッションの扱いは正しいところが不明。
	ローカル・ファイルを扱うにはバージョンを下げなければならないという巷の情報に従ったところ、アイコンなどの環境が壊れた。
	最低限の修正を施したが完全な復旧は行わない。
	後にMANAGE_EXTERNAL_STORAGEを登録する趣旨のことが本家のサイトにあったので追加した。
	 */
	private boolean checkPermission() {
//DEBUG		Call requires API level 30 (current min is 28): `android.os.Environment#isExternalStorageManager`
//DEBUG		boolean flag = Environment.isExternalStorageManager();
		boolean isReq = false;
		if (Build.VERSION.SDK_INT >= 23) {
			String[] aperm = {
					Manifest.permission.READ_EXTERNAL_STORAGE,
					Manifest.permission.WRITE_EXTERNAL_STORAGE,
					Manifest.permission.MANAGE_EXTERNAL_STORAGE,
			};
			if (ActivityCompat.checkSelfPermission(this, aperm[2])
					!= PackageManager.PERMISSION_GRANTED) {
			// 許可されていない時
//DEBUG!!				if (!ActivityCompat.shouldShowRequestPermissionRationale(this, aperm[1])) {
				// 「今後、表示しない」チェックなしの時
					ActivityCompat.requestPermissions(MainActivity.this,
							aperm,
							REQUEST_CODE);
					// onRequestPermissionsResult で継続する
					isReq = true;
//DEBUG!!				}
			}
		}
		return isReq;
	}

	@Override
	public boolean onCreateOptionsMenu(Menu menu) {
		MenuInflater inflater = getMenuInflater();
		inflater.inflate(R.menu.menu_options, menu);
		MenuItem mi = menu.findItem(R.id.miConvert);
//
		mi.setTitle(MainActivity._values.convert);
		mi = menu.findItem(R.id.miSettings);
		mi.setTitle(MainActivity._values.settings);
		mi = menu.findItem(R.id.miHelp);
		mi.setTitle(MainActivity._values.help);
		mi = menu.findItem(R.id.miQuit);
		mi.setTitle(MainActivity._values.quit);
		return true;
	}

	/* 更新後の選択画面への反映 */
	void restoreRowInfo(RowInfo[] ainfo) {
		int i, len = ainfo.length;
		for(i=0;i<len;i++) {
			TreeNode node = _hNode.get(ainfo[i].path);
			if (node == null) {
				if (MainActivity._ERROR_LOG) {
					RLog.e("Dcs", "restoreRowInfo: node notfound! " + ainfo[i].path);
				}
			} else {
				node.rowInfo = ainfo[i];
			}
			ainfo[i] = _ainfo[i] = null;
		}
		len = _ainfo.length;
		for(;i<len;i++) {
			_ainfo[i] = null;
		}
		_hNode.clear();
		_hNode = null;
		_ainfo = null;
	}

	/* ConvertActivity 以外での RowInfo のクリア */
	void clearRowInfo(RowInfo[] ainfo) {
		int i, len = ainfo.length;
		for(i=0;i<len;i++) {
			ainfo[i] = null;
		}
		_hNode.clear();
		_hNode = null;
		_ainfo = null;
	}

	// オプション・アイテム・リスナー
	@Override
	public boolean onOptionsItemSelected(MenuItem item) {
try {
	int itemId = item.getItemId();
	if (itemId == R.id.miConvert) { // 変換
		WaitDialog[] wdlg = new WaitDialog[1];
		// 処理中ダイアログ
		AlertDialog.Builder builder = new AlertDialog.Builder(MainActivity.this)
			.setPositiveButton(_values.ok, new DialogInterface.OnClickListener() {
				@Override
				public void onClick(DialogInterface dialog, int which) {
					wdlg[0].setWaiting();
					// 時間差による処理中ダイアログ表示
					checkToShowWaitDialog(wdlg[0]);
//
					new Thread() {
						@Override
						public void run() {
							try {
								// 処理対象ファイルを確認せずに更新
								executeWithoutConfirmation(_ainfo, wdlg);
								synchronized (wdlg[0]) {
									wdlg[0].dismiss(); // 処理中ダイアログを閉じる
								}
							} catch(Exception e) {
								if (MainActivity._ERROR_LOG) {
									RLog.e("Dcs", e.getMessage(), e);
								}
							}
						}
					}.start();
				}
			})
			.setNegativeButton(_values.cancel, new DialogInterface.OnClickListener() {
				@Override
				public void onClick(DialogInterface dialog, int which) {
					/* RowInfo のクリア */
					clearRowInfo(_ainfo);
				}
			});
		AlertDialog dlg = builder.create();
		wdlg[0] = createWaitingDialog(true);
		wdlg[0].setWaiting();
//
		new Thread(){
			@Override
			public void run() {
				try {
					// 選択リスト取得
					_ainfo = getSelectedInfoList(wdlg);
					try {
						Thread.sleep(100);
					} catch(InterruptedException e) {
						if (MainActivity._ERROR_LOG) {
							RLog.e("Dcs", e.getMessage(), e);
						}
					}
					synchronized (wdlg[0]) {
						wdlg[0].dismiss(); // 処理中ダイアログを閉じる
					}
					int size = _ainfo.length;
					if (size <= _MAX_CONVERT_COUNT) {
					// 処理対象ファイルが最大数以下の時
						Bundle bundle = new Bundle();
						bundle.putParcelableArray(MainActivity.KEY_LIST, _ainfo);
						Intent intent = new Intent(MainActivity.this, ConvertActivity.class);
						intent.putExtras(bundle);
						startActivity(intent);
					} else {
					// 処理対象ファイルが最大数を超えた時
						new AsyncTask<Void, Void, Boolean>() {
							@Override
							protected Boolean doInBackground(Void... params) {
								return true;
							}

							@Override
							protected void onPostExecute(Boolean result) {
								try {
									// "処理対象ファイルが最大数を超えました(%d > %d)。確認せずに更新しますか？";
									String msg = String.format(_values.updateWithoutConfirmation, size, _MAX_CONVERT_COUNT);
									dlg.setMessage(msg);
									dlg.show();
								} catch(Exception e) {
									if (MainActivity._ERROR_LOG) {
										RLog.e("Dcs", e.getMessage(), e);
									}
								}
							}
						}.execute();
					}
				} catch(Exception e) {
					if (MainActivity._ERROR_LOG) {
						RLog.e("Dcs", e.getMessage(), e);
					}
				}
			}
		}.start();
	} else if (itemId == R.id.miSettings) { // 設定
		Intent intent = new Intent(MainActivity.this, SettingsActivity.class);
		startActivity(intent);
	} else if (itemId == R.id.miHelp) { // ヘルプ
		Intent intent = new Intent(MainActivity.this, HelpActivity.class);
		startActivity(intent);
	} else if (itemId == R.id.miQuit) { // 終了
		super.finish();
	} else {
		if (MainActivity._ERROR_LOG) {
			RLog.e("Dcs","onOptionsItemSelected: Unknown itemId found!("+itemId+")");
		}
	}
} catch(Exception e) {
	if (MainActivity._ERROR_LOG) {
		RLog.e("Dcs", e.getMessage(), e);
	}
	}
		return super.onOptionsItemSelected(item);
	}

	// 時間差による処理中ダイアログ表示
	void checkToShowWaitDialog(WaitDialog wdlg) {
		new Thread() {
			public void run() {
				try {
					Thread.sleep(_CHECK_TIME);
				} catch (InterruptedException e) {
                    if (MainActivity._ERROR_LOG) {
                        RLog.e("Dcs", e.getMessage(), e);
                    }
				}
				synchronized (wdlg) {
					if (wdlg.isWaiting()) {
						// 終了していない場合、表示
						new AsyncTask<Void, Void, Boolean>() {
							@Override
							protected Boolean doInBackground(Void... params) {
								return true;
							}

							@Override
							protected void onPostExecute(Boolean result) {
								try {
									// 終了していない場合、表示
									wdlg.show();
								} catch (Exception e) {
									if (MainActivity._ERROR_LOG) {
										RLog.e("Dcs", e.getMessage(), e);
									}
								}
							}
						}.execute();
					}
				}
			}
		}.start();
	}

	class WaitDialog {
		AlertDialog _dlg;
		boolean _isWaiting;
		boolean _isCancel;

		WaitDialog(String title, String text, boolean isOpen) {
			AlertDialog.Builder builder = new AlertDialog.Builder(MainActivity.this)
					.setMessage(title)
					.setPositiveButton(text, new DialogInterface.OnClickListener() {
						@Override
						public void onClick(DialogInterface dialog, int which) {
							_isCancel = true;
						}
					});
			_dlg = builder.create();
			if (isOpen) {
				// 時間差による処理中ダイアログ表示
				checkToShowWaitDialog(this);
			}
		}

		public void setMessage(String message) {
			_dlg.setMessage(message);
		}

		public void show() {
			_dlg.show();
		}

		public void dismiss() {
			_isWaiting = false;
			_dlg.dismiss();
		}

		public void setWaiting() {
			_isWaiting = true;
		}

		public boolean isWaiting() {
			return _isWaiting;
		}

		public boolean isCancel() {
			return _isCancel;
		}
	}

	// 処理中ダイアログ生成／表示
	WaitDialog createDialog(String title, String text, boolean isOpen) {
		return new WaitDialog(title, text, isOpen);
	}

	WaitDialog createPermiDialog() {
		return createDialog(_values.permiTitle, _values.ok, false);
	}

	WaitDialog createWaitingDialog(boolean isOpen) {
		return createDialog(_values.waitTitle, _values.stop, isOpen);
	}

	@Override
	public void onRequestPermissionsResult(int requestCode, @NonNull String[] permission, @NonNull int[] grantResults) {
		super.onRequestPermissionsResult(requestCode, permission, grantResults);
try {
	// 外部ストレージのアクセス可
		// 初期表示
		initializing();
} catch(Exception e) {
	if (MainActivity._ERROR_LOG) {
		RLog.e("Dcs", e.getMessage(), e);
	}
}
	}

	StackTraceElement[] _logStackTrace;

	File getPrivateDir() {
		File[] list = getExternalCacheDirs();
// /storage/emulated/0/Android/data/rabbit.dcs/cache
// /storage/0BE7-2E16/Android/data/rabbit.dcs/cache
		int i, len = list.length;
		for(i=0;i<len;i++) {
//			if (list[i].getPath().startsWith("/storage/emulated/0/")) {
			if (list[i].getPath().matches(_INTERNAL_STORAGE_RKEY)) {
				return list[i];
			}
		}
		return null;
	}

	void createLog() {
		try {
			File cdir = getPrivateDir();
			File edir = new File(cdir, _LOG_DIR);
			edir.mkdirs();
			File[] clist = edir.listFiles();
			if (clist != null) {
			// getExternalCacheDirs からのパスなので(clist != null)は不要
				int i, len = clist.length;
				for(i=0;i<len;i++) {
					clist[i].delete();
				}
			}
			RLog rlog = new RLog(new File(edir.getPath(), _LOG_NAME).getPath());
			rlog.setLogPrint(false);
//!!			rlog.setFilePrint(false);
		} catch (IOException e) {
			_logStackTrace = e.getStackTrace();
		}
	}

	// 処理対象ファイルを確認せずに更新
	void executeWithoutConfirmation(RowInfo[] ainfo, WaitDialog[] wdlg) {
		try {
			int[] astype = new int[1];
			int[] adtype = new int[1];
			int[] atype = new int[1];
			int[] altype =
					(String.valueOf(true).equals(getProperty(_property_key_ln_judge))?
							new int[1]: null);
			DcsConvertObserver[] observer = new DcsConvertObserver[1];
			Dcs dcs = ConvertActivity.createDcs(astype, adtype, observer);
//
			int i, len = ainfo.length;
			for(i=0;i<len;i++) {
				if (wdlg[0].isCancel()) {
					break;
				}
/*DEL _PROGRESS_CYCLE*
				if ((i + 1) % _PROGRESS_CYCLE == 0) {
					wdlg[0].setMessage(_values.waitTitle + (i + 1));
				}
/*DEL _PROGRESS_CYCLE*/
				RowInfo info = ainfo[i];
				dcs.convertFile(info.path, astype[0], adtype[0], false, observer[0], observer[0], info);
				if (info.os != null) {
					// OutputStream が設定されている場合、閉じる
					try {
						info.os.close();
					} catch (IOException e) {
						if (MainActivity._ERROR_LOG) {
							RLog.e("Dcs", e.getMessage(), e);
						}
					}
					info.os = null;
					// RowInfo を更新
					MainActivity.judgesFile(dcs, atype, altype, info);
				}
			}
			dcs.close();
			/* RowInfo のクリア */
			clearRowInfo(ainfo);
			RecyclerView rvList = findViewById(R.id.rvList);
			RecyclerView.Adapter adapter = rvList.getAdapter();
			if (adapter == null) {
				if (MainActivity._ERROR_LOG) {
					RLog.e("Dcs", "executeWithoutConfirmation: RecyclerView.getAdapter() == null");
				}
			} else {
				new AsyncTask<Void, Void, Boolean>() {
					@Override
					protected Boolean doInBackground(Void... params) {
						return true;
					}

					@Override
					protected void onPostExecute(Boolean result) {
						try {
							// 表示されている明細を更新
							int i, len = rvList.getChildCount();
							for(i=0;i<len;i++) {
								adapter.notifyItemChanged(i);
							}
						} catch(Exception e) {
							if (MainActivity._ERROR_LOG) {
								RLog.e("Dcs", e.getMessage(), e);
							}
						}
					}
				}.execute();
			}
		} catch (DcsException e) {
			if (MainActivity._ERROR_LOG) {
				RLog.e("Dcs", e.getMessage(), e);
			}
		}
	}

	// ルート・リスト取得
	ArrayList<File> getRootList() {
		File idir = null, edir = null;
		// /storage/emulated/0/Android/data/rabbit.dcs/files
		// /storage/0BE7-2E16/Android/data/rabbit.dcs/files
		File[] dirExternalFiles = getExternalFilesDirs(null);
		int i, len = dirExternalFiles.length;
		for(i=0;i<len;i++) {
			if (_wdlg != null && _wdlg[0].isCancel()) {
				break;
			}
			if (dirExternalFiles[i] == null) {
				continue;
			}
			String str = dirExternalFiles[i].getPath();
			int pos = str.indexOf("/Android/");
			// /storage/emulated/0/Android/data/rabbit.dcs/files
			// ...................+
			File dir = new File(str.substring(0, pos));
			// "/storage/emulated/0/"
			if (str.matches(_INTERNAL_STORAGE_RKEY)) {
			// 内部ストレージの時
				idir = dir;
			} else {
				edir = dir;
			}
		}
		ArrayList<File> list = new ArrayList<>();
		if (idir != null) {
			list.add(idir);
		}
		if (edir != null) {
			list.add(edir);
		}
		return list;
	}

	// ルート表示
	void drawRoot() {
		// 現在のレベルからルートまで後退
		_stepCount = - _node.getLevel();
		if (_stepCount != 0) {
		// ルートに移動した時
			savePosition();
		}
		_node = _node.getLevelParent(0);
		LinearLayout llPath = findViewById(R.id.llPath);
		llPath.removeAllViews();
		ArrayList<File> list = getRootList();
		draw(list);
	}

	// パス位置に移動
	public void gotoWithIndex(int index) {
		if (index < 0) {
		// 先頭の時(childCount(1)-2)
			synchronized (_wdlg[0]) {
				_wdlg[0].dismiss();
				// ２回目以降はルートも生成(_wdlg[0].isCancel()判定のため)
				_wdlg[0] = createWaitingDialog(true);
				_wdlg[0].setWaiting();
			}
			drawRoot();
			return;
		}
		savePosition();
		LinearLayout llPath = findViewById(R.id.llPath);
		int len = llPath.getChildCount();
		if (len - 1 > index) {
			llPath.removeViews(index + 1, len - 1 - index);
		}
		// 現在のレベルからインデックスまで後退
		_stepCount = index - (len - 1);
		// ルートはパスに含まれない(+1)
		_node = _node.getLevelParent(index + 1);
		TextView view = (TextView) llPath.getChildAt(index);
		view.setTextColor(_text_highlight_color);
		view.setTypeface(Typeface.DEFAULT_BOLD);
		drawByInfo(_node.rowInfo);
	}

	// RowInfo による表示
	void drawByInfo(RowInfo info) {
		ArrayList<File> list = new ArrayList<>();
		File[] childs = info.file.listFiles();
RLog.d("Dcs", ">drawByInfo "+childs.length+": "+info.file.getPath());
		if (childs == null) {
			if (MainActivity._ERROR_LOG) {
				RLog.e("Dcs","drawByInfo "+info.path+" listFiles == null!");
			}
			return;
		}
		synchronized (_wdlg[0]) {
			_wdlg[0].dismiss();
			_wdlg[0] = createWaitingDialog(true);
			_wdlg[0].setWaiting();
		}

		new Thread() {
			@Override
			public void run() {
				int i, len = childs.length;
				for(i=0;i<len;i++) {
RLog.d("Dcs", "drawByInfo "+(i+1)+"/"+childs.length+": "+childs[i].getPath());
					if (_wdlg[0].isCancel()) {
						break;
					}
/*DEL _PROGRESS_CYCLE*
					if ((i + 1) % _PROGRESS_CYCLE == 0) {
						_wdlg[0].setMessage(_values.waitTitle + (i+1)+ "/" + len);
					}
/*DEL _PROGRESS_CYCLE*/
					list.add(childs[i]);
				}
				draw(list);
			}
		}.start();
	}

	// リジューム用に縦位置を保存
	void savePosition() {
		RecyclerView rvList = findViewById(R.id.rvList);
		int len = rvList.getChildCount();
		if (len > 0) {
			View child = rvList.getChildAt(len - 1);
			int y = rvList.getChildAdapterPosition(child);
			// X は無効
			int x = rvList.computeHorizontalScrollOffset();
			_node.rowInfo.point = new Point(x, y);
		}
	}

	Semaphore _wsem = new Semaphore(0);

	WaitDialog[] _wdlg;

	// ページ表示
	 void draw(ArrayList<File> list) {
		int level = _node.getLevel();
		boolean isRoot = (level == 0);
        if (_wdlg == null) {
			_wdlg = new WaitDialog[] {
					createWaitingDialog(false),
			};
		} else {
			try {
				// 取得
				_wsem.acquire();
			} catch (InterruptedException e) {
				if (MainActivity._ERROR_LOG) {
					RLog.e("Dcs", e.getMessage(), e);
				}
			}
        }
		new Thread() {
			@Override
			public void run() {
			try {
				ImageButton ibRevert = findViewById(R.id.ibRevert);
				new AsyncTask<Void, Void, Boolean>() {
					@Override
					protected Boolean doInBackground(Void... params) {
						return true;
					}

					@Override
					protected void onPostExecute(Boolean result) {
						try {
							// Root の時は Revert を非表示
							ibRevert.setVisibility(isRoot ? View.INVISIBLE : View.VISIBLE);
						} catch (Exception e) {
							if (MainActivity._ERROR_LOG) {
								RLog.e("Dcs", e.getMessage(), e);
							}
						}
					}
				}.execute();
//
				ArrayList<RowInfo> ilist = new ArrayList<>();
				Dcs dcs = new Dcs();
				// １ノード（１ページ）生成
				createOneNode(dcs, ilist, list, _node, false);
				dcs.close();
//
				int i, size = ilist.size();
				if (size > _MAX_DISPLAY_COUNT) {
				// 表示ファイルが最大数を超えた時
					for (i = size - 1; i >= _MAX_DISPLAY_COUNT; i--) {
						ilist.remove(i);
					}
					new AsyncTask<Void, Void, Boolean>() {
						@Override
						protected Boolean doInBackground(Void... params) {
							return true;
						}

						@Override
						protected void onPostExecute(Boolean result) {
							try {
								AlertDialog.Builder builder = new AlertDialog.Builder(MainActivity.this)
								.setPositiveButton(_values.ok, null);
								AlertDialog dlg = builder.create();
                                // "表示ファイルが最大数を超えました(%d > %d)。";
								String msg = String.format(_values.maximumNumberOfDisplayFilesHasBeenExceeded, size, _MAX_DISPLAY_COUNT);
								dlg.setMessage(msg);
								synchronized (_wdlg[0]) {
									_wdlg[0].dismiss();
								}
								dlg.show();
							} catch (Exception e) {
								if (MainActivity._ERROR_LOG) {
									RLog.e("Dcs", e.getMessage(), e);
								}
							}
						}
					}.execute();
				}
				new AsyncTask<Void, Void, Boolean>() {
					@Override
					protected Boolean doInBackground(Void... params) {
						return true;
					}

					@Override
					protected void onPostExecute(Boolean result) {
						try {
							setSelectedCount(_node.getSelectedChildCount());
							RecyclerView rvList = findViewById(R.id.rvList);
							rvList.setLayoutManager(new LinearLayoutManager(MainActivity.this));
							MainAdapter<ArrayList<RowInfo>> adapter = new MainAdapter<>(ilist);
							_selectionControler.setAdapter(adapter);
						} catch (Exception e) {
							if (MainActivity._ERROR_LOG) {
								RLog.e("Dcs", e.getMessage(), e);
							}
						}
					}
				}.execute();
				// カレント・パス
				_prop.setProperty(_property_key_cpath, getCurrentPath());
			} catch (Exception e) {
				if (MainActivity._ERROR_LOG) {
					RLog.e("Dcs", e.getMessage(), e);
				}
			} finally {
				synchronized (_wdlg[0]) {
					_wdlg[0].dismiss();
				}
				_wsem.release();
			}
			}
		}.start();
	}

	FileFilter createFileFilter() {
		return new FileFilter() {
			@Override
			public boolean accept(File file) {
				return (file.isDirectory() || isMatch(file.getName()));
			}
		};
	}

	// １ノード（１ページ）生成
	void createOneNode(Dcs dcs, ArrayList<RowInfo> ilist, File pdir, TreeNode node, boolean isSelected) {
		// createMatchList によりディレクトリーを展開
		createOneNode(dcs, ilist, createMatchList(pdir), node, isSelected);
	}

	// １ノード（１ページ）生成
	void createOneNode(Dcs dcs, ArrayList<RowInfo> ilist, ArrayList<File> flist, TreeNode node, boolean isSelected) {
		try {
			RowInfo info;
			int[] atype = new int[1];
			int[] altype =
					(String.valueOf(true).equals(getProperty(_property_key_ln_judge)) ?
							new int[1] : null);
			// ディレクトリーの先読み用
			FileFilter filter = createFileFilter();
			int level = node.getLevel();
			boolean isRoot = (level == 0);
			File[] afile = dcsSort(flist, isRoot);
			boolean existsExternal = false;
			int i, len = afile.length;
			for (i = 0; i < len; i++) {
				if (_wdlg[0].isCancel()) {
					break;
				}
/*DEL _PROGRESS_CYCLE*
				if ((i + 1) % _PROGRESS_CYCLE == 0) {
					_wdlg[0].setMessage(_values.waitTitle + (i + 1) + "/" + len);
				}
/*DEL _PROGRESS_CYCLE*/
				File file = afile[i];
				boolean isNew = false;
				// ルート(level == 0)の時はパス
				TreeNode cnode = node.getChild(level == 0 ? file.getPath() : file.getName());
				if (cnode == null) {
					// 新規の時
					cnode = new TreeNode(new RowInfo());
					node.add(cnode);
					isNew = true;
				}
				if (isSelected
						|| _stepCount > 0
						&& cnode.rowInfo.clickTime < node.rowInfo.clickTime) {
					// 前進した時
					// かつ時間が古い時
					cnode.rowInfo.isSelected = node.rowInfo.isSelected;
					cnode.rowInfo.clickTime = node.rowInfo.clickTime;
					cnode.rowInfo.icon = R.drawable.dir_normal;
				}
				info = cnode.rowInfo;
				info.file = file;
				info.path = file.getPath();
				Locale locale = Locale.getDefault();
				SimpleDateFormat sdf = new SimpleDateFormat("yy/MM/dd HH:mm:ss", locale);
				info.date = sdf.format(file.lastModified());
//
				if (file.isDirectory()) {
				// ディレクトリーの時
					info.isDirectory = true;
					File[] childs = file.listFiles(filter);

					if (childs == null || childs.length == 0) {
					// 無効の時
						if (isNew) {
							info.icon = R.drawable.dir_empty;
						}
						info.isEmptyDirectory = true;
						info.size = (childs == null? "can't access": "0");
					} else {
					// 有効の時
						if (isNew) {
							info.icon = R.drawable.dir_normal;
						}
//!!						info.vcnt = getValidCount(childs);
						info.vcnt = childs.length;
						info.ecnt = childs.length;
//!!						info.size = String.format(locale, "%d/%d", info.vcnt, info.ecnt);
						info.size = String.format(locale, "%d", info.vcnt);
					}
					if (isRoot) {
						info.fileName = file.getPath();
						info.displayName = getRootName(file.getPath());
						existsExternal = info.displayName.equals(_values.externalStorage);
					} else {
						info.fileName = file.getName();
						if (info.fileName == null || info.fileName.length() == 0) {
							// 名前がない場合、パスとする
							info.fileName = file.getPath();
						}
						info.displayName = info.fileName;
					}
					info.typeName = "";
					info.ltype = "";
					info.mac = "";
					info.message = "";
				} else {
				// ディレクトリーではない時
					// ファイルと見なす(リンク、デバイス等を含む)
					info.fileName = info.displayName = file.getName();
					// アイコンをトグルで設定
					if (info.file.canRead()) {
						info.icon = icons[_ii++ & _ICON_SIZE_1];
						// RowInfo を更新
						MainActivity.judgesFile(dcs, atype, altype, info);
						info.message = "";
					} else {
						info.icon = R.drawable.file_blank;
						info.size = String.format(locale, "%3d", info.file.length());
						info.message = "Can't read";
					}
				}
				if (ilist != null) {
					ilist.add(info);
				}
			}
			if (ilist != null) {
				if (isRoot && !existsExternal) {
					info = new RowInfo();
					info.isDirectory = true;
					info.displayName = _values.nothingExternalStorage;
					info.icon = R.drawable.dir_empty;
					info.date = "";
					info.size = "";
					info.typeName = "";
					info.ltype = "";
					info.mac = "";
					info.message = "";
					ilist.add(info);
				}
			}
		} catch(Exception e) {
			if (MainActivity._ERROR_LOG) {
				RLog.e("Dcs", e.getMessage(), e);
			}
		}
	}

	// コード・タイプ判定
	static void judgesFile(Dcs dcs, int[] atype, int[] altype, RowInfo info) {
		// コード・タイプを判定
		String sstype = getContext().getProperty(_property_key_stype);
		int stype = SettingsActivity.nameToType(sstype);
		dcs.judgesFile(atype, altype, info.file.getPath(), stype);
		Locale locale = Locale.getDefault();
		info.type = atype[0];
		info.typeName = DcsStatus.typeToName(atype[0]);
		info.size = String.format(locale, "%3d", info.file.length());
		info.sts = 0;
		if (dcs.utfmacExists()) {
			// UTF8-MACの場合、ワーニング
			info.mac = "UTF8-MAC";
			info.sts |= RowInfo.STS_WUTF8_MAC;
		} else {
			info.mac = "";
		}
		if (altype == null) {
			info.ltype = "";
		} else {
			StringBuilder sb = new StringBuilder();
			if (altype[0] == Dcs.ECTYPE) {
				// 文字タイプを特定できない場合、ワーニング
				sb.append(" CR,LF-Unknown");
				info.sts |= RowInfo.STS_WNIDENT;
			} else {
			// 文字タイプを特定できた時
				int acnt = 0;
				if ((altype[0] & Dcs.LTYPE_LF) != 0) {
					sb.append(" LF,");
					acnt++;
				}
				if ((altype[0] & Dcs.LTYPE_CR) != 0) {
					sb.append(" CR,");
					acnt++;
				}
				if ((altype[0] & Dcs.LTYPE_CRLF) != 0) {
					sb.append(" CR/LF,");
					acnt++;
				}
				if (sb.length() == 0) {
				// CR, LF なしの時
					sb.append(" CR,LF-Nothing");
					info.sts |= RowInfo.STS_WNOLN;
				} else {
				// CR, LF 複数の時
					if (acnt != 1) {
						info.sts |= RowInfo.STS_WMIXEDLN;
					}
					/* 末尾の ',' を削除 */
					sb.setLength(sb.length() - 1);
				}
			}
			info.ltype = sb.toString();
		}
		if (info.file.canWrite()) {
			info.message = "";
		} else {
			info.message = "Can't write";
		}
	}

	// 有効件数取得
	int getValidCount(File[] childs) {
		int vcnt, i, len = childs.length;
		for(i=vcnt=0;i<len;i++) {
			if (childs[i].isDirectory()) {
			// ディレクトリーの時
				if (childs[i].list() != null) {
				// 無効ではない時
					vcnt++;
				}
			} else if (isMatch(childs[i].getName())) {
				// ファイルの場合、選択条件にマッチすること
				vcnt++;
			}
		}
		return vcnt;
	}

	// 抽出条件判定
	public boolean isMatch(String name) {
		int pos = name.lastIndexOf('.');
		if (pos < 0) {
			// サフィックスがない場合、タイプなしが既選択ならば可
			return (_isNoType);
		}
		// 大文字／小文字区別せず
		String suffix = name.substring(pos).toLowerCase();
		boolean isMatch = (_hSuffix.get(suffix) != null);
		// 有効サフィックスにマッチするか
		// 又は、無効サフィックスにマッチしなかった場合は真
		return(_isSelectTtype == isMatch);
	}

	// 明細ディレクトリーのクリックによるパスの先端生成
	public void addPathInfo(RowInfo info) {
		TextView tv = new TextView(this);
		tv.setTextSize(_PATH_TEXT_SIZE);
		tv.setFocusable(true);
		tv.setTextColor(MainActivity._text_highlight_color);
		tv.setTypeface(Typeface.DEFAULT_BOLD);
		// 既存の先端の色をデフォルト値に戻す
		LinearLayout llPath = findViewById(R.id.llPath);
		int ccnt = llPath.getChildCount();
		if (ccnt > 0) {
			TextView ctv = (TextView) llPath.getChildAt(ccnt - 1);
			ctv.setTextColor(Color.BLACK);
			ctv.setTypeface(Typeface.DEFAULT);
		}
		TreeNode cnode = _node.getChild(info.fileName);
		if (cnode == null) {
			// 子にない場合、パスの先端を追加
			cnode = new TreeNode(info);
			_node.add(cnode);
		}
		// クリック時に再表示するリスナーを登録
		tv.setOnClickListener(new View.OnClickListener() {
			@Override
			public void onClick(View view) {
				try {
					LinearLayout llPath = findViewById(R.id.llPath);
					int idx = llPath.indexOfChild(view);
					gotoWithIndex(idx);
				} catch(Exception e) {
					if (MainActivity._ERROR_LOG) {
						RLog.e("Dcs", e.getMessage(), e);
					}
				}
			}
		});
		// Root の時だけ絶対パス
		tv.setText(llPath.getChildCount() == 0 ?
			(getRootName(info.file.getPath())+" "): ("» "+info.file.getName()+" "));
		llPath.addView(tv);
		// 先端を表示
		HorizontalScrollView hsvPath = findViewById(R.id.hsvPath);
		hsvPath.post(new Runnable() {
			@Override
			public void run() {
				tv.setActivated(true);  //DEBUG 有効か？
				tv.requestFocus();	  //DEBUG 有効か？
				hsvPath.smoothScrollTo(llPath.getWidth(), 0);
			}
		});
		savePosition();
		_node = cnode;
		// １つ進んだ
		_stepCount = 1;
		// クリックされたディレクトリーを展開
		drawByInfo(info);
	}

	// パスからルート名を取得
	static String getRootName(String path) {
		// "/storage/emulated/0/"
		return((path + '/').matches(_INTERNAL_STORAGE_RKEY)?
				_values.internalStorage: _values.externalStorage);
	}

	public Point getPosition() {
		return _node.rowInfo.point;
	}

	public int getSelectedCount() {
		TextView tvCount = findViewById(R.id.tvCount);
		return Integer.parseInt(((String)tvCount.getText()).trim());
	}

	public void setSelectedCount(int count) {
		TextView tvCount = findViewById(R.id.tvCount);
		tvCount.setText(String.valueOf(count));
	}

	public void addHolder(MainAdapter.ViewHolder holder) {
		_selectionControler.addHolder(holder);
	}

	public void removeHolder(MainAdapter.ViewHolder holder) {
		_selectionControler.removeHolder(holder);
	}

	/* すべて選択 ON/OFF による更新 */
	@Override
	public void selectAll(ArrayList<RecyclerView.ViewHolder> list, boolean isChecked) {
		// ViewHolder を更新
		CheckBox cbLast = null;
		RecyclerView rvList = findViewById(R.id.rvList);
		int i, len = list.size();
		for(i=0;i<len;i++) {
			MainAdapter.ViewHolder holder = (MainAdapter.ViewHolder) list.get(i);
			RowInfo info = holder.getRowInfo();
			if (!info.isEmptyDirectory) {
				View parent = MainActivity.getParent(holder.tvInfo, R.id.llRow);
				// onCheckboxClicked と共通化すべき
				if (holder.rowInfo.icon == R.drawable.dir_half) {
				// 半選択の時
					holder.rowInfo.icon = R.drawable.dir_normal;
					RecyclerView.Adapter adapter = rvList.getAdapter();
					if (adapter == null) {
						if (MainActivity._ERROR_LOG) {
							RLog.e("Dcs", "selectAll: RecyclerView.getAdapter() == null");
						}
					} else {
						adapter.notifyItemChanged(rvList.indexOfChild(parent));
					}
				}
				CheckBox cbSelect = parent.findViewById(R.id.cbSelect);
				cbSelect.setChecked(isChecked);
				info.isSelected = isChecked;
				cbLast = cbSelect;
			}
		}
		// ノードの配下を更新
		long clickTime = System.nanoTime();
		len = _node.getChildCount();
		for(i=0;i<len;i++) {
			TreeNode cnode = _node.getChild(i);
			// チェック情報を更新
			cnode.rowInfo.isSelected = isChecked;
			cnode.rowInfo.clickTime = clickTime;
			// 孫は削除
			cnode.removeAll();
		}
		if (cbLast != null) {
			// 有効選択があった場合、親の選択状態を更新する
			cbLast.setChecked(!isChecked);
			cbLast.performClick();
		}
	}

	static class SortInfo {
		String key;
		File file;
		boolean isDirectory;

		SortInfo(String key, File file) {
			this.key = key;
			this.file = file;
			this.isDirectory = file.isDirectory();
		}
	}

	// Rootはフルパス、以外は名前
	// 大文字／小文字区列なし
	// "/" やディレクトリー配下削除
	File[] dcsSort(ArrayList<File> list, boolean isRoot) {
		try {
		int i, j, len = list.size();
		SortInfo[] sinfo = new SortInfo[len];
		for(i=j=0;i<len;i++) {
			if (_wdlg[0].isCancel()) {
				break;
			}
/*DEBUG*
ここは有効にならない
			if ((i + 1) % _PROGRESS_CYCLE == 0) {
				if (_wdlg[0].isShowing()) {
					_wdlg[0].setMessage(_values.waitTitle + ": 1 step " + (i + 1) + "/" + len);
				}
			}
/*DEBUG*/
			File file = list.get(i);
			String name = file.getName();
			boolean isDirectory = file.isDirectory();
			if (!(isDirectory || isMatch(name))) {
			// 対象外の時
				continue;
			}
			// Root の場合はパス、それ以外は名前を設定
			if (isRoot) {
				name = file.getPath();
			}
			// ディレクトリーの場合は末尾に区切り文字 '/' を付加（"/" は "//"になるが対象外）
			sinfo[j++] = new SortInfo(isDirectory? name+"/": name, file);
		}
		SortInfo[] tsinfo = new SortInfo[j];
		System.arraycopy(sinfo, 0, tsinfo, 0, j);
		sinfo = tsinfo;
		len = j;
		if (!isRoot) {
			Arrays.sort(sinfo, new Comparator<SortInfo>() {
				public int compare(SortInfo b, SortInfo n) {
					// ディレクトリーを先頭にする（デバック中は末尾）
					if (b.isDirectory && !n.isDirectory) {
						// directory > file
						return (_DEBUG_SORT_DIR? 1: -1);
					}
					if (!b.isDirectory && n.isDirectory) {
						// file < directory
						return (_DEBUG_SORT_DIR? -1: 1);
					}
					return b.key.compareTo(n.key);
				}
			});
		}
		int dcnt = 0;
		for(i=0;i<len;i++) {
			if (_wdlg[0].isCancel()) {
				break;
			}
/*DEBUG*
ここは有効にならない
			if ((i + 1) % _PROGRESS_CYCLE == 0) {
				if (_wdlg[0].isShowing()) {
					_wdlg[0].setMessage(_values.waitTitle + ": 2 step " + (i + 1) + "/" + len);
				}
			}
/*DEBUG*/
			if (sinfo[i] == null || !sinfo[i].isDirectory) {
			// ディレクトリー以外はチェックなし
				continue;
			}
			if (sinfo[i].key.equals("//")) {
				// ルート("//")の場合は対象外
				sinfo[i] = null;
				dcnt++;
				continue;
			}
			for(j=i+1;j<len;j++) {
				if (!sinfo[j].key.startsWith(sinfo[i].key)) {
					// 子のファイル・パスのでなければ抜ける
					break;
				}
				// 子のファイル・パスの場合、対象外とする
				sinfo[j] = null;
				dcnt++;
			}
		}
		// 削除したファイルを除く
		File[] nfile = new File[len - dcnt];
		for(i=j=0;i<len;i++) {
			if (sinfo[i] == null) {
				continue;
			}
			nfile[j++] = sinfo[i].file;
		}
		return nfile;
	} catch(Exception e) {
		if (MainActivity._ERROR_LOG) {
			RLog.e("Dcs", e.getMessage(), e);
		}
	}
		return null;
	}

	// id を持つ親を取得
	static View getParent(View view, int id) {
		View pview = view;
		while(pview.getId() != id) {
			pview = (View) pview.getParent();
		}
		return pview;
	}
	View.OnTouchListener _listener = new View.OnTouchListener() {
		@Override
		public boolean onTouch(View view, MotionEvent evt) {
			switch (evt.getAction()) {
				default:
					break;
				case MotionEvent.ACTION_DOWN:
					_scrollManager.isContains((RecyclerView) MainActivity.getParent(view, R.id.rvList), evt.getRawX(), evt.getRawY(), "initializing");
					break;
				case MotionEvent.ACTION_MOVE:
					_scrollManager.moveTo((RecyclerView) MainActivity.getParent(view, R.id.rvList), evt.getRawX(), evt.getRawY(), false, "initializing");
					break;
				case MotionEvent.ACTION_UP:
					_scrollManager._isInside = false;
					break;
				case MotionEvent.ACTION_CANCEL:
					break;
			}
			return false;
		}
	};

	void setListeners(View parent, MainAdapter.ViewHolder holder) {
		// コード変換と覗き見
		View[] aview = new View[1];
		aview[0] = parent.findViewById(R.id.tvInfo);
		GestureDetector gestureDetector = new GestureDetector(this, new GestureDetector.SimpleOnGestureListener() {
			@Override
			public boolean onDown(MotionEvent evt) {
				_scrollManager.isContains(_binding.rvList, evt.getRawX(), evt.getRawY(), "GestureDetector");
				return true;
			}

			@Override
			public boolean onScroll(MotionEvent evt1, MotionEvent evt2, float distanceX, float distanceY) {
				_scrollManager.moveTo(_binding.rvList, evt1.getRawX(), evt1.getRawY(), false, "GestureDetector");
				return true;
			}

			@Override
			public void onLongPress(MotionEvent evt) {
				// 長押しによるコード変換
				Semaphore sem = new Semaphore(0);
				holder.rowInfo.isMoved = false;
//
				new Thread() {
					@Override
					public void run() {
						try {
							try {
								// 取得（ファイル書き込み待ち）
								sem.acquire();
							} catch (InterruptedException e) {
								if (MainActivity._ERROR_LOG) {
									RLog.e("Dcs", e.getMessage(), e);
								}
							}
							// １項目を再描画
							redraw(parent, holder);
						} catch(Exception e) {
							if (MainActivity._ERROR_LOG) {
								RLog.e("Dcs", e.getMessage(), e);
							}
						}
					}
				}.start();
//
				View parent = MainActivity.getParent(aview[0], R.id.llRow);
				parent.setBackgroundColor(MainActivity.getSeasonColor(1.2F));
				new Thread() {
					public void run() {
						try {
							// １ファイルの	コード変換
							int[] astype = new int[1];
							int[] adtype = new int[1];
							DcsConvertObserver[] observer = new DcsConvertObserver[1];
							Dcs dcs = ConvertActivity.createDcs(astype, adtype, observer);
							dcs.convertFile(holder.rowInfo.path, astype[0], adtype[0], false, observer[0], observer[0], holder.rowInfo);
							if (holder.rowInfo.os != null) {
								// Out0putStream が設定されている場合、閉じる
								try {
									holder.rowInfo.os.close();
								} catch (IOException e) {
									if (MainActivity._ERROR_LOG) {
										RLog.e("Dcs", e.getMessage(), e);
									}
								}
								holder.rowInfo.os = null;
							}
							holder.rowInfo.file = new File(holder.rowInfo.path);
							dcs.close();
						} catch (DcsException e) {
							if (MainActivity._ERROR_LOG) {
								RLog.e("Dcs", e.getMessage(), e);
							}
						}
						// 解放
						sem.release();
					}
				}.start();
				super.onLongPress(evt);
			}

			@Override
			public boolean onDoubleTap(MotionEvent evt) {
				return true;
			}

			@Override
			public boolean onSingleTapUp(MotionEvent evt) {
				// 別アプリで非同期にされている可能性あり
				if (holder.rowInfo.isDirectory) {
				// ディレクトリーの時
					if (!holder.rowInfo.isEmptyDirectory) {
						if (!holder.rowInfo.isMoved) {
							// 移動ではない時
							// パスの先端を生成
							addPathInfo(holder.rowInfo);
						}
					}
				} else {
				// ファイルの時
					String sdtype = getProperty(MainActivity._property_key_dtype);
					// １項目を再描画
					int stype = redraw(parent, holder);
					// 覗き見
					peekText(findViewById(R.id.ibRoot), holder.rowInfo.path, stype, String.format(_values.pressAndHoldToConvert, sdtype));
				}
				return true;
			}
		});
		View.OnTouchListener onFileTouchListener = new View.OnTouchListener() {

			@Override
			public boolean onTouch(View view, MotionEvent evt) {
				return gestureDetector.onTouchEvent(evt);
			}
		};

		//チェックボックスcbSelectは含まない
		int[] aid = {
				R.id.llRow,
				R.id.rlRow,
				R.id.v,
				R.id.ivIcon,
				R.id.hsvRow,
				R.id.tvInfo,
		};
		WindowManager wm = (WindowManager)getSystemService(Context.WINDOW_SERVICE);
		Display disp = wm.getDefaultDisplay();
		Point rp = new Point();
		disp.getRealSize(rp);
		int[] sl = new int[2];

		int i, len = aid.length;
		for (i = 0; i < len; i++) {
			View view = parent.findViewById(aid[i]);
			int vid = view.getId();
			if (vid == R.id.hsvRow) {
				// HorizontalScrollView の最短幅（右端）
				// 画面座標に変換
				view.getLocationOnScreen(sl);
				// _HSV_ROWOFFSET: 正確な幅が取れない！
				view.setMinimumWidth(rp.x - (sl[0] == 0? _HSV_ROWOFFSET: sl[0]));
				view.setHorizontalScrollBarEnabled(false);
				view.setOnTouchListener(_listener);
			} else {
				if (vid == R.id.tvInfo) {
					view.getLocationOnScreen(sl);
					view.setMinimumWidth(rp.x - (sl[0] == 0? _HSV_ROWOFFSET: sl[0]));
				}
				// 別アプリで非同期にされている可能性あり
				view.setOnTouchListener(onFileTouchListener);
			}
		}
		// 選択チェック・ボックス
		CheckBox cbSelect = parent.findViewById(R.id.cbSelect);
		cbSelect.setOnClickListener(new View.OnClickListener() {
			@Override
			public void onClick(View view) {
				_selectionControler.onSelected((CheckBox) view);
				onCheckboxClicked(cbSelect);
			}
		});
	}

	// １項目を再描画
	int redraw(View parent, MainAdapter.ViewHolder holder) {
		Semaphore sem = new Semaphore(0);
		int[] atype = new int[1];
		int[] altype =
				(String.valueOf(true).equals(getProperty(_property_key_ln_judge)) ?
						new int[1] : null);
        RecyclerView rvList = findViewById(R.id.rvList);
        new Thread() {
            @Override
            public void run() {
                try {
                    // 再描画
                    parent.setBackgroundColor(getSeasonColor());
                    Locale locale = Locale.getDefault();
                    SimpleDateFormat sdf = new SimpleDateFormat("yy/MM/dd HH:mm:ss", locale);
                    holder.rowInfo.date = sdf.format(holder.rowInfo.file.lastModified());
//
                    Dcs dcs = new Dcs();
                    // RowInfo を更新
                    MainActivity.judgesFile(dcs, atype, altype, holder.rowInfo);
                    dcs.close();
//
                    RecyclerView.Adapter adapter = rvList.getAdapter();
                    if (adapter == null) {
                        if (MainActivity._ERROR_LOG) {
                            RLog.e("Dcs", "redraw: RecyclerView.getAdapter() == null");
                        }
                    } else {
                        new AsyncTask<Void, Void, Boolean>() {
                            @Override
                            protected Boolean doInBackground(Void... params) {
                                return true;
                            }

                            @Override
                            protected void onPostExecute(Boolean result) {
                                try {
									WindowManager wm = (WindowManager)getSystemService(Context.WINDOW_SERVICE);
									Display disp = wm.getDefaultDisplay();
									Point rp = new Point();
									disp.getRealSize(rp);
									int[] sl = new int[2];
//
									View hsvRow = parent.findViewById(R.id.hsvRow);
									// 画面座標に変換
									hsvRow.getLocationOnScreen(sl);
									// HorizontalScrollView の最短幅（右端）
									// 気休め
									hsvRow.setMinimumWidth(rp.x - (sl[0] == 0? _HSV_ROWOFFSET: sl[0]));
                                    // adapter.notifyItemChanged() は２ページ以降更新されない、バグ
                                    adapter.notifyItemRangeChanged(rvList.getChildAdapterPosition(parent), 1);
                                } catch(Exception e) {
                                    if (MainActivity._ERROR_LOG) {
                                        RLog.e("Dcs", e.getMessage(), e);
                                    }
                                }
                            }
                        }.execute();
                    }
					// 解放
                    sem.release();
                } catch (Exception e) {
                    if (MainActivity._ERROR_LOG) {
                        RLog.e("Dcs", e.getMessage(), e);
                    }
                }
            }
        }.start();
//
		try {
			// 取得
			sem.acquire();
		} catch (InterruptedException e) {
			if (MainActivity._ERROR_LOG) {
				RLog.e("Dcs", e.getMessage(), e);
			}
		}
		return atype[0];
	}

	static Snackbar _peekText;

	// 覗き見
	static void peekText(View view, String path, int type, String guidance) {
		try {
			if (guidance == null) {
				guidance = "";
			}
			int glen = guidance.length();

			SpannableStringBuilder spannable = new SpannableStringBuilder(guidance);
			if (glen > 0) {
			// ガイダンスがある時
				spannable.setSpan(
						new ForegroundColorBoldSpan(Color.BLUE),
						0, glen,
						Spannable.SPAN_EXCLUSIVE_EXCLUSIVE
				);
				spannable.setSpan(
						new StyleSpan(Typeface.BOLD),
						0, glen,
						Spannable.SPAN_EXCLUSIVE_EXCLUSIVE
				);
			}
			File file = new File(path);
			BufferedInputStream br = new BufferedInputStream(new FileInputStream(file));
			// ファイル・サイズ
			long flen = file.length();
			// 1024 バイト以下を読み込んで表示
			byte[] buf = new byte[Math.min((int)flen, 1024)];
			// 入力サイズ
			int rlen = br.read(buf);
			br.close();
			if (_peekText != null) {
				// 表示されていたら中断する
				_peekText.dismiss();
			}
			if (rlen <= 0) {
				//  "長押しで UTF8_BOM に変換",
				_peekText = Snackbar.make(view, spannable, Snackbar.LENGTH_SHORT);
			} else {
				Charset cs;
				switch(type) {
					case Dcs.JIS:
					case Dcs.JIS2:
						cs = Charset.forName("ISO-2022-JP");
						break;
					case Dcs.SJIS:
						cs = Charset.forName("MS932");
						break;
					case Dcs.EUC:
					case Dcs.EUC1:
					case Dcs.EUC2:
					case Dcs.EUC3:
						cs = Charset.forName("EUC-JP");
						break;
					case Dcs.UTF16BE:
					case Dcs.UTF16BE_BOM:
						cs = StandardCharsets.UTF_16BE;
						break;
					case Dcs.UTF16LE:
					case Dcs.UTF16LE_BOM:
						cs = StandardCharsets.UTF_16LE;
						break;
					case Dcs.UTF8:
					case Dcs.UTF8_BOM:
						cs = StandardCharsets.UTF_8;
						break;
					default:
						cs = StandardCharsets.US_ASCII;
						break;
				}
				String str = new String(buf, cs);
				if (rlen < flen) {
				// 全て読み込んでいない時
					int len = str.length();
					if (len >= 2) {
						// ２文字以上の場合、泣き別れを考慮して最後の文字を削除
						str = str.substring(0, len - 1);
					}
				}
				if (glen > 0) {
				// ガイダンスがある時
					spannable.append("\n\n");
				}
				spannable.append(str);
				//  "長押しで UTF8_BOM に変換",
				_peekText = Snackbar.make(view, spannable, Snackbar.LENGTH_LONG);
//
				View sview = _peekText.getView();
				TextView tv = (TextView) sview.findViewById(com.google.android.material.R.id.snackbar_text);
				// コード名の長さによる見出しの改行を考慮(+1)
				tv.setMaxLines(spannable.toString().split("\n|\r\n|\r").length + 1);
			}
			_peekText.setAction(MainActivity._values.close, new View.OnClickListener() {
				@Override
				public void onClick(View view) {
					_peekText.dismiss();
				}
			});
			_peekText.show();
		} catch(IOException e) {
			if (MainActivity._ERROR_LOG) {
				RLog.e("Dcs", e.getMessage(), e);
			}
		}
	}

	// チェックボックスのチェック処理
	public void onCheckboxClicked(CheckBox cbSelect) {
		View vp = getParent((View) cbSelect.getParent(), R.id.llRow);
		RecyclerView rvList = findViewById(R.id.rvList);
		MainAdapter.ViewHolder holder = (MainAdapter.ViewHolder) rvList.getChildViewHolder((View) vp);
		// selectAll と共通化すべき
		if (holder.rowInfo.icon == R.drawable.dir_half) {
		// 半選択の時
			holder.rowInfo.icon = R.drawable.dir_normal;
			RecyclerView.Adapter adapter = rvList.getAdapter();
			if (adapter == null) {
				if (MainActivity._ERROR_LOG) {
					RLog.e("Dcs", "onCheckboxClicked: RecyclerView.getAdapter() == null");
				}
			} else {
				adapter.notifyItemChanged(rvList.indexOfChild(vp));
			}
		}
		// チェックボックスの選択値を RowInfo に設定（isSelected は不可！）
		setCheckStatus(holder.rowInfo, cbSelect.isChecked());
	}

	// チェックボックスの選択値を RowInfo に設定（isSelected は不可！）
	void setCheckStatus(RowInfo rowInfo, boolean checked) {
		int i, len;
		rowInfo.isSelected = checked;
		rowInfo.clickTime = System.nanoTime();
		TreeNode pnode, cnode = _node.getChild(rowInfo.fileName);
		// 配下は全選択又は全未選択
		// 子をすべて削除
		cnode.removeAll();
		// 遡って半選択の判定を行なう
		pnode = _node;

loop_end:
		while(pnode != null) {
			int valid = 0, sel = 0;
			len = pnode.getChildCount();
			for(i=0;i<len;i++) {
				cnode = pnode.getChild(i);
				if (!cnode.rowInfo.isEmptyDirectory) {
				// 有効の時
					valid++;
					if (cnode.rowInfo.isSelected) {
					// 選択されている時
						sel++;
					}
					if (cnode.rowInfo.icon == R.drawable.dir_half) {
					// 半選択の時
						do {
							// Root まで半選択
							pnode.rowInfo.icon = R.drawable.dir_half;
							pnode.rowInfo.isSelected = true;
						} while((pnode = pnode.getParent()) != null);
						break loop_end;
					}
				}
			}
			if (sel == 0) {
			// 無選択の時
				pnode.rowInfo.icon = R.drawable.dir_normal;
				pnode.rowInfo.isSelected = false;
			} else if (sel == valid) {
			// 全選択の時
				pnode.rowInfo.icon = R.drawable.dir_normal;
				pnode.rowInfo.isSelected = true;
			} else {
			// 半選択の時
				pnode.rowInfo.icon = R.drawable.dir_half;
				pnode.rowInfo.isSelected = true;
			}
			pnode = pnode.getParent();
		}
	}

	// 選択リスト取得
	RowInfo[] getSelectedInfoList(WaitDialog[] wdlg) throws DcsException {
		_hNode = new Hashtable<>();
		Dcs dcs = new Dcs();
		TreeNode node = _node.getLevelParent(0);
		ArrayList<RowInfo> list = new ArrayList<>();
		int acnt = 0;
		while((node = node.getNext()) != null) {
			if (wdlg[0].isCancel()) {
				break;
			}
			if (node.rowInfo.isSelected) {
			// 選択されている時
				if (node.rowInfo.isDirectory) {
				// ディレクトリーの時
					if (!node.rowInfo.isEmptyDirectory
							&& node.getChildCount() == 0) {
						// 展開していない場合、展開する
						expand(dcs, node);
					}
				} else {
					// 選択されたファイルを抽出
					list.add(node.rowInfo);
					_hNode.put(node.rowInfo.path, node);
/*DEL _PROGRESS_CYCLE*
					if (++acnt % _PROGRESS_CYCLE == 0) {
						wdlg[0].setMessage(_values.waitTitle + acnt);
					}
/*DEL _PROGRESS_CYCLE*/
				}
			}
		}
		dcs.close();
		int len = list.size();
		RowInfo[] ainfo = list.toArray(new RowInfo[len]);
		Arrays.sort(ainfo, new Comparator<RowInfo>() {
			@Override
			public int compare(RowInfo prev, RowInfo next) {
				// 大文字／小文字区別せず
				return prev.path.compareToIgnoreCase(next.path);
			}
		});
		return ainfo;
	}

	// 選択リスト生成
	ArrayList<File> createMatchList(File pdir) {
		File[] childs = pdir.listFiles(new FileFilter() {
			@Override
			public boolean accept(File file) {
				return (file.isDirectory() || isMatch(file.getName()));
			}
		});
		ArrayList<File> list = new ArrayList<>();
		if (childs != null) {
			int i, len = childs.length;
			for(i=0;i<len;i++) {
				list.add(childs[i]);
			}
		}
		return list;
	}

	// ノード展開
	// getSelectedInfoList から呼ばれた後
	// 選択状態にしてあるファイルを抽出
	void expand(Dcs dcs, TreeNode node) {
		createOneNode(dcs, null, node.rowInfo.file, node, true);
		int i, len = node.getChildCount();
		for(i=0;i<len;i++) {
			if (_wdlg[0].isCancel()) {
				break;
			}
			TreeNode cnode = node.getChild(i);
			if (cnode.rowInfo.isDirectory
					&& cnode.getChildCount() > 0) {
				// 孫がある場合、子ノード展開
				expand(dcs, cnode);
			}
		}
	}

	String getProperty(String key) {
		return _prop.getProperty(key);
	}

	void setProperty(String key, String value) {
		_prop.setProperty(key, value);
	}

	void resetPropertys() {
		_prop.clear();
		File pdir = getFilesDir();
		File pfile = new File(pdir, _PROPERTIES_NAME);
		pfile.delete();
		loadProperties();
	}

	void loadProperties() {
		try {
			Properties prop = _prop;
			File pdir = getFilesDir();
			File pfile = new File(pdir, _PROPERTIES_NAME);
			pdir.mkdirs();

			if (!pfile.exists()) {
				setDefaultProperties();
				createTypeHash();
				return;
			}
			BufferedInputStream bis = new BufferedInputStream(new FileInputStream(pfile));
			prop.loadFromXML(bis);
			bis.close();
			createTypeHash();

		} catch (Exception e) {
if (MainActivity._ERROR_LOG) {
RLog.e("Dcs", e.getMessage(), e);
}
			setDefaultProperties();
			createTypeHash();
		}
	}

	void createTypeHash() {
		_isNoType = String.valueOf(true).equals(getProperty(_property_key_notype));
		_isSelectTtype = String.valueOf(true).equals(getProperty(_property_key_condttype));
		Hashtable<String, String> hSuffix = _hSuffix = new Hashtable<>();
		int i, len;
		try {
			String[] suffixes = SettingsActivity.toArrayTypeText(getProperty(_isSelectTtype? _property_key_ttypes: _property_key_btypes));
			len = suffixes.length;
			for(i=0;i<len;i++) {
				// 大文字／小文字区別せず
				hSuffix.put(suffixes[i].toLowerCase(), suffixes[i]);
			}
		} catch(IOException e) {
if (MainActivity._ERROR_LOG) {
RLog.e("Dcs", e.getMessage(), e);
}
			// タイプなしを含まず、バイナリーを除く hSuffixゼロ件
			_isNoType = false;
			_isSelectTtype = false;
			hSuffix.clear();
		}
	}

	void setDefaultProperties() {
		Properties prop = _prop;
//随時設定		prop.setProperty(_property_key_cpath, Environment.getDownloadCacheDirectory().getParent());
		prop.setProperty(_property_key_condttype, String.valueOf(true));
		prop.setProperty(_property_key_help_zoom, String.valueOf(100));
		prop.setProperty(_property_key_help_y, String.valueOf(0));
		prop.setProperty(_property_key_ttypes, SettingsActivity.editTypeText(SettingsActivity.TXT_SUFFIX, true));
		prop.setProperty(_property_key_btypes, SettingsActivity.editTypeText(SettingsActivity.BIN_SUFFIX, true));
		prop.setProperty(_property_key_stype, SettingsActivity.S_UNKNOWN);
		prop.setProperty(_property_key_dtype, SettingsActivity.S_UTF8_BOM);
		prop.setProperty(_property_key_ln_judge, String.valueOf(false));
		prop.setProperty(_property_key_ln_edit, String.valueOf(false));
		prop.setProperty(_property_key_ln_dest, SettingsActivity.S_DEFAULT_LN);
		prop.setProperty(_property_key_mac_judge, String.valueOf(false));
		prop.setProperty(_property_key_mac1char, String.valueOf(false));
		prop.setProperty(_property_key_overr_same, String.valueOf(false));
		prop.setProperty(_property_key_overr_similar, String.valueOf(false));
	}

	void saveProperties() throws IOException {
		Properties prop = _prop;

		// 改行をシステム値に統一
		String ln = System.getProperty("line.separator");
		if (ln == null) {
			ln = "\n";
		}
		Enumeration<Object> keys = prop.keys();
		while(keys.hasMoreElements()) {
			String key = (String) keys.nextElement();
			String str = prop.getProperty(key);
			str = str.replaceAll("(\\n|\\r\\n|\\r)", ln);
			prop.setProperty(key, str);
		}
		File pdir = getFilesDir();
		File pfile = new File(pdir, _PROPERTIES_NAME);
		BufferedOutputStream bos = new BufferedOutputStream(new FileOutputStream(pfile));
		prop.storeToXML(bos, getString(R.string.app_name)+ln+getString(R.string.copyright), "UTF-8");
		bos.close();
	}

	// リジューム用だったが、選択条件再現やファイル情報更新等、実現困難なため不使用
	String getCurrentPath() {
		String lsep = System.getProperty("file.separator");
		LinearLayout llPath = findViewById(R.id.llPath);
		int i, len = llPath.getChildCount();
		StringBuilder sb = new StringBuilder();
		for(i=0;i<len;i++) {
			TextView tv = (TextView) llPath.getChildAt(i);
			String name = (String) tv.getText();
			sb.append((i==0? name: name.substring(2)).trim()).append(lsep);
		}
		return sb.toString();
	}

	static int _seasonIndex = -1;

	static int getSeasonIndex() {
		// Zero origin なので春は 2 （３月）から始まる
		return (_seasonIndex >= 0? _seasonIndex:
			(_seasonIndex = (((Calendar.getInstance().get(Calendar.MONTH) + 10 ) / 3) % 4)));
	}

	static int getSeasonColor() {
		switch(getSeasonIndex()) {
			default:
			case S_SPRING:	// 春
                return Color.rgb(225, 250, 154);
			case S_SUMMER:	// 夏
                return Color.rgb(215, 219, 250);
			case S_FALL:	// 秋
                return Color.rgb(250, 216, 120);
			case S_WINTER:	// 冬
                return Color.rgb(232, 232, 236);
		}
	}

	static int getSeasonColor(float weight) {
		int color = getSeasonColor();
		int r = Math.min(0xff, Math.round(weight * ((color >> 16) & 0xff)));
		int g = Math.min(0xff, Math.round(weight * ((color >> 8) & 0xff)));
		int b = Math.min(0xff, Math.round(weight * (color & 0xff)));
		return ((0xff<<24)|(r << 16)|(g<<8)|b);
	}
}
