package rabbit.dcs;

import android.content.Context;
import android.graphics.Color;
import android.graphics.Point;
import android.text.Spannable;
import android.text.SpannableStringBuilder;
import android.text.style.ForegroundColorSpan;
import android.view.Display;
import android.view.LayoutInflater;
import android.view.View;
import android.view.ViewGroup;
import android.view.WindowManager;
import android.widget.CheckBox;
import android.widget.HorizontalScrollView;
import android.widget.ImageView;
import android.widget.TextView;

import androidx.annotation.NonNull;
import androidx.recyclerview.widget.RecyclerView;

import java.util.ArrayList;
import java.util.Locale;

public class MainAdapter<onCheckboxClicked>
        extends RecyclerView.Adapter<MainAdapter.ViewHolder>
        implements DcsAdapter {

    ArrayList<RowInfo> _rowInfos;

    boolean _isInit;

    public MainAdapter(ArrayList<RowInfo> rowInfos) {
        this._rowInfos = rowInfos;
    }

    static class ViewHolder extends RecyclerView.ViewHolder implements DcsViewHolder {
        RowInfo rowInfo;
        ImageView ivIcon;
        CheckBox cbSelect;
        TextView tvInfo;

        ViewHolder(View v) {
            super(v);
            ivIcon = v.findViewById(R.id.ivIcon);
            cbSelect = v.findViewById(R.id.cbSelect);
            tvInfo = v.findViewById(R.id.tvInfo);
//            tvInfo.setTypeface(Typeface.MONOSPACE, Typeface.BOLD);
        }

        @Override
        public RowInfo getRowInfo() {
            return rowInfo;
        }
    }

    @Override
    public void onViewRecycled (@NonNull ViewHolder holder) {
            MainActivity.getContext().removeHolder(holder);
    }

    // Create new views (invoked by the layout manager)
    @Override
    @NonNull
    public ViewHolder onCreateViewHolder(@NonNull ViewGroup parent, int viewType) {
        try {

        // create a new view
        View view = LayoutInflater.from(parent.getContext())
                .inflate(R.layout.row_main, parent, false);
        // set the view's size, margins, paddings and layout parameters
            return new ViewHolder(view);
        } catch(Exception e) {
            if (MainActivity._ERROR_LOG) {
                RLog.e("Dcs", e.getMessage(), e);
            }
        }
        return null;
    }

    // Replace the contents of a view (invoked by the layout manager)
    @Override
    public void onBindViewHolder(@NonNull ViewHolder holder, int position) {
        try {
        MainActivity mainActivity = MainActivity.getContext();
        // - get element from your dataset at this position
        // - replace the contents of the view with that element
        RecyclerView rvList = mainActivity.findViewById(R.id.rvList);
        int fgcolor = Color.rgb(0x10, 0x10, 0x10);
        int bgcolor = MainActivity.getSeasonColor();
        int bgcolor0 = MainActivity.getSeasonColor(1.1F);
        /*DEBUGここから背景色 1/2*/
        holder.tvInfo.setTextColor(fgcolor);

        holder.tvInfo.setBackgroundColor(bgcolor);
        /*DEBUGここまで背景色 1/2*/
        RowInfo info = _rowInfos.get(position);
        holder.rowInfo = info;

        holder.ivIcon.setImageResource(info.icon);
        holder.cbSelect.setChecked(info.isSelected);
        String strName = info.displayName;
        String strDate = info.date;
        String strSize = String.format("%3s", info.size);
        String strType = info.typeName;
        String strLType = info.ltype;
        String strMac = info.mac;
        String strMessage = info.message;
        SpannableStringBuilder ssb = new SpannableStringBuilder();
        int colorName = 0;
        boolean noListeners = false;
        boolean isStatusCheck = false;
        holder.cbSelect.setVisibility(View.VISIBLE);

        if (info.isDirectory) {
            if (info.fileName == null
             && MainActivity._values.nothingExternalStorage.equals(info.displayName)) {
            // SDカード未装着の時
                // リスナーなし
                noListeners = true;
//
                WindowManager wm = (WindowManager)mainActivity.getSystemService(Context.WINDOW_SERVICE);
                Display disp = wm.getDefaultDisplay();
                Point rp = new Point();
                disp.getRealSize(rp);
                int[] sl = new int[2];
                // MainActivity.setListeners と同値
                // HorizontalScrollView の最短幅（右端）
                View view = MainActivity.getParent(holder.tvInfo, R.id.hsvRow);
                // 画面座標に変換
                view.getLocationOnScreen(sl);
                // _HSV_ROWOFFSET: 正確な幅が取れない！
                view.setMinimumWidth(rp.x - (sl[0] == 0? MainActivity._HSV_ROWOFFSET: sl[0]));
                view.setHorizontalScrollBarEnabled(false);
//
                view = MainActivity.getParent(holder.tvInfo, R.id.tvInfo);
                view.getLocationOnScreen(sl);
                view.setMinimumWidth(rp.x - (sl[0] == 0? MainActivity._HSV_ROWOFFSET: sl[0]));
                view.setHorizontalScrollBarEnabled(false);
                // チェックボックスを非表示
                holder.cbSelect.setVisibility(View.INVISIBLE);
                colorName = 0xff0000ff;
                strSize = "";
                strType = "";
                info.isEmptyDirectory = true;
            } else if (info.isEmptyDirectory) {
            // 無効ディレクトリーの時
                strSize = "";
                strType = "";
                // チェックボックスを非表示
                holder.cbSelect.setVisibility(View.INVISIBLE);
            } else {
        	strSize = String.format("%3s", info.size);
                strType = "";
            }
        } else if (!info.file.canRead() || !info.file.canWrite()) {
        // ファイルの時、かつ（入力又は出力不可）の時
            // チェックボックスを非表示
            holder.cbSelect.setVisibility(View.INVISIBLE);
        } else {
        // ファイルの時、かつ入出力可の時
            // ワーニングの背景色を設定
            isStatusCheck = true;
        }
            ssb.append(strName).append('\n');
        int len = ssb.length();
        if (colorName != 0) {
            ssb.setSpan(
                    new ForegroundColorSpan(colorName),
                    0, // start
                    len, // end
                    Spannable.SPAN_EXCLUSIVE_EXCLUSIVE
            );
        }
        ssb.append(strDate).append(' ').append(strSize);
        ssb.setSpan(
                new RelativeSizeMonospaceSpan(0.8f),
                    len, // start
                    ssb.length(), // end
                    Spannable.SPAN_EXCLUSIVE_EXCLUSIVE
        );
        // ワーニングの背景色を設定
        MainActivity.setStatusColor(ssb, isStatusCheck? info.sts: RowInfo.STS_SUCCESS, strType, strLType, strMac, strMessage);
        holder.tvInfo.setText(ssb);
        View parent = (View) holder.ivIcon.getParent();
        View parentR = (View) holder.cbSelect.getParent();
        View parent2 = (View) holder.tvInfo.getParent();
        MainActivity.setParentBackground(parent, bgcolor0);
        HorizontalScrollView svRow = (HorizontalScrollView) MainActivity.getParent(parent2, R.id.hsvRow);
        svRow.setBackgroundColor(bgcolor);
        /*DEBUG*
        svRow.setBackgroundColor(0xff8000ff);
        parent.setBackgroundColor(0xffff0000);
        parentR.setBackgroundColor(0xff00ff00);
        parent2.setBackgroundColor(0xff00ffff);
        parent3.setBackgroundColor(0xffffff00);
        ((View)parent2.getParent()).setBackgroundColor(0xffff0080);
        /*DEBUGここから背景色 2/2*/
        parent.setBackgroundColor(bgcolor);
        parentR.setBackgroundColor(bgcolor);
        parent2.setBackgroundColor(bgcolor);
        MainActivity.setParentBackground(parent, bgcolor0);
        /*DEBUGここまで背景色 2/2*/
        mainActivity.addHolder(holder);
        if (!_isInit && position == 0) {
        // 先頭行の時
            _isInit = true;
            // 縦方向の位置を前回に合わせる
            rvList.post(new Runnable() {
                @Override
                public void run() {
                    Point p = mainActivity.getPosition();

                    if (p != null) {
                        rvList.scrollToPosition(p.y);
                    }
                }
            });
        }
        if (!noListeners) {
            View vp = MainActivity.getParent(parent, R.id.llRow);
            mainActivity.setListeners(vp, holder);
        }

        } catch(Exception e) {
            if (MainActivity._ERROR_LOG) {
                RLog.e("Dcs", e.getMessage(), e);
            }
        }
    }

    // Return the size of your dataset (invoked by the layout manager)
    @Override
    public int getItemCount() {
        return _rowInfos.size();
    }

    @Override
    public RowInfo getRowInfo(int index) {
        return _rowInfos.get(index);
    }
 }
