package rabbit.dcs;

import android.util.Log;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintWriter;
import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.util.Date;

public class RLog {

    static String _path;
    static PrintWriter _pw;

    static boolean _isLogPrint = false;

    static boolean _isFilePrint = true;

    public static boolean reopen(File dfile) {
        synchronized (_path) {
            try{
                _pw.close();
                BufferedReader br = new BufferedReader(new FileReader(new File(_path)));
                PrintWriter pw = _pw = new PrintWriter(new BufferedWriter(new FileWriter(dfile)));
                String str;
                while((str = br.readLine()) != null) {
                    pw.println(str);
                }
                pw.flush();
                br.close();
                new File(_path).delete();
                _path = dfile.getPath();
            } catch(Exception e) {
                return false;
            }
        }
        return true;
    }

    public void setLogPrint(boolean isLogPrint) {
        _isLogPrint = isLogPrint;
    }

    public void setFilePrint(boolean isFilePrint) {
        _isFilePrint = isFilePrint;
    }

    public RLog(String path) throws IOException {
        _path = path;
        File file = new File(path);
        if (file.exists()) {
            if (file.length() > 0) {
                String name = file.getName();
                int pos = name.lastIndexOf('.');
                String prefix = name.substring(0, pos);
                String suffix = name.substring(pos);

                SimpleDateFormat sdf = new SimpleDateFormat("yyMMdd_HHmmss");
                String ymd = sdf.format(new Date());

                String nname = String.format("%s_%s%s",
                        prefix, ymd, suffix);
                file.renameTo(new File(file.getParent(), nname));
            }
        }
        _pw = new PrintWriter(new BufferedWriter(new FileWriter(file)));
    }

    public static void e(String key, String value) {
        String str = getDateFormat(new Date())+": "+value;
        if (_isLogPrint) {
            Log.e(key, str);
        }
        if (_pw == null) {
            return;
        }
        if (_isFilePrint) {
            _pw.println(key+":e/"+str);
            _pw.flush();
        }
    }

    public static void e(String key, String value, Throwable e) {
        String str = getDateFormat(new Date())+": "+value;
        if (_isLogPrint) {
            Log.e(key, str, e);
        }
        if (_pw == null) {
            return;
        }
        if (_isFilePrint) {
            _pw.println(key+":e/"+str);
            StackTraceElement[] s = e.getStackTrace();
            int i, len = s.length;
            for(i=0;i<len;i++) {
                _pw.println(s[i].toString());
            }
            _pw.flush();
        }
    }

    public static void d(String key, String value) {
        String str = getDateFormat(new Date())+": "+value;
        if (_isLogPrint) {
            Log.d(key, str);
        }
        if (_pw == null) {
            return;
        }
        if (_isFilePrint) {
            _pw.println(key+":d/"+key+": "+str);
            _pw.flush();
        }
    }

    public static void d(String key, String value, Throwable e) {
        String str = getDateFormat(new Date())+": "+value;
        if (_isLogPrint) {
            Log.d(key, str, e);
        }
        if (_pw == null) {
            return;
        }
        if (_isFilePrint) {
            _pw.println(key+":d/"+key+": "+str);
            StackTraceElement[] s = e.getStackTrace();
            int i, len = s.length;
            for(i=0;i<len;i++) {
                _pw.println(s[i].toString());
            }
            _pw.flush();
        }
    }

    public static String getDateFormat(Date date) {
        String str = DateFormat.getDateTimeInstance().format(date);
        int index = str.indexOf(' ');
        if (index >= 0 && str.charAt(index+2) == ':') {
            index++;
            // 時刻の先頭に 0 を挿入
            str = str.substring(0,index)+'0'+str.substring(index);
        }
        return str;
    }
}
