package rabbit.dcs;

import android.content.Context;
import android.graphics.Point;
import android.util.DisplayMetrics;
import android.view.Display;
import android.view.View;
import android.view.WindowManager;
import android.webkit.WebView;

import androidx.recyclerview.widget.RecyclerView;

public class ScrollManager {

    float _bx, _by;
    float _cx, _cy;
    int _bpos;
    boolean _isInside;

    public ScrollManager() {
    }

    public void isContains(WebView wview, float x, float y, String dbg) {
        _isInside = false;
        MainActivity mainActivity = MainActivity.getContext();
        WindowManager wm = (WindowManager)mainActivity.getSystemService(Context.WINDOW_SERVICE);
//
        // 画面座標に変換
        int[] sl = new int[2];
        wview.getLocationOnScreen(sl);
        int vsw = wview.getVerticalScrollbarWidth();
        // 画面左端からの位置
        int rw = wview.getWidth() + sl[0];
        if (x < rw - vsw || x >= rw) {
            return;
        }
        // WebView の高さ
        int rh = wview.getHeight();
        DisplayMetrics dm = mainActivity.getResources().getDisplayMetrics();
        // 総明細の高さ
        int ah = (int)(dm.scaledDensity * wview.getContentHeight());
        if (ah - 3 <= rh) {
        // スクロール・バーが表示されない時(誤差：ah - 1 == rh)
            return;
        }
        // RecyclerView / Adapter
        float p = (float)rh / ah;
        //スクロール・バーの長さ
        float vsh = p * rh;
        //スクロール・バーの長さ誤差補正
        float ecv = (1 - p*p) * 170;
        int sy = wview.getScrollY();
        if (y < -ecv + sl[1] + p * sy || y >= ecv + sl[1] + p * sy + vsh) {
            return;
        }
        _isInside = true;
        _bx = _cx = x;
        _by = _cy = y;
        _bpos = sy;
    }

    public void moveTo(WebView wview, float x, float y, boolean isAdd, String dbg) {
        if (!_isInside) {
            return;
        }
        if (isAdd) {
            x += _cx;
            y += _cy;
        }
        MainActivity mainActivity = MainActivity.getContext();
        int rh = wview.getHeight();
        DisplayMetrics dm = mainActivity.getResources().getDisplayMetrics();
        // 総明細の高さ
        int ah = (int)(dm.scaledDensity * wview.getContentHeight());
        // RecyclerView / Adapter
        float p = (float)ah / rh;
        wview.setScrollY((int)Math.max(0, Math.min(ah - rh, p * (y - _by) + _bpos)));
        _cx = x;
        _cy = y;
    }

    public void isContains(RecyclerView wview, float x, float y, String dbg) {
        _isInside = false;
        int ccnt = wview.getChildCount();
        if (ccnt == 0) {
            return;
        }
        RecyclerView.Adapter adapter = wview.getAdapter();
        if (adapter == null) {
            if (MainActivity._ERROR_LOG) {
                RLog.e("Dcs", "isContains: RecyclerView.getAdapter() == null");
            }
            return;
        }
        MainActivity mainActivity = MainActivity.getContext();
        WindowManager wm = (WindowManager)mainActivity.getSystemService(Context.WINDOW_SERVICE);
        Display disp = wm.getDefaultDisplay();

        Point realSize = new Point();
        disp.getRealSize(realSize);
//
        // 画面座標に変換
        int[] sl = new int[2];
        wview.getLocationOnScreen(sl);
        int vsw = wview.getVerticalScrollbarWidth();
        // 画面左端からの位置
        int rw = wview.getWidth() + sl[0];
        if (x < rw - vsw || x >= rw) {
            return;
        }
        // RecyclerView の高さ
        int rh = wview.getHeight();
        View view = wview.getChildAt(0);
        int vh = view.getHeight();
        int icnt = adapter.getItemCount();
        // 総明細の高さ
        int ah = vh * icnt;
        if (ah <= rh) {
            // スクロール・バーが表示されない時
            return;
        }
        // RecyclerView / Adapter
        float p = (float)rh / ah;
        //スクロール・バーの長さ
        float vsh = p * rh;
        //スクロール・バーの長さ誤差補正
        float ecv = (1 - p*p) * 170;
        // 最上端表示のビューの全体の位置
        int pos = wview.getChildAdapterPosition(view);
        if (y < -ecv + sl[1] + p * vh * pos || y >= ecv + sl[1] + p * vh * pos + vsh) {
            return;
        }
        _isInside = true;
        _bx = _cx = x;
        _by = _cy = y;
        _bpos = pos;
    }

    public void moveTo(RecyclerView wview, float x, float y, boolean isAdd, String dbg) {
        if (!_isInside) {
            return;
        }
        int ccnt = wview.getChildCount();
        if (ccnt == 0) {
            return;
        }
        RecyclerView.Adapter adapter = wview.getAdapter();
        if (adapter == null) {
            if (MainActivity._ERROR_LOG) {
                RLog.e("Dcs", "moveTo: RecyclerView.getAdapter() == null");
            }
            return;
        }
        if (isAdd) {
            x += _cx;
            y += _cy;
        }
        int rh = wview.getHeight();
        // 最上端表示のビュー
        View view = wview.getChildAt(0);
        // 最上端表示のビューの全体の位置
        int vh = view.getHeight();
        int icnt = adapter.getItemCount();
        // 総明細の高さ
        int ah = vh * icnt;
        // RecyclerView / Adapter
        float p = (float)rh / ah;
        int cpos = Math.round((y - _by) / (p * vh));
        wview.scrollToPosition(Math.max(0, Math.min(icnt - 1, y > _by ? cpos + ccnt - 1 + _bpos: cpos + _bpos)));
        _cx = x;
        _cy = y;
    }
}
