package rabbit.dcs;

import android.view.View;
import android.widget.CheckBox;

import androidx.recyclerview.widget.RecyclerView;

import java.util.ArrayList;

/*
以下の View を操作する
R.id.rvList
R.id.cbSelectAll
R.id.bOk

setListener(CheckBox cbSelect) ですべての CheckBox を登録する
 */

public abstract class SelectionControler {

    public abstract int getSelectedCount();
    public abstract void setSelectedCount(int count);
    public abstract RecyclerView getRecyclerViewList();
    public abstract CheckBox getCheckBoxSelectAll();

    public void onSelected(CheckBox cbSelect) {

        int scnt = getSelectedCount();
        boolean isChecked = cbSelect.isChecked();
        if (isChecked) {
            scnt++;
        } else {
            scnt--;
        }
        setSelectedCount(scnt);
//
        RecyclerView rvList = getRecyclerViewList();
        DcsAdapter adapter = (DcsAdapter) rvList.getAdapter();
		if (adapter == null) {
if (MainActivity._ERROR_LOG) {
RLog.e("Dcs", "SelectionControler.onSelected: RecyclerView.getAdapter() == null");
}
		} else {
        	// 全ての項目を調べる
        	int vcnt, i, len = adapter.getItemCount();
        	for(i=vcnt=0;i<len;i++) {
            	RowInfo info = adapter.getRowInfo(i);
            	if (!info.isEmptyDirectory) {
            	// Row が有効の時
                	vcnt++;
            	}
        	}
        	CheckBox cbSelectAll = getCheckBoxSelectAll();
        	if (scnt == 0) {
        	// すべて未選択の時
            	cbSelectAll.setChecked(false);
        	} else if (scnt >= vcnt) {
        	// すべて選択している時
            	cbSelectAll.setChecked(true);
        	}
		}
    }

    ArrayList<RecyclerView.ViewHolder> _list;

    public void addHolder(RecyclerView.ViewHolder holder) {
        _list.add(holder);
    }

    public void removeHolder(RecyclerView.ViewHolder holder) {
        boolean flag;
        flag = _list.remove(holder);
        if (!flag) {
            if (MainActivity._ERROR_LOG) {
                DcsViewHolder dcsholder = (DcsViewHolder) holder;
                RLog.e("Dcs", "Holder not found! "+dcsholder.getRowInfo().path);
            }
        }
    }


    public void setAdapter(RecyclerView.Adapter adapter) {
        RowInfo info;
        if (_list != null) {
            _list.clear();
        }
        _list = new ArrayList<>();
        // 全ての項目を調べる
        CheckBox cbSelectAll = getCheckBoxSelectAll();
        RecyclerView rvList = getRecyclerViewList();
        rvList.setAdapter(adapter);
        int vcnt, scnt, i, len = adapter.getItemCount();
        for(i=vcnt=scnt=0;i<len;i++) {
            info = ((DcsAdapter)adapter).getRowInfo(i);
            if (!info.isEmptyDirectory) {
                vcnt++;
                if (info.isSelected) {
                    scnt++;
                }
            }
        }
        setSelectedCount(scnt);
        // すべて選択されている場合、チェック
        cbSelectAll.setChecked(scnt > 0 && scnt == vcnt);
    }

    SelectionOwner _owner;

    public SelectionControler(SelectionOwner owner) {
        _owner = owner;

        CheckBox cbSelectAll = getCheckBoxSelectAll();
        cbSelectAll.setOnClickListener(new View.OnClickListener() {
            @Override
            public void onClick(View view) {
                try {
                    setSelectedCount(0);
                    long clickTime = System.nanoTime();
                    boolean isChecked = cbSelectAll.isChecked();

                    RecyclerView rvList = getRecyclerViewList();
                    DcsAdapter adapter = (DcsAdapter) rvList.getAdapter();
                    if (adapter == null) {
                        if (MainActivity._ERROR_LOG) {
                            RLog.e("Dcs", "SelectionControler: cbSelectAll.onClick: RecyclerView.getAdapter() == null");
                        }
                    } else {
                        int i, scnt, len = adapter.getItemCount();
                        for (i = scnt = 0; i < len; i++) {
                            RowInfo info = adapter.getRowInfo(i);
                            if (!info.isEmptyDirectory) {
                                info.clickTime = clickTime;
                                info.isSelected = isChecked;
                                scnt++;
                            }
                        }
                        _owner.selectAll(_list, isChecked);
                        // 上記 CheckBox.performClick()
                        // で合計が更新されるので、ここで正しい値を設定する
                        setSelectedCount(isChecked ? scnt : 0);
                    }
                } catch(Exception e) {
                    if (MainActivity._ERROR_LOG) {
                        RLog.e("Dcs", e.getMessage(), e);
                    }
                }
            }
        });
    }
}
