package rabbit.dcs;

import android.app.AlertDialog;
import android.content.DialogInterface;
import android.content.Intent;
import android.content.pm.ActivityInfo;
import android.os.Bundle;
import android.view.GestureDetector;
import android.view.MotionEvent;
import android.view.View;
import android.widget.ArrayAdapter;
import android.widget.Button;
import android.widget.CheckBox;
import android.widget.RadioButton;
import android.widget.Spinner;
import android.widget.TextView;
import android.widget.Toast;

import androidx.appcompat.app.AppCompatActivity;

import java.io.IOException;
import java.io.StreamTokenizer;
import java.io.StringReader;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Comparator;

public class SettingsActivity extends AppCompatActivity {

    private static SettingsActivity _settingsActivity;

    private static final int TIP_LENGTH = 10;

    static final String TN_MAYBE			= "?";
    static final String TN_EMPTY			= "Empty";
    static final String TN_UNKNOWN		= "Unknown";
    static final String TN_ASCII			= "ascii";
    static final String TN_JIS				= "jis_cp50221";
    static final String TN_JIS2			= "jis_iso2022jp_2004";
    static final String TN_SJIS			= "sjis";
    static final String TN_EUC				= "euc";
    static final String TN_EUC1			= "euc_cp20932";
    static final String TN_EUC2			= "euc_cp51932";
    static final String TN_EUC3			= "euc_ms";
    static final String TN_UTF16BE		= "utf16be";
    static final String TN_UTF16BE_BOM	= "utf16be_bom";
    static final String TN_UTF16LE		= "utf16le";
    static final String TN_UTF16LE_BOM	= "utf16le_bom";
    static final String TN_UTF8			= "utf8";
    static final String TN_UTF8_BOM		= "utf8_bom";

    static final String S_UNKNOWN = TN_UNKNOWN;
    static final String S_JIS = TN_JIS.toUpperCase();
    static final String S_JIS2 = TN_JIS2.toUpperCase();
    static final String S_SJIS = TN_SJIS.toUpperCase();
    static final String S_EUC = TN_EUC.toUpperCase();
    static final String S_EUC1 = TN_EUC1.toUpperCase();
    static final String S_EUC2 = TN_EUC2.toUpperCase();
    static final String S_EUC3 = TN_EUC3.toUpperCase();
    static final String S_UTF16BE = TN_UTF16BE.toUpperCase();
    static final String S_UTF16BE_BOM = TN_UTF16BE_BOM.toUpperCase();
    static final String S_UTF16LE = TN_UTF16LE.toUpperCase();
    static final String S_UTF16LE_BOM = TN_UTF16LE_BOM.toUpperCase();
    static final String S_UTF8 = TN_UTF8.toUpperCase();
    static final String S_UTF8_BOM = TN_UTF8_BOM.toUpperCase();

    static final String S_LF = "LF";
    static final String S_CR = "CR";
    static final String S_CRLF = "CR/LF";
    static final String S_DEFAULT_LN = S_LF;

    static TypeNameInfo[] _info = {
            new TypeNameInfo(Dcs.JIS, Dcs.ND_JIS, S_JIS),
            new TypeNameInfo(Dcs.JIS2, Dcs.ND_JIS2, S_JIS2),
            new TypeNameInfo(Dcs.SJIS, Dcs.ND_SJIS, S_SJIS),
            new TypeNameInfo(Dcs.EUC, Dcs.ND_EUC, S_EUC),
            new TypeNameInfo(Dcs.EUC1, Dcs.ND_EUC1, S_EUC1),
            new TypeNameInfo(Dcs.EUC2, Dcs.ND_EUC2, S_EUC2),
            new TypeNameInfo(Dcs.EUC3, Dcs.ND_EUC3, S_EUC3),
            new TypeNameInfo(Dcs.UTF16BE, Dcs.ND_UTF16BE, S_UTF16BE),
            new TypeNameInfo(Dcs.UTF16BE_BOM, Dcs.ND_UTF16BE_BOM, S_UTF16BE_BOM),
            new TypeNameInfo(Dcs.UTF16LE, Dcs.ND_UTF16LE, S_UTF16LE),
            new TypeNameInfo(Dcs.UTF16LE_BOM, Dcs.ND_UTF16LE_BOM, S_UTF16LE_BOM),
            new TypeNameInfo(Dcs.UTF8, Dcs.ND_UTF8, S_UTF8),
            new TypeNameInfo(Dcs.UTF8_BOM, Dcs.ND_UTF8_BOM, S_UTF8_BOM),
    };

    static final String BIN_SUFFIX =
            ".au .avi .bin .bmp .class .cls .com .dat .dic .dll .exe"
                    +"\n.gif .gz .jar .jpeg .jpg .lzh .mid .midi .mp3 .mp4 .mpeg .mpg"
                    +"\n.o .obj .png .ram .so .tar .tgz .wav .xls .xlsm .xlsx .Z .zip";

    static final String TXT_SUFFIX =
            ".bak .bat .c .cgi .cpp .css .csv .doc .h .htm .html"
                    + "\n.ini .java .js .log .mak .old .org .tmp .txt";

    String[] _ttypes;
    String[] _btypes;

    public static SettingsActivity getContext() {
        return _settingsActivity;
    }

    public static int nameToType(String name) {
        TypeNameInfo[] info = _info;
        int i, len = info.length;
        for(i=0;i<len;i++) {
            if (info[i].name.equals(name)) {
                return info[i].type;
            }
        }
        return Dcs.UNKNOWN;
    }

    public static int nameToLtype(String sln) {
        return(S_LF.equals(sln)? Dcs.LTYPE_LF:
                S_CRLF.equals(sln)? Dcs.LTYPE_CRLF:
                        S_CR.equals(sln)? Dcs.LTYPE_CR: Dcs.LTYPE_NON);
    }

    @Override
    protected void onStart() {
        super.onStart();
    }

    @Override
    protected void onResume() {
        super.onResume();
    }

    @Override
    protected void onStop() {
        super.onStop();
    }

    @Override
    protected void onDestroy () {
        super.onDestroy();
//DEBUG        setRequestedOrientation(ActivityInfo.SCREEN_ORIENTATION_UNSPECIFIED);
    }

    @Override
    protected void onCreate(Bundle savedInstanceState) {
        super.onCreate(savedInstanceState);

		setRequestedOrientation(ActivityInfo.SCREEN_ORIENTATION_PORTRAIT);
        setContentView(R.layout.activity_settings);
//
        _settingsActivity = this;
        setProperties();
        setTitle(MainActivity._values.title);
        TextView tvGuidance = findViewById(R.id.tvGuidance);
        tvGuidance.setText(MainActivity._values.setJudgmentAndConversionConditions);
        if (MainActivity._values instanceof ValuesJp) {
            tvGuidance.setTextSize(16);
        } else {
            tvGuidance.setTextSize(18);
        }
        int bgcolor = MainActivity.getSeasonColor(MainActivity._W_BGCOLOR);
        View view = findViewById(R.id.activity_settings);
        view.setBackgroundColor(bgcolor);
        Button bOk = findViewById(R.id.bOk);
        Button bCancel = findViewById(R.id.bCancel);
        Button bReset = findViewById(R.id.bReset);
        Button bText = findViewById(R.id.bText);
        Button bBinary = findViewById(R.id.bBinary);
        MainActivity.setButtonColor(bOk);
        MainActivity.setButtonColor(bCancel);
        MainActivity.setButtonColor(bReset);
        MainActivity.setButtonColor(bText);
        MainActivity.setButtonColor(bBinary);
    }

    class DcsOnTouchListener implements View.OnTouchListener {
        GestureDetector _gestureDetector;
        DcsSimpleOnGestureAdapter _onGestureAdapter;

        DcsOnTouchListener(GestureDetector gestureDetector, DcsSimpleOnGestureAdapter onGestureAdapter) {
            _gestureDetector = gestureDetector;
            _onGestureAdapter= onGestureAdapter;
        }

        @Override
        public boolean onTouch(View view, MotionEvent evt) {
            if (evt.getAction() == MotionEvent.ACTION_DOWN) {
                _onGestureAdapter.notifyView(view);
            }
            return _gestureDetector.onTouchEvent(evt);
        }
    }

	Toast _toolTip;

    class DcsSimpleOnGestureAdapter extends GestureDetector.SimpleOnGestureListener {
        View _view;
        boolean _isText;
        boolean _isLongPressed;

        DcsSimpleOnGestureAdapter(boolean isText) {
            _isText = isText;
        }

        public boolean isLongPressed() {
            return _isLongPressed;
        }

        @Override
        public boolean onDown(MotionEvent e) {
            _isLongPressed = false;
            return true;
        }

        @Override
        public boolean onScroll(MotionEvent e1, MotionEvent e2, float distanceX, float distanceY) {
            return true;
        }

        @Override
        public void onLongPress(MotionEvent evt) {
			if (_toolTip != null) {
				// 表示されていたら中断する
				_toolTip.cancel();
			}
            _toolTip = Toast.makeText(SettingsActivity.this,
                    _isText? _ttypes[0]: _btypes[0], Toast.LENGTH_LONG);
			_toolTip.show();

            _isLongPressed = true;
            super.onLongPress(evt);
        }

        @Override
        public boolean onDoubleTap(MotionEvent evt) {
            return true;
        }

        @Override
        public boolean onSingleTapUp(MotionEvent evt) {
            if (_view instanceof RadioButton) {
                ((RadioButton) _view).setChecked(true);
            } else {
                _view.performClick();
            }
            checkEnabled(_view);
            return true;
        }

        public void notifyView(View view) {
            _view = view;
        }
    }

    void setProperties() {
        MainActivity mainActivity = MainActivity.getContext();

// 編集用ファイル・タイプ
        _ttypes = new String[] { mainActivity.getProperty(MainActivity._property_key_ttypes) };
        _btypes = new String[] { mainActivity.getProperty(MainActivity._property_key_btypes) };
// 有効タイプ（テキスト）選択
        RadioButton rbText = findViewById(R.id.rbIncluding);
// 無効タイプ（バイナリー）選択
        RadioButton rbBinary = findViewById(R.id.rbExcluding);
// 有効タイプ（テキスト）編集ボタン
        Button bText = findViewById(R.id.bText);
// 無効タイプ（バイナリー）編集ボタン
        Button bBinary = findViewById(R.id.bBinary);

        DcsSimpleOnGestureAdapter textDcsSimpleOnGestureAdapter = new DcsSimpleOnGestureAdapter(true);
        GestureDetector textGestureDetector = new GestureDetector(this, textDcsSimpleOnGestureAdapter);
        View.OnTouchListener textOnTouchListener = new DcsOnTouchListener(textGestureDetector, textDcsSimpleOnGestureAdapter);

        DcsSimpleOnGestureAdapter binaryDcsSimpleOnGestureAdapter = new DcsSimpleOnGestureAdapter(false);
        GestureDetector binaryGestureDetector = new GestureDetector(this, binaryDcsSimpleOnGestureAdapter);
        View.OnTouchListener binaryOnTouchListener = new DcsOnTouchListener(binaryGestureDetector, binaryDcsSimpleOnGestureAdapter);
        rbText.setOnTouchListener(textOnTouchListener);
        rbBinary.setOnTouchListener(binaryOnTouchListener);
        bText.setOnTouchListener(textOnTouchListener);
        bBinary.setOnTouchListener(binaryOnTouchListener);
//
        String[] astype = {
                S_UNKNOWN,
                S_JIS,
                S_JIS2,
                S_SJIS,
                S_EUC,
                S_EUC1,
                S_EUC2,
                S_EUC3,
                S_UTF16BE,
                S_UTF16LE,
                S_UTF8,
        };
        String[] adtype = {
                S_JIS,
                S_JIS2,
                S_SJIS,
                S_EUC,
                S_EUC1,
                S_EUC2,
                S_EUC3,
                S_UTF16BE,
                S_UTF16BE_BOM,
                S_UTF16LE,
                S_UTF16LE_BOM,
                S_UTF8,
                S_UTF8_BOM,
        };
// 入力タイプ選択
        ArrayAdapter<String> aaStype
                = new ArrayAdapter<>(this, android.R.layout.simple_spinner_item, astype);
        aaStype.setDropDownViewResource(android.R.layout.simple_spinner_dropdown_item);
        Spinner spStype = findViewById(R.id.spSType);
        spStype.setAdapter(aaStype);
        int pos = aaStype.getPosition(mainActivity.getProperty(MainActivity._property_key_stype));
        spStype.setSelection(Math.max(0, pos));
// 出力タイプ選択
        ArrayAdapter<String> aaDtype
                = new ArrayAdapter<>(this, android.R.layout.simple_spinner_item, adtype);
        aaDtype.setDropDownViewResource(android.R.layout.simple_spinner_dropdown_item);
        Spinner spDtype = findViewById(R.id.spDType);
        spDtype.setAdapter(aaDtype);
        pos = aaDtype.getPosition(mainActivity.getProperty(MainActivity._property_key_dtype));
        spDtype.setSelection(Math.max(0, pos));
// タイプなし
        boolean isNoType = String.valueOf(true).equals(mainActivity.getProperty(MainActivity._property_key_notype));
        CheckBox cbNoType = findViewById(R.id.cbNoType);
        cbNoType.setChecked(isNoType);
//
// 改行判定
        boolean isLnJudge = String.valueOf(true).equals(mainActivity.getProperty(MainActivity._property_key_ln_judge));
        CheckBox cbLnJudge = findViewById(R.id.cbLnJudge);
        cbLnJudge.setChecked(isLnJudge);
//
// 改行編集
        CheckBox cbLnEdit = findViewById(R.id.cbLnEdit);
// 改行文字選択
        String[] aln = {
                S_LF,
                S_CR,
                S_CRLF,
        };
        ArrayAdapter<String> aaLnDest
                = new ArrayAdapter<>(this, android.R.layout.simple_spinner_item, aln);
        aaLnDest.setDropDownViewResource(android.R.layout.simple_spinner_dropdown_item);
        Spinner spLnDest = findViewById(R.id.spLnDest);
        spLnDest.setAdapter(aaLnDest);
        pos = aaLnDest.getPosition(mainActivity.getProperty(MainActivity._property_key_ln_dest));
        spLnDest.setSelection(Math.max(0, pos));
// UTF8-MAC １文字変換
        boolean isMac1char = String.valueOf(true).equals(mainActivity.getProperty(MainActivity._property_key_mac1char));
        CheckBox cbMac1char = findViewById(R.id.cbMac1char);
        cbMac1char.setChecked(isMac1char);
//
// 同一コード上書き
        boolean isOverrSame = String.valueOf(true).equals(mainActivity.getProperty(MainActivity._property_key_overr_same));
        CheckBox cbOverrSame = findViewById(R.id.cbOverrSame);
        cbOverrSame.setChecked(isOverrSame);
// 類似コード上書き
        boolean isOverrSimilar = String.valueOf(true).equals(mainActivity.getProperty(MainActivity._property_key_overr_similar));
        TextView tvOverrSimilar = findViewById(R.id.tvOverrSimilar);
        String str = "JIS(CP50221, ISO-2022-JP-2004),\nEUC(Default, PC20932, PC51932, EUC-MS)";
        tvOverrSimilar.setTooltipText(str);
        CheckBox cbOverrSimilar = findViewById(R.id.cbOverrSimilar);
        cbOverrSimilar.setChecked(isOverrSimilar);
        cbOverrSimilar.setTooltipText(str);

// リスナー登録
// テキスト／バイナリー
        View.OnClickListener clickListener = new View.OnClickListener() {
            @Override
            public void onClick(View view) {
                checkEnabled(view);
            }
        };
        // テキスト／バイナリーのラジオ・ボタンの選択状態により
        // ボタンの有効／無効の同期をとる
        if (String.valueOf(true).equals( mainActivity.getProperty(MainActivity._property_key_condttype))) {
            // テキストのラジオ・ボタン選択の場合、テキストのボタン有効
            rbText.performClick();
        } else {
            // バイナリーのラジオ・ボタン選択の場合、バイナリーのボタン有効
            rbBinary.performClick();
        }
        bText.setOnClickListener(new View.OnClickListener() {
            @Override
            public void onClick(View view) {
                createEditTypes(true);
            }
        });
        bBinary.setOnClickListener(new View.OnClickListener() {
            @Override
            public void onClick(View view) {
                createEditTypes(false);
            }
        });
// 入出力タイプ（Spinner は AdapterView.OnItemSelectedListener ）
// 改行編集
        TextView tvLnEdit = findViewById(R.id.tvLnEdit);
        tvLnEdit.setOnClickListener(clickListener);
        cbLnEdit.setOnClickListener(clickListener);
        // 改行編集のチェック・ボックスの選択状態により
        // ボタンの有効／無効の同期をとる
        if (!String.valueOf(true).equals( mainActivity.getProperty(MainActivity._property_key_ln_edit))) {
            // 改行編集のチェック・ボックス未選択の場合、予め選択しておく
            cbLnEdit.setChecked(true);
        }
        // 改行編集のチェック・ボックス選択の場合、改行のスピンナーを有効にする
        cbLnEdit.performClick();
// 改行判定
        TextView tvLnJudge = findViewById(R.id.tvLnJudge);
        tvLnJudge.setOnClickListener(clickListener);
// UTF8-MAC １文字変換
        TextView tvMac1char = findViewById(R.id.tvMac1char);
        tvMac1char.setOnClickListener(clickListener);
        cbMac1char.setOnClickListener(clickListener);
// 同一コード上書き
        TextView tvOverrSame = findViewById(R.id.tvOverrSame);
        tvOverrSame.setOnClickListener(clickListener);
        cbOverrSame.setOnClickListener(clickListener);
// 類似コード上書き
        tvOverrSimilar.setOnClickListener(clickListener);
        cbOverrSimilar.setOnClickListener(clickListener);
// 確認／取消し／初期値に戻す
        Button bOk = findViewById(R.id.bOk);
        bOk.setOnClickListener(new View.OnClickListener() {
            @Override
            public void onClick(View view) {
                // 「確認」の場合、プロパティーを更新
                mainActivity.setProperty(MainActivity._property_key_condttype, String.valueOf(rbText.isChecked()));
                mainActivity.setProperty(MainActivity._property_key_ttypes, _ttypes[0]);
                mainActivity.setProperty(MainActivity._property_key_btypes, _btypes[0]);
                mainActivity.setProperty(MainActivity._property_key_notype, String.valueOf(cbNoType.isChecked()));

                mainActivity.setProperty(MainActivity._property_key_stype, (String) spStype.getSelectedItem());
                mainActivity.setProperty(MainActivity._property_key_dtype, (String) spDtype.getSelectedItem());
                mainActivity.setProperty(MainActivity._property_key_notype, String.valueOf(cbNoType.isChecked()));

                mainActivity.setProperty(MainActivity._property_key_ln_judge, String.valueOf(cbLnJudge.isChecked()));

                mainActivity.setProperty(MainActivity._property_key_ln_edit, String.valueOf(cbLnEdit.isChecked()));
                mainActivity.setProperty(MainActivity._property_key_ln_dest, (String) spLnDest.getSelectedItem());
                mainActivity.setProperty(MainActivity._property_key_mac1char, String.valueOf(cbMac1char.isChecked()));

                mainActivity.setProperty(MainActivity._property_key_overr_same, String.valueOf(cbOverrSame.isChecked()));
                mainActivity.setProperty(MainActivity._property_key_overr_similar, String.valueOf(cbOverrSimilar.isChecked()));
                try {
                    mainActivity.saveProperties();
                    mainActivity.loadProperties();
                } catch (IOException e) {
                    if (MainActivity._ERROR_LOG) {
                        RLog.e("Dcs", e.getMessage(), e);
                    }
                }
                finish();
            }
        });
        Button bCancel = findViewById(R.id.bCancel);
        bCancel.setOnClickListener(new View.OnClickListener() {
            @Override
            public void onClick(View view) {
                // 「取消し」の場合、何もしない
                finish();
            }
        });
        Button bReset = findViewById(R.id.bReset);
        bReset.setOnClickListener(new View.OnClickListener() {
            @Override
            public void onClick(View view) {
                new AlertDialog.Builder(getContext())
//                              "設定"
                        .setTitle(MainActivity._values.settings)
//                              "設定を初期値に戻しますか？"
                        .setMessage(MainActivity._values.doYouWantToSetTheDefaultValue)
//                              "確認"
                        .setPositiveButton(MainActivity._values.ok, new DialogInterface.OnClickListener() {
                            @Override
                            public void onClick(DialogInterface dialog, int which) {
                                MainActivity.getContext().resetPropertys();
                                setProperties();
                            }
                        })
//                              "取消し"
                        .setNegativeButton(MainActivity._values.cancel, null)
                        .show();
            }
        });
//
        TextView tv = findViewById(R.id.tvExtractionCondition);
        tv.setText(MainActivity._values.extractionCondition); // "抽出条件"
        rbText.setText(MainActivity._values.including); // "含む"
        rbBinary.setText(MainActivity._values.excluding); // "除く"
        bText.setText(MainActivity._values.text); // "テキスト　"
        bBinary.setText(MainActivity._values.binary); // "バイナリー"
        cbNoType.setText(MainActivity._values.noType); // "タイプなし"
        bOk.setText(MainActivity._values.ok); // "確認"
        bCancel.setText(MainActivity._values.cancel); // "取消し"
//
        tv = findViewById(R.id.tvInputType);
        tv.setText(MainActivity._values.inputType); // "入力タイプ"
        tv = findViewById(R.id.tvOutputType);
        tv.setText(MainActivity._values.outputType); // "出力タイプ"
//
        tvLnJudge.setText(MainActivity._values.linFeedJudgment); // "改行判定"
        bReset.setText(MainActivity._values.reset); // "初期値に戻す"
        tvLnEdit.setText(MainActivity._values.linFeedEditing); // "改行編集"
        tvMac1char.setText(MainActivity._values.utf8mac1char); // "UTF8-MAC １文字"
//
        tv = findViewById(R.id.tvValue2);
        tv.setText(MainActivity._values.value2b); // "上書き　"
        tvOverrSame.setText(MainActivity._values.sameCode); // "同じコード"
        tvOverrSimilar.setText(MainActivity._values.similarCode); // "類似コード"
        if (MainActivity._values instanceof ValuesJp) {
            rbText.setTextSize(12);
            rbBinary.setTextSize(12);
            tvMac1char.setTextSize(11);
        } else {
            rbText.setTextSize(14);
            rbBinary.setTextSize(14);
            tvMac1char.setTextSize(12);
        }
        checkEnabled(rbText);
    }

    void createEditTypes(boolean isText) {
        EditActivity.EditInfo info = new EditActivity.EditInfo();
        info.isText = isText;
        if (isText) {
//                "テキスト・タイプ編集";
            info.title = MainActivity._values.textTypesEditing;
//                "選択対象のテキスト・タイプを編集して下さい。";
            info.guidance = MainActivity._values.editTheSelectedTextTypes;
            info.value = MainActivity.getContext().getProperty(MainActivity._property_key_ttypes);
        } else {
//                "バイナリー・タイプ編集";
            info.title = MainActivity._values.binaryTypesEditing;
//                "除外するバイナリー・タイプを編集して下さい。";
            info.guidance = MainActivity._values.editBinaryTypesToExclude;
            info.value = MainActivity.getContext().getProperty(MainActivity._property_key_btypes);
        }
        Bundle bundle = new Bundle();
        bundle.putParcelable(MainActivity.KEY_EDIT, info);
        Intent intent = new Intent(SettingsActivity.this, EditActivity.class);
        intent.putExtras(bundle);
        startActivity(intent);
    }

    private void checkEnabled(View view) {
        boolean flag;
        RadioButton rb;
        Button tb, b;
        CheckBox cb;
        Spinner sp;

        int vid = view.getId();
        if (vid == R.id.rbIncluding || vid == R.id.rbExcluding) {
            rb = (RadioButton) view;
            // テキスト選択の場合は真
            flag = (vid == R.id.rbIncluding && rb.isChecked());
            b = findViewById(R.id.bText);
            tb = findViewById(R.id.bBinary);
            // テキストの可／不可
            b.setEnabled(flag);
            // バイナリーの可／不可
            tb.setEnabled(!flag);
            MainActivity.setButtonColor(b);
            MainActivity.setButtonColor(tb);
        } else if (vid == R.id.tvLnJudge) {
            cb = findViewById(R.id.cbLnJudge);
            cb.setChecked(!cb.isChecked());
        } else if (vid == R.id.tvMac1char) {
            cb = findViewById(R.id.cbMac1char);
            cb.setChecked(!cb.isChecked());
        } else if (vid == R.id.cbLnEdit || vid == R.id.tvLnEdit) {
            if (vid == R.id.tvLnEdit) {
                // TextView クリックの場合、CheckBox を反転する
                cb = findViewById(R.id.cbLnEdit);
                cb.setChecked(!cb.isChecked());
            } else {
                cb = (CheckBox) view;
            }
            sp = findViewById(R.id.spLnDest);
            sp.setEnabled(cb.isChecked());
        } else if (vid == R.id.cbOverrSame || vid == R.id.tvOverrSame) {
            if (vid == R.id.tvOverrSame) {
                // TextView クリックの場合、CheckBox を反転する
                cb = findViewById(R.id.cbOverrSame);
                cb.setChecked(!cb.isChecked());
//            } else {
//                cb = (CheckBox) view;
            }
        } else if (vid == R.id.cbOverrSimilar || vid == R.id.tvOverrSimilar) {
            if (vid == R.id.tvOverrSimilar) {
                // TextView クリックの場合、CheckBox を反転する
                cb = findViewById(R.id.cbOverrSimilar);
                cb.setChecked(!cb.isChecked());
//            } else {
//                cb = (CheckBox) view;
            }
        }
    }

    static String editTypeText(String text, boolean isLineBreak) {
        try {
            String[] astr = toArrayTypeText(text);
            StringBuilder sb = new StringBuilder();
            int cnt = 0;
            int i, len = astr.length;
            for(i=0;i<len;i++) {
                sb.append(astr[i]);
                if (isLineBreak && ++cnt % TIP_LENGTH == 0) {
                    sb.append('\n');
                } else {
                    sb.append(' ');
                }
                sb.append(' ');
            }
            if (isLineBreak && cnt % TIP_LENGTH != 0) {
                sb.setLength(sb.length() - 1);
                sb.append('\n');
            }
            return sb.toString();
        } catch (IOException e) {
			if (MainActivity._ERROR_LOG) {
				RLog.e("Dcs", e.getMessage(), e);
			}
        }
        return null;
    }

    // ".txt" 型の配列を生成
    static String[] toArrayTypeText(String text) throws IOException {
        StreamTokenizer tok = new StreamTokenizer(new StringReader(text));
        ArrayList<String> list = new ArrayList<>();
        while(tok.nextToken() != StreamTokenizer.TT_EOF) {
            if (tok.ttype != StreamTokenizer.TT_WORD) {
                continue;
            }
            String str = tok.sval.replaceFirst("^\\.*", ".");
            if (str.length() == 0) {
                continue;
            }
            list.add(str);
        }
        int i, j, len = list.size();
        String[] a = list.toArray(new String[len]);
        Arrays.sort(a, new Comparator<String>() {
            @Override
            public int compare(String o1, String o2) {
                return o1.compareToIgnoreCase(o2);
            }
        });
        // 重複値を削除
        int cnt = 0;
        String bstr= null;
        for(i=0;i<len;i++) {
            if (a[i].equals(bstr)) {
                // 直前値と等しい時
                a[i] = null;
                cnt++;
                continue;
            }
            bstr = a[i];
        }
        String[] b = new String[len - cnt];
        for(i=j=0;i<len;i++) {
            if (a[i] != null) {
                b[j++] = a[i];
            }
        }
        return b;
    }

    static String editPath(String path, String name) {
        String parent = path.substring(0, path.length() - name.length());
        return name + " - " + parent;
    }

    public void setTypes(String text, boolean isText) {
        // プロパティーの更新
        if (isText) {
            _ttypes[0] = editTypeText(text, true);
        }  else {
            _btypes[0] = editTypeText(text, true);
        }
    }
}
