package rabbit.dcs;

import java.io.File;
import java.io.FileFilter;
import java.util.ArrayList;

public class TreeNode {
    TreeNode _parent;
    RowInfo rowInfo;
    ArrayList<TreeNode> _childs = new ArrayList<>();

    public TreeNode(RowInfo rowInfo) {
        this.rowInfo = rowInfo;
    }

    public int getChildCount() {
        return _childs.size();
    }

    public TreeNode getChild(int index) {
        return _childs.get(index);
    }

    public TreeNode getChild(String name) {
        int i, len = _childs.size();
        for(i=0;i<len;i++) {
            TreeNode node = _childs.get(i);
            if (node.rowInfo.fileName.equals(name)) {
                return node;
            }
        }
        return null;
    }

    public int getSelectedChildCount() {
        int scnt, i, len = _childs.size();
        for(i=scnt=0;i<len;i++) {
            TreeNode node = _childs.get(i);
            if (!node.rowInfo.isEmptyDirectory
                    && node.rowInfo.isSelected) {
                scnt++;
            }
        }
        return scnt;
    }

    public int getChildIndex(TreeNode node) {
        return _childs.indexOf(node);
    }

    public TreeNode getFirstChild() {
        return _childs.get(0);
    }

    public TreeNode getLastChild() {
        return _childs.get(_childs.size() - 1);
    }

    public TreeNode getParent() {
        return _parent;
    }

    public void add(TreeNode node) {
        if (node._parent == this) {
            return;
        }
        _childs.add(node);
        if (node._parent != null) {
            node._parent.remove(node._parent.getChildIndex(node));
        }
        node._parent = this;
    }

    public void add(int index, TreeNode node) {
        if (node._parent == this) {
            return;
        }
        _childs.add(index, node);
        if (node._parent != null) {
            node._parent.remove(node._parent.getChildIndex(node));
        }
        node._parent = this;
    }

    public void set(int index, TreeNode node) {
        if (node._parent == this) {
            return;
        }
        _childs.set(index, node);
        if (node._parent != null) {
            node._parent.remove(node._parent.getChildIndex(node));
        }
        node._parent = this;
    }

    public boolean remove(TreeNode node) {
        if (node._parent == this) {
            node._parent = null;
            return _childs.remove(node);
        }
        return false;
    }

    public TreeNode remove(int index) {
        return _childs.remove(index);
    }

    public void removeAll() {
        _childs.clear();
    }

    public TreeNode getRoot() {
        TreeNode tnode, node = this;
        int cnt = 0;
        while ((tnode = node.getParent()) != null) {
            node = tnode;
        }
        return node;
    }

    public TreeNode getNext() {
        TreeNode tnode, node = this;
        if (_childs.size() > 0) {
            // 子がある時
            return _childs.get(0);
        }
        // 親以前の先祖をたどって、その子の弟を探す
        TreeNode ancestor = _parent;
        TreeNode child = this;
        for(;;) {
            if (ancestor == null) {
                // 先祖がない時
                return null;
            }
            int size = ancestor._childs.size();
            int index = ancestor.getChildIndex(child);
            if (index + 1 < size) {
                // 弟がある時
                return ancestor._childs.get(index + 1);
            }
            child = ancestor;
            ancestor = ancestor._parent;
        }
    }

    public int getLevel() {
        TreeNode node = this;
        int cnt = 0;
        while ((node = node.getParent()) != null) {
            cnt++;
        }
        return cnt;
    }

    public TreeNode getLevelParent(int index) {
        TreeNode node = this;
        int i, level = getLevel();
        for(;level>index; level--) {
            node = node.getParent();
        }
        return node;
    }

    static MainActivity _mainActivity;

    public int getFileCount() {
        _mainActivity = MainActivity.getContext();
        return getFileCount(0, rowInfo.file);
    }

    static int getFileCount(int scnt, File file) {
        File[] childs = file.listFiles(new FileFilter() {
            @Override
            public boolean accept(File file) {
                return (file.isDirectory() || _mainActivity.isMatch(file.getName()));
            }
        });
        if (childs == null) {
            return scnt;
        }
        int i, len = childs.length;
        for(i=0;i<len;i++) {
            if (childs[i].isDirectory()) {
                scnt = getFileCount(scnt, childs[i]);
            } else {
                scnt++;
            }
        }
        return scnt;
    }
}
