/*
 * _eucto_unile.c
 *
 * Copyright 2019-2023 Ichiji Tadokoro. All Rights Reserved.
 */

#include <ctype.h>
#include "_dcs.h"

int _eucto_unile(Dcs *dcs, bool is_bom) {
	int c, tsts, sts = DCS_SUCCESS;
	const unsigned char *tp, *sp, *ep;
	unsigned char *dbuf, *dp;
	int ltype = dcs->_type;
	ll_list *tlist, *clist = ll_first((ll_list *)dcs->_block);
	BufferInfo *tbinfo, *binfo = ll_get(clist);

	sp = binfo->buf + binfo->offset;
	ep = binfo->buf + binfo->length;
	reset_no_char(dcs, sp);
	dbuf = dp = _buffser_info.buf;
	if (is_bom) {
		*dp++ = UTF16_B1;
		*dp++ = UTF16_B0;
	}
	for(;;) {
		/* 残りが最大長に満たない時、次の入力バッファに移動 */
		check_ll_buf(dcs, sp, ep, 4, clist, tlist, binfo, tbinfo);
		/* 出力バッファ・サイズを超える時、ライター出力 */
		check_info_buf_break(dcs, dp, dbuf, sizeof(_buffser_info.buf) - _MAX_UNILE_OFFSET, sts);
		if (sp >= ep) {
			break;
		}
		tp = sp;
		if ((c = (*sp++ & 0xff)) >= 0xa1 || c == 0x8e) {
			if (sp >= ep) {
				sts = DCS_EUEXPEOD;
				break;
			}
			c = ((c<<8)|(*sp++ & 0xff));
			euc_to_unile_long_sp(dcs, dp, c, tp);
			if (!dcs_is_suppress(dcs) || !is_unile_no_char(dp)) {
				dp += 2;
			}
		} else if (c == 0x8f) {				/* EUC３バイト第１ﾊﾞｲﾄ */
			if (sp + 2 > ep) {
				sts = DCS_EUEXPEOD;
				break;
			}
			c = (0x8f0000 | char_to_short(sp));
			euc_to_unile_long_sp(dcs, dp, c, tp);
			if (!dcs_is_suppress(dcs) || !is_unile_no_char(dp)) {
				dp += 2;
			}
			sp += 2;
		} else {
			cnvln_stype_to_unile(dp, c, sp, ep, ltype, euc);
		}
	}
	if (sts == DCS_SUCCESS || sts == DCS_EUEXPEOD) {
		/* 出力バッファの残りをライター出力 */
		check_info_buf_rem(dcs, dp, dbuf, tsts);
		if (tsts != DCS_SUCCESS) {
			sts = tsts;
		}
	}
	DCS_RETURN(dcs, sts);
}
