/*
 * dcs_euc_strln.c
 *
 * Copyright 2019-2023 Ichiji Tadokoro. All Rights Reserved.
 */

#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>
#include "_dcs.h"

int dcs_euc_strln(Dcs *dcs, char *dbuf, size_t dsize, size_t *pdsize, char **sbuf, size_t ssize) {
	int sts = 0;
	int c, cnt = 0;
	char *ptr = NULL;
	const unsigned char *sp, *ep, *tbuf;

	if (ssize > dsize) {
		ssize = dsize;
	}
	*pdsize = 0;
	sp = tbuf = (unsigned char *)*sbuf;
	ep = (unsigned char *)*sbuf + ssize;
	while(sp < ep) {
		if ((c = (*sp++ & 0xff)) >= 0xa1) {
			if (isgraph((c & 0x7f)) != 0) {		/* EUC第１ﾊﾞｲﾄ */
				if (sp - tbuf >= ssize) {
//					sts = DCS_EUEXPEOD;
					sp--;
					break;
				}
				if ((c = (*sp++ & 0xff)) < 0xa1 || !isgraph((c & 0x7f))) {	/* EUC第２ﾊﾞｲﾄ */
					sp--;
					sts = DCS_ECHAR;
					break;
				}
				else {
					cnt++;
				}
			}
			else {
				sp--;
				sts = DCS_ECHAR;
				break;
			}
		}
		else if (c == 0x8e) {				/* EUCｶﾅ第１ﾊﾞｲﾄ */
			if (sp - tbuf >= ssize) {
//				sts = DCS_EUEXPEOD;
				sp--;
				break;
			}
			c = (*sp++ & 0xff);
			if (!is_sjis_kana(c)) {		/* 半角ｶﾅ */
				sp--;
				sts = DCS_ECHAR;
				break;
			}
			else {
				cnt++;
			}
		}
		else if (c == 0x8f) {				/* EUC３バイト第１ﾊﾞｲﾄ */
			if (sp - tbuf + 1 >= ssize) {
//				sts = DCS_EUEXPEOD;
				sp--;
				break;
			}
			sp += 2;
			cnt++;
		}
		else if (c == '\n') {
			ptr = (char *)sp - 1;
			break;
		}
		else if (c == '\r') {
			ptr = (char *)sp - 1;
			if (sp - tbuf < ssize && *sp == '\n') {
				sp++;
			}
			break;
		} else {
			cnt++;
		}
	}
	if (ptr == NULL) {
	/* 改行がなかった時 */
		ptr = (char *)sp;
	}
	memcpy(dbuf, tbuf, ptr - (char *)tbuf);
	*pdsize = ptr - (char *)tbuf;
	/* 次の行頭 */
	*sbuf = (char *)sp;
	if (sts < 0) {
		DCS_RETURN(dcs, sts);
	}

	DCS_RETURN_COUNT(dcs, sts, cnt);
}
