/*
 * dcs_jis_strln.c
 *
 * Copyright 2019-2023 Ichiji Tadokoro. All Rights Reserved.
 */

#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>
#include "_dcs.h"

#define _ESC_CHAR	0x1b
#define SO_CHAR		0x0e	/* JIS半角ｶﾅｼﾌﾄOUT */
#define SI_CHAR		0x0f	/* JIS半角ｶﾅｼﾌﾄIN */

int dcs_jis_strln(Dcs *dcs, char *dbuf, size_t dsize, size_t *pdsize, char **sbuf, size_t ssize, int *shtype) {
	int sts = 0;
	int c, cnt = 0, flag_shift = *shtype;
	char *ptr = NULL;
	const unsigned char *sp, *ep, *tbuf;

	if (ssize > dsize) {
		ssize = dsize;
	}
	*pdsize = 0;
	sp = tbuf = (unsigned char *)*sbuf;
	ep = (unsigned char *)*sbuf + ssize;
	while(sp  < ep) {
		if ((c = *sp++) == _ESC_CHAR) {	/* ｴｽｹｰﾌﾟ(0x1b) */
			if (sp - (unsigned char *)*sbuf + 2 > ssize) {
				goto loop_end;
			}
			switch(c = *sp++) {
			case '$':	/* JIS漢字ｼﾌﾄIN */
				switch(c = *sp++) {
				case 'J':	/* 78JIS??? */
				case '@':	/* 78JIS */
				case 'B':	/* 83JIS */
					flag_shift = STYPE_X0208_1983;
					break;
				case '(':	/* JIS X 0213 */
					if (sp + 1 > ep) {
						goto loop_end;
					} else {
						switch(c = *sp++) {
						case 'D': /* 補助漢字 */
							flag_shift = STYPE_X0212;
							break;
						case 'Q': /* 第1面 */
						case 'O': /* JIS X 0213:2000第1面 */
							set_jis2022jp200401_to(dcs);
							flag_shift = STYPE_X0213_2004;
							break;
						case 'P': /* 第2面 */
							set_jis2022jp200402_to(dcs);
							flag_shift = STYPE_X0213_2000_2;
							break;
						default:
							sp--;
							break;
						}
					}
					break;
				default:
					sp--;
					break;
				}
				break;
			case '(':	/* JIS漢字ｼﾌﾄOUT */
				switch(c = *sp++) {
				case 'J':	/* JISﾛｰﾏ字 */
				case 'H':	/* JISﾛｰﾏ字 */
				case 'B':	/* ASCII */
					flag_shift = STYPE_ASCII;
					break;
				case 'I':	/* JIS半角ｶﾅ */
					flag_shift = STYPE_KANA;
					break;
				default:
					sp--;
					break;
				}
				break;
			default:
				sp--;
				break;
			}
		} else { /* if ((c = *sp++) == _ESC_CHAR) */
			switch(flag_shift) {
			case STYPE_X0212:
			case STYPE_X0208_1983:
			case STYPE_X0213_2004:
			case STYPE_X0213_2000_2:
				for(;;) {
					if (c > 0x7e || isgraph(c) == 0) {
						sp--;
						sts = DCS_ECHAR;
						goto loop_end;
					}
					else if (sp - tbuf >= ssize) {
						sts = DCS_EUEXPEOD;
						goto loop_end;
					}
					else if ((c = *sp++) > 0x7e || isgraph(c) == 0) {
						sp--;
						sts = DCS_ECHAR;
						goto loop_end;
					} else {
						cnt++;
					}
					if (sp - tbuf >= ssize) {
						goto loop_end;
					} else if ((c = *sp++) == _ESC_CHAR) {
						sp--;
						break;
					}
				}
				break;
			case STYPE_KANA:
				for(;;) {
					if (is_jis_kana(c) == 0) {
						sp--;
						sts = DCS_ECHAR;
						goto loop_end;
					} else  {
						cnt++;
					}
					if (sp - tbuf >= ssize) {
						goto loop_end;
					}
					else if (*sp == _ESC_CHAR) {
						break;
					}
					c = *sp++;
				}
				break;
			case STYPE_ASCII:
			default:
				for (;;) {
					if (c > _JIS_ROMAN_MAX) {
						sp--;
						sts = DCS_ECHAR;
						goto loop_end;
					}
					if (c != SO_CHAR) {
						if (c == '\n') {
							ptr = (char *)sp - 1;
							goto loop_end;
						}
						else if (c == '\r') {
							ptr = (char *)sp - 1;
							if (sp - tbuf < ssize && *sp == '\n') {
								sp++;
							}
							goto loop_end;
						}
						cnt++;
					}
					else {		/* JIS半角ｶﾅｼﾌﾄOUT */
						for(;;) {
							if (sp - tbuf >= ssize) {
								goto loop_end;
							}
							else if ((c = *sp++) == SI_CHAR) {	/* JIS半角ｶﾅｼﾌﾄIN */
								break;
							}
							else if (is_jis_kana(c) == 0) {
								sp--;
								sts = DCS_ECHAR;
								goto loop_end;
							} else {
								cnt++;
							}
						}
					}
					if (sp - tbuf >= ssize) {
						goto loop_end;
					}
					else if (*sp == _ESC_CHAR) {
						break;
					}
					c = *sp++;
				}
				break;
			}
		} /* if ((c = *sp++) == _ESC_CHAR) */
	}
loop_end:

	if (ptr == NULL) {
	/* 改行がなかった時 */
		ptr = (char *)sp;
	}
	memcpy(dbuf, tbuf, ptr - (char *)tbuf);
	*pdsize = ptr - (char *)tbuf;
	/* 次の行頭 */
	*sbuf = (char *)sp;
	*shtype = flag_shift;
	if (sts < 0) {
		DCS_RETURN(dcs, sts);
	}
	DCS_RETURN_COUNT(dcs, sts, cnt);
}
