/*
 * dcs_sjis_strln.c
 *
 * Copyright 2019-2023 Ichiji Tadokoro. All Rights Reserved.
 */

#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include "_dcs.h"

int dcs_sjis_strln(Dcs *dcs, char *dbuf, size_t dsize, size_t *pdsize, char **sbuf, size_t ssize) {
	int sts = 0;
	int c, cnt = 0;
	char *ptr = NULL;
	const unsigned char *sp, *ep, *tbuf;

	if (ssize > dsize) {
		ssize = dsize;
	}
	*pdsize = 0;
	sp = tbuf = (unsigned char *)*sbuf;
	ep = (unsigned char *)*sbuf + ssize;
	while(sp < ep) {
		c = (*sp++ & 0xff);
		if ((c & 0x80) == 0) {
			if (c == '\n') {
				ptr = (char *)sp - 1;
				break;
			} else if (c == '\r') {
				ptr = (char *)sp - 1;
				if (sp - tbuf < ssize && *sp == '\n') {
					sp++;
				}
				break;
			}
			cnt++;
		} else if (is_sjis_kana(c)) {
			cnt++;
		} else if (sp - tbuf >= ssize) {
			/* 行の途中で切れている */
//			sts = DCS_EUEXPEOD;
			sp--;
			break;
		} else {
			c = ((c << 8)|(*sp++ & 0xff));
			if (is_sjis_short(dcs, c)) {
				cnt++;
			} else {
				sp--;
				sts = DCS_ECHAR;
				break;
			}
		}
	}
	if (ptr == NULL) {
	/* 改行がなかった時 */
		ptr = (char *)sp;
	}
	memcpy(dbuf, tbuf, ptr - (char *)tbuf);
	*pdsize = ptr - (char *)tbuf;
	/* 次の行頭 */
	*sbuf = (char *)sp;
	if (sts < 0) {
		DCS_RETURN(dcs, sts);
	}
	DCS_RETURN_COUNT(dcs, sts, cnt);
}
