/*
 * dcs_utf8_strln.c
 *
 * Copyright 2019-2023 Ichiji Tadokoro. All Rights Reserved.
 */

#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include "_dcs.h"

int dcs_utf8_strln(Dcs *dcs, char *dbuf, size_t dsize, size_t *pdsize, char **sbuf, size_t ssize) {
	int cnt = 0, sts = 0;
	int c, c2, c3, c4;
	char *ptr = NULL;
	const unsigned char *sp, *ep, *tbuf;

	if (ssize > dsize) {
		ssize = dsize;
	}
	*pdsize = 0;
	tbuf = (unsigned char *)*sbuf;
	if ((*tbuf & 0xff) == UTF8_B0
	 && (*(tbuf + 1) & 0xff) == UTF8_B1
	 && (*(tbuf + 2) & 0xff) == UTF8_B2) {
		sp = tbuf + 3;
	} else {
		sp = tbuf;
	}
	ep = (unsigned char *)*sbuf + ssize;
	while(sp < ep) {
		c = *sp++;
		if ((c & 0x80) == 0) {
		/* 0xxxxxxx: １バイト文字 */
			if (c == '\n') {
				ptr = (char *)sp - 1;
				break;
			}
			else if (c == '\r') {
				ptr = (char *)sp - 1;
				if (sp - tbuf < ssize && *sp == '\n') {
					sp++;
				}
				break;
			}
			cnt++;
		} else if (sp - tbuf >= ssize) {
//			sts = DCS_EUEXPEOD;
			sp--;
			break;
		} else {
			c2 = *sp++;
			if ((c & 0xe0) == 0xc0) {
				if ((c2 & 0xc0) == 0x80) {
				/* 110xxxxx 10xxxxxx: ２バイト文字 */
				} else {
					sts = DCS_ECHAR;
					break;
				}
			} else {
				if (sp - tbuf >= ssize) {
//					sts = DCS_EUEXPEOD;
//					sp -= 2;
					break;
				}
				c3 = *sp++;
				if ((c & 0xf0) == 0xe0 && (c2 & 0xc0) == 0x80 && (c3 & 0xc0) == 0x80) {
				/* 1110xxxx 10xxxxxx 10xxxxxx: ３バイト文字 */
				} else {
					if (sp - tbuf >= ssize) {
	//					sts = DCS_EUEXPEOD;
//						sp -= 3;
						break;
					}
					c4 = *sp++;
					if ((c & 0xf8) == 0xf0 && (c2 & 0xc0) == 0x80 && (c3 & 0xc0) == 0x80 && (c4 & 0xc0) == 0x80) {
					/* 11110xxx 10xxxxxx 10xxxxxx 10xxxxxx: ４バイト文字 */
					} else {
						sts = DCS_ECHAR;
						break;
					}
				}
			}
			cnt++;
		}
	}
	if (ptr == NULL) {
	/* 改行がなかった時 */
		ptr = (char *)sp;
	}
	memcpy(dbuf, tbuf, ptr - (char *)tbuf);
	*pdsize = ptr - (char *)tbuf;
	/* 次の行頭 */
	*sbuf = (char *)sp;
	if (sts < 0) {
		DCS_RETURN(dcs, sts);
	}
	DCS_RETURN_COUNT(dcs, sts, cnt);
}
