/*
 * jdcs.h
 *
 * Copyright 2019-2023 Ichiji Tadokoro. All Rights Reserved.
 */

#ifndef JDCS_H_
#define JDCS_H_

#ifdef __cplusplus
extern "C" {
#endif

#include "dcs.h"

#define JDCS_VERSION            ((float)2.11)

/* STATUS NUMBERS */
#define JDCS_SUCCESS		0 /* 処理成功 */
#define JDCS_EOF			-1 /* 入力ファイルの終わり */
#define JDCS_ENOTEXISTS		-2 /* ファイルは存在せず */
#define JDCS_ENOTFILE		-3 /* 通常ファイルではない */
#define JDCS_EALLOC			-4 /* メモリー・アロケート・エラー */
#define JDCS_EROPEN			-5 /* 入力ファイル・オープン・エラー */
#define JDCS_EWOPEN			-6 /* 出力ファイル・オープン・エラー */
#define JDCS_EREAD			-7 /* 入力エラー */
#define JDCS_EWRITE			-8 /* 出力エラー */

#define JDCS_EIOMODE			-10 /* 入出力モード誤り */
#define JDCS_EIOFDRANGE		-11 /* ファイル記述子範囲誤り */
#define JDCS_EIONOTOPEN		-12 /* 未オープン・ファイルへのアクセス */
#define JDCS_EIOMAX			-13 /* 上限を超えたファイル・オープン */

#define JDCS_ECTYPE			-20 /* コード・タイプ誤り */
#define JDCS_EBOM			-21 /* BOM 不正（未使用） */
#define JDCS_EUEXPEOD		-22 /* 予期せぬデータの終わり */
#define JDCS_ECHAR			-23 /* コード値エラー */
#define JDCS_ELTYPE			-24 /* 改行コード・タイプ誤り */

#define JDCS_ECANOBS		-30 /* 監視者による取り消し */

#define JDCS_EJNI_ARRAY_LENGTH		-50
#define JDCS_EJNI_FIND_CLASS		-51
#define JDCS_EJNI_NEW_BYTE_ARRAY	-52
#define JDCS_EJNI_NEW_OBJECT_ARRAY	-53
#define JDCS_EJNI_ALLOC_OBJECT		-54
#define JDCS_EJNI_NEW_GLOBAL_REF	-55

/* コード判定ステータス */
#define JDCS_UNKNOWN			-1 /* コード・タイプ不明 */
#define JDCS_EMPTY			0 /* 空データ */
#define JDCS_ASCII			1 /* ASCII */
#define JDCS_JIS				2 /* JIS CP50221 */
#define JDCS_JIS2			3 /* JIS ISO-2022-JP-2004 */
#define JDCS_SJIS			4 /* SJIS */
#define JDCS_EUC				5 /* EUC-JP */
#define JDCS_EUC1			6 /* EUC CP20932 */
#define JDCS_EUC2			7 /* EUC CP51932 */
#define JDCS_EUC3			8  /* eucJP-ms */
#define JDCS_UTF16BE			11 /* UTF-16 ビッグ・エンディアン */
#define JDCS_UTF16BE_BOM		12 /* UTF-16 ビッグ・エンディアン（BOM付）*/
#define JDCS_UTF16LE			13 /* UTF-16 リトル・エンディアン・ */
#define JDCS_UTF16LE_BOM		14 /* UTF-16 リトル・エンディアン（BOM付）*/
#define JDCS_UTF8			15 /* UTF-8 */
#define JDCS_UTF8_BOM		16 /* UTF-8（BOM付）*/
#define JDCS_NORMAL			00000000077	/* 確定コード・タイプの最大値 */

/* 確定コード・タイプ */
#define JDCS_CONFIRMED_CODE_TYPE(sts)	(JDCS_UNKNOWN < (sts) && (sts) <= JDCS_NORMAL)

/* 未確定コード・タイプ（マスク値） */
#define JDCS_ND_ASCII		00000000100 /* ASCII */
#define JDCS_ND_JIS			00000000200 /* JIS CP50221 */
#define JDCS_ND_JIS2			00000000400 /* JIS ISO-2022-JP-2004 */
#define JDCS_ND_JIS_ALL		00000000600 /* JIS ALL */
#define JDCS_ND_SJIS			00000001000 /* SJIS */
#define JDCS_ND_EUC			00000002000 /* EUC-JP */
#define JDCS_ND_EUC1			00000004000 /* EUC CP20932 */
#define JDCS_ND_EUC2			00000010000 /* EUC CP51932 */
#define JDCS_ND_EUC3			00000020000 /* eucJP-ms */
#define JDCS_ND_EUC_ALL		00000036000 /* EUC ALL */
#define JDCS_ND_UTF16BE		00000200000 /* UTF-16 ビッグ・エンディアン */
#define JDCS_ND_UTF16BE_BOM	00000400000 /* UTF-16 ビッグ・エンディアン（BOM付）*/
#define JDCS_ND_UTF16LE		00001000000 /* UTF-16 リトル・エンディアン・ */
#define JDCS_ND_UTF16LE_BOM	00002000000 /* UTF-16 リトル・エンディアン（BOM付）*/
#define JDCS_ND_UTF8			00004000000 /* UTF-8 */
#define JDCS_ND_UTF8_BOM		00010000000 /* UTF-8（BOM付）*/
#define JDCS_ND_UTF_BOM		00012400000 /* UTF（BOM付）*/
#define JDCS_ND_ALL			00027777700 /* ND ALL */

/* 変換監視継続、停止ステータス */
#define JDCS_OBS_STOP			-1		/* 変換監視停止 */
#define JDCS_OBS_CONTINUE		0		/* 変換監視継続 */

/* Charset 定義ビット(_unibe_def_table.h マスク値) */
/* UTF 以外は未実装 */
#define JCSB_UTF16				0000001
#define JCSB_UTF8				0000002
#define JCSB_UTF				0000003
#define JCSB_JIS				0000010
#define JCSB_SJIS				0000020
#define JCSB_EUC				0000040
#define JCSB_EUC1				0000100
#define JCSB_EUC2				0000200
#define JCSB_EUC3				0000400
#define JCSB_OTHER				0000770
#define JCSB_ALL				0000777
#define JCSB_EX					0001000
#define JCSB_EX2				0002000
#define JCSB_UTFEX			(JCSB_EX | JCSB_UTF)
#define JCSB_UTFEX2		(CSB_EX2 | CSB_UTF)

#ifdef __cplusplus
}
#endif

#endif /* #ifndef JDCS_H_ */
