/*
 * jdcs_convert_file.c
 *
 * Copyright 2019-2023 Ichiji Tadokoro. All Rights Reserved.
 */

#include "_jdcs.h"

extern JniConvertInfo *_jni_info;

/* 変換監視 */
extern int _jdcs_observer(int sts, void *user_info);

/* 変換出力 */
extern int _jdcs_writer(const char *buf, size_t size, void *user_info);

/*
 * Class:     rabbit_dcs_Dcs
 * Method:    convertFile
 * Signature: ([BIIZLrabbit/dcs/ConvertObserver;Lrabbit/dcs/ConvertWriter;Ljava/lang/Object;)I
 */
JNIEXPORT jint JNICALL Java_rabbit_dcs_Dcs_convertFile
  (JNIEnv *jenv, jobject jobj, jbyteArray jpath,
	jint jstype, jint jdtype, jboolean jisForced, jobject jobserver, jobject jwriter, jobject user_info) {
	_JDcsInfo *info;
	int sts, len;
	jclass obcls, wrcls;
	JniConvertInfo *jni_info;
	char path[PATH_MAX];

	GET_DCS_INFO_BY_OBJECT(jenv, jobj, info);
	if ((info->dcs->_jni_info = jni_info = _jni_info = calloc(sizeof(JniConvertInfo), 1)) == NULL) {
		sts = DCS_EALLOC;
	} else {

		obcls = (*jenv)->GetObjectClass(jenv, jobserver);
		wrcls = (*jenv)->GetObjectClass(jenv, jwriter);
		jni_info->mid_ob = (*jenv)->GetMethodID(jenv, obcls, "observe", OBSERVER_SIGNATURE);
		jni_info->mid_wr = (*jenv)->GetMethodID(jenv, wrcls, "write", WRITER_SIGNATURE);
		jni_info->jenv = jenv;
		jni_info->jobj = jobj;
		jni_info->info = info;
		jni_info->jobserver = jobserver;
		jni_info->jwriter = jwriter;
		jni_info->jbuf = (*jenv)->NewByteArray(jenv, _SBUF_SIZE);
		len = min(sizeof(path) - 1, (*jenv)->GetArrayLength(jenv, jpath));
		(*jenv)->GetByteArrayRegion(jenv, jpath, 0, len, path);
		*(path + len) = 0;

		_set_data_type(info->dcs, _DTYPE_JNI_FILE);
		sts = dcs_convert_file(info->dcs, path, jstype, jdtype, jisForced, _jdcs_observer, _jdcs_writer, user_info);
		_clear_data_type(info->dcs, _DTYPE_JNI_FILE);

		free(info->dcs->_jni_info);
		info->dcs->_jni_info = NULL;
	}
	return SET_AND_RETURN_STS(jenv, jobj, info, sts);
}
